/*
 * (c) Copyright 2003 - 2005 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "bfirconf.h"
#include "elog.h"
#include "bit.h"

#if BF_MAXFILTERS > VCU_MAX_CHANNELS
#define FILTER_MASK_SIZE BF_MAXFILTERS
#else
#define FILTER_MASK_SIZE VCU_MAX_CHANNELS
#endif

struct filt2filt_scale {
    int filt;
    int outfilt;
    double scale;
};

struct filter {
    struct {
#define FTYPE_PHYS_EQ 1
#define FTYPE_LOGI_EQ 2
#define FTYPE_XTC 3
#define FTYPE_CONNECT 4
#define FTYPE_REVERB 5        
        int type;
        int arg[3];
    } debug;
    int index;
    int coeff;
#define CONN_FILTER 0
#define CONN_VIRTUAL 1    
    uint32_t connidx[2][2][FILTER_MASK_SIZE/32];
};

static char *
strmask(const uint32_t mask[],
        int size)
{
    static char *str[8] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
    static int idx = 0;
    char *p;
    int n;

    if (idx == 8) {
        idx = 0;
    }
    str[idx] = realloc(str[idx], size * 10);
    p = str[idx];
    n = -1;
    while ((n = bit_find(mask, n + 1, size - 1)) != -1) {
        sprintf(p, "%d ", n);
        p += strlen(p);
    }
    if (p != str[idx]) {
        p[-1] = '\0';
    }
    return str[idx++];
}

static const char *
strtype(int type)
{
    switch (type) {
    case VCU_CH_AMBIOPOLE: return "ambiopole";
    case VCU_CH_REVERB: return "reverb";
    case VCU_CH_AMBISONIC: return "ambisonic";
    case VCU_CH_PASSTHRU: return "passthru";
    case VCU_CH_PHYSICAL: return "physical";
    }
    return "#invalid#";
}

static char *
filter_debug_str(const struct filter *f)
{
    static char str[128];
    
    switch (f->debug.type) {
    case FTYPE_PHYS_EQ:
        sprintf(str, "physical %s eq for phys %d -> virt %d",
                f->debug.arg[0] == IN ? "input" : "output",
                f->debug.arg[1], f->debug.arg[2]);
        break;
    case FTYPE_LOGI_EQ:
        sprintf(str, "logical %s eq for type %s index %d",
                f->debug.arg[0] == IN ? "input" : "output",
                strtype(f->debug.arg[1]), f->debug.arg[2]);
        break;
    case FTYPE_XTC:
        sprintf(str, "xtc filter %d/4 for ambiopole index %d",
                f->debug.arg[1] + 1, f->debug.arg[0]);
        break;
    case FTYPE_CONNECT:
        sprintf(str, "dummy filter init by virt link %d -> %d",
                f->debug.arg[0], f->debug.arg[1]);
        break;
    case FTYPE_REVERB:
        sprintf(str, "reverb filter %d/%d for reverb index %d",
                f->debug.arg[1] + 1, f->debug.arg[2], f->debug.arg[0]);
        break;
    default:
        sprintf(str, "INVALID FILTER DEBUG DATA");
        break;
    }
    return str;
}

static void
print_filter(FILE *stream,
             const struct brutefir_config *config,
             const struct runtime_map *rmap,
             const struct filt2filt_scale scales[],
             int scales_count,
             const struct filter *f)
{
    char s[20 * FILTER_MASK_SIZE], *p;
    int io, n, i, conn;
    double scale;
    
    fprintf(stream, "filter %d { # %s\n"
            "  coeff: %d;\n  process: -1;\n  delay: 0;\n",
            f->index, filter_debug_str(f), f->coeff);
    FOR_IN_AND_OUT {
        for (conn = 0; conn < 2; conn++) {
            n = -1;
            p = s;
            while ((n = bit_find(f->connidx[io][conn], n + 1,
                                 FILTER_MASK_SIZE - 1)) != -1)
            {
                scale = 1.0;
                if (conn == CONN_VIRTUAL) {
                    if (io == OUT) {
                        scale = rmap->out_scales[f->index][n];
                    }
                    if (config->ampl[io][n] == 0.0) {                        
                        if (scale != 1.0) {
                            sprintf(p, "%d//%g, ", n, scale);
                        } else {
                            sprintf(p, "%d, ", n);
                        }
                    } else {
                        if (scale != 1.0) {
                            sprintf(p, "%d/%g/%g, ", n, -config->ampl[io][n],
                                    scale);
                        } else {
                            sprintf(p, "%d/%g, ", n, -config->ampl[io][n]);
                        }
                    }
                } else {
                    if (io == IN) {
                        for (i = 0; i < scales_count; i++) {
                            if (scales[i].outfilt == f->index &&
                                scales[i].filt == n)
                            {
                                scale *= scales[i].scale;
                            }
                        }
                    }
                    if (scale != 1.0) {
                        sprintf(p, "%d//%g, ", n, scale);
                    } else {
                        sprintf(p, "%d, ", n);
                    }
                }
                p += strlen(p);
            }
            if (p != s) {
                p[-2] = '\0';
                fprintf(stream, "  %s_%s: %s;\n", io == IN ? "from" : "to",
                        conn == CONN_VIRTUAL ?
                        (io == IN ? "inputs" : "outputs") : "filters", s);
            }
        }
    }
    fprintf(stream, "};\n");
}

static void
write_coeffs_config(FILE *stream,
                    const struct brutefir_config *config,
                    const struct brutefir_coeff coeffs[],
                    int n_coeffs)
{
    int n, i;

    for (n = 0; n < n_coeffs; n++) {
        fprintf(stream, "coeff %d { # %s id %d index %d (part %d of %d)\n\
  format: \"processed\";\n\
  attenuation: 0.0;\n\
  blocks: %d;\n\
  filename: ",
                n,
                coeffs[n].type < 0 ? "dummy" :
                bfirdata_typename(coeffs[n].type),
                coeffs[n].id,
                coeffs[n].index,
                coeffs[n].sub_index + 1,
                coeffs[n].type < 0 ? 1 : bfirdata_filter_count(coeffs[n].type),
                coeffs[n].n_blocks);
        for (i = 0; i < coeffs[n].n_segments - 1; i++) {
            fprintf(stream, "%d/%d/%d, ",
                    coeffs[n].segids[i],
                    coeffs[n].segoffsets[i],
                    coeffs[n].seglengths[i]);
        }
        fprintf(stream, "%d/%d/%d;\n};\n",
                coeffs[n].segids[i],
                coeffs[n].segoffsets[i],
                coeffs[n].seglengths[i]);        
    }
    if (config->dyneq_count > 0) {
        for (n = 0; n < config->dyneq_count; n++) {
            
            fprintf(stream, "coeff %d { # dyneq shadow\n\
  format: \"processed\";\n\
  shared_mem: true;\n\
  attenuation: 0.0;\n\
  blocks: %d;\n\
  filename: \"dirac pulse\";\n\
};\n",
                    n_coeffs + n,
                    config->dyneq[n]->blocks);
        }
    }
}

static bool_t
connect_io(struct filter *virt2filter[2][VCU_CH_COUNT][VCU_MAX_CHANNELS],
           struct filter filters[],
           int *filter_index,
           int type,
           const uint32_t vmap_in[],
           const uint32_t vmap_out[],
           double outscale,
           struct filt2filt_scale **scales,
           int *scales_count,
           struct runtime_map *rmap)
{
    struct filter *f;
    int i, j;

    elog(LOGDEBUG, "connect inputs <%s> to outputs <%s>",
         strmask(vmap_in, VCU_MAX_CHANNELS),
         strmask(vmap_out, VCU_MAX_CHANNELS));
    
    i = -1;
    while ((i = bit_find(vmap_in, i + 1, VCU_MAX_CHANNELS - 1)) != -1) {
        j = -1;
        while ((j = bit_find(vmap_out, j + 1, VCU_MAX_CHANNELS - 1)) != -1) {
            if (virt2filter[IN][type][i] == NULL &&
                virt2filter[OUT][type][j] == NULL)
            {
                /* we need a dummy filter to be able to connect in with out */
                if (*filter_index >= BF_MAXFILTERS) {
                    return -1;
                }
                elog(LOGDEBUG, "created dummy filter %d to connect input %d "
                     "with output %d", *filter_index, i, j);
                f = &filters[*filter_index];
                f->debug.type = FTYPE_CONNECT;
                f->debug.arg[0] = i;
                f->debug.arg[1] = j;
                f->index = *filter_index;
                f->coeff = -1;
                virt2filter[IN][type][i] = f;
                virt2filter[OUT][type][j] = f;
                (*filter_index)++;
            }
                        
            if (virt2filter[IN][type][i] != NULL &&
                virt2filter[OUT][type][j] != NULL)
            {
                if (virt2filter[IN][type][i] == virt2filter[OUT][type][j]) {
                    elog(LOGDEBUG, "input %d and output %d added to filter %d",
                         i, j, virt2filter[IN][type][i]->index);
                    f = virt2filter[IN][type][i];
                    bit_set(f->connidx[IN][CONN_VIRTUAL], i);
                    bit_set(f->connidx[OUT][CONN_VIRTUAL], j);
                    bit_set(rmap->virt2filt[IN][i], f->index);
                    bit_set(rmap->virt2filt[OUT][j], f->index);
                    rmap->out_scales[f->index][j] = outscale;
                } else {
                    elog(LOGDEBUG, "input %d filter %d connected to output %d "
                         "filter %d", i, virt2filter[IN][type][i]->index,
                         j, virt2filter[OUT][type][j]->index);
                    bit_set(virt2filter[IN][type][i]->connidx[OUT][CONN_FILTER],
                            virt2filter[OUT][type][j]->index);
                    bit_set(virt2filter[OUT][type][j]->connidx[IN][CONN_FILTER],
                            virt2filter[IN][type][i]->index);
                    if (outscale != 1.0) {
                        *scales = realloc(*scales, (*scales_count + 1) *
                                          sizeof(struct filt2filt_scale));
                        (*scales)[*scales_count].filt =
                            virt2filter[IN][type][i]->index;
                        (*scales)[*scales_count].outfilt =
                            virt2filter[OUT][type][j]->index;
                        (*scales)[*scales_count].scale = outscale;
                        (*scales_count)++;
                    }
                }
            } else if (virt2filter[IN][type][i] != NULL) {
                elog(LOGDEBUG, "connect output %d to input %d filter %d",
                     j, i, virt2filter[IN][type][i]->index);
                bit_set(virt2filter[IN][type][i]->connidx[OUT][CONN_VIRTUAL],
                        j);
                bit_set(rmap->virt2filt[OUT][j],
                        virt2filter[IN][type][i]->index);
                rmap->out_scales[virt2filter[IN][type][i]->index][j] = outscale;
            } else { /* virt2filter[OUT][type][j] != NULL */
                elog(LOGDEBUG, "connect input %d to output %d filter %d",
                     i, j, virt2filter[OUT][type][j]->index);
                bit_set(virt2filter[OUT][type][j]->connidx[IN][CONN_VIRTUAL],
                        i);
                bit_set(rmap->virt2filt[IN][i],
                        virt2filter[OUT][type][j]->index);
                rmap->out_scales[virt2filter[OUT][type][j]->index][j] =
                    outscale;
            }
        }
    }
    return true;
}

static void
map_filter(const struct brutefir_config *config,
           struct filter *virt2filter[2][VCU_CH_COUNT][VCU_MAX_CHANNELS],
           struct filter *f,
           int type,
           int index_in,
           int index_out,
           struct runtime_map *rmap)
{
    int io, i;
    
    i = -1;
    FOR_IN_AND_OUT {
        while ((i = bit_find(config->chmap.logi2virt[io][type]
                             [io == IN ? index_in : index_out],
                             i + 1, config->chmap.virt_count[io] - 1)) != -1)
        {
            if (virt2filter[io][type][i] != NULL) {
                bit_set(f->connidx[io][CONN_FILTER],
                        virt2filter[io][type][i]->index);
                bit_set(virt2filter[io][type][i]->connidx[!io][CONN_FILTER],
                        f->index);
            } else {
                bit_set(f->connidx[io][CONN_VIRTUAL], i);
                bit_set(rmap->virt2filt[io][i], f->index);
            }
        }
    }    
}

bool_t
bfirconf_config_write(FILE *stream,
                      const struct brutefir_config *config,
                      const struct brutefir_coeff coeffs[],
                      int n_coeffs,
                      struct runtime_map *rmap)
{
    struct filter *virt2filter[2][VCU_CH_COUNT][VCU_MAX_CHANNELS], *filters, *f;
    int n, i, j, io, filter_offset, inputs, used_channels, type, scales_count;
    int newmap[VCU_MAX_CHANNELS], physmap[VCU_MAX_CHANNELS], filter_index;
    int virt_count, virt_base, phys_base;
    uint32_t vmap[VCU_MAX_CHANNELS/32];
    bool_t virt_is_phys, found_link;
    struct filt2filt_scale *scales;

    elog(LOGDEBUG, "bfirconf_config_write() enter");
    
    filters = malloc(BF_MAXFILTERS * sizeof(struct filter));
    memset(filters, 0, BF_MAXFILTERS * sizeof(struct filter));
    memset(virt2filter, 0, sizeof(virt2filter));
    memset(rmap, 0, sizeof(struct runtime_map));
    scales_count = 0;
    scales = NULL;
    for (n = 0; n < BF_MAXFILTERS; n++) {
        for (i = 0; i < VCU_MAX_CHANNELS; i++) {
            rmap->out_scales[n][i] = 1.0;
        }
    }
    /*
     * Generate filters
     */
    filter_index = 0;
    /* generate physical equalisers (attached to virtual channels) */
    elog(LOGDEBUG, "generate physical equalisers");
    FOR_IN_AND_OUT {
        n = -1;
        while ((n = bit_find(config->eqactive[io][VCU_CH_PHYSICAL], n + 1,
                             config->tot_open_channels[io])) != -1)
        {
            found_link = false;
            i = -1;
            while ((i = bit_find(config->chmap.phys2virt[io][n], i + 1,
                                 config->chmap.virt_count[io] - 1)) != -1)
            {
                found_link = true;
                if (filter_index >= BF_MAXFILTERS) {
                    elog(LOGDEBUG, "Too many filters");
                    free(scales);
                    free(filters);
                    return false;
                }
                bit_set(rmap->eq_phys2filt[io][n], filter_index);
                f = &filters[filter_index];
                f->debug.type = FTYPE_PHYS_EQ;
                f->debug.arg[0] = io;
                f->debug.arg[1] = n;
                f->debug.arg[2] = i;
                f->index = filter_index;
                f->coeff = config->eqcoeffs[io][VCU_CH_PHYSICAL][n];
                bit_set(f->connidx[io][CONN_VIRTUAL], i);
                for (type = 0; type < VCU_CH_COUNT; type++) {
                    virt2filter[io][type][i] = f;
                }
                bit_set(rmap->virt2filt[io][i], filter_index);
                filter_index++;
            }
            if (!found_link) {
                elog(LOGERROR, "Did not find any phys2virt mapping for "
                     "io %d idx %d", io, n);
                free(scales);
                free(filters);
                return false;
            }
        }
    }

    /* generate logical equalisers */
    elog(LOGDEBUG, "generate logical equalisers");
    memset(rmap->eq_logi2filt, -1, sizeof(rmap->eq_logi2filt));
    FOR_IN_AND_OUT {
        for (type = 0; type < VCU_CH_COUNT; type++) {
            n = -1;
            while ((n = bit_find(config->eqactive[io][type], n + 1,
                                 config->channel_count[io][type] - 1)) != -1)
            {
                if (filter_index >= BF_MAXFILTERS) {
                    elog(LOGDEBUG, "Too many filters");
                    free(scales);
                    free(filters);
                    return false;
                }
                rmap->eq_logi2filt[io][type][n] = filter_index;
                f = &filters[filter_index];
                f->debug.type = FTYPE_LOGI_EQ;
                f->debug.arg[0] = io;
                f->debug.arg[1] = type;
                f->debug.arg[2] = n;
                f->index = filter_index;
                f->coeff = config->eqcoeffs[io][type][n];
                found_link = false;
                i = -1;
                while ((i = bit_find(config->chmap.logi2virt[io][type][n],
                                     i + 1, config->chmap.virt_count[io] - 1))
                       != -1)
                {
                    found_link = true;
                    if (virt2filter[io][type][i] != NULL) {
                        bit_set(f->connidx[io][CONN_FILTER],
                                virt2filter[io][type][i]->index);
                        bit_set(virt2filter[io][type][i]->
                                connidx[!io][CONN_FILTER], f->index);
                    } else {
                        bit_set(f->connidx[io][CONN_VIRTUAL], i);
                        bit_set(rmap->virt2filt[io][i], filter_index);
                    }
                    virt2filter[io][type][i] = f;
                }
                if (!found_link) {
                    elog(LOGERROR, "Did not find any logi2virt mapping for "
                         "io %d type %d idx %d", io, type, n);
                    free(scales);
                    free(filters);
                    return false;
                }
                filter_index++;
            }
        }
    }

    /* generate ambiopole filters */
    elog(LOGDEBUG, "generate ambiopole filters");
    memset(rmap->xtc2filt, -1, sizeof(rmap->xtc2filt));
    type = VCU_CH_AMBIOPOLE;
    for (n = 0; n < config->channel_count[OUT][type] >> 1; n++) {
        if (config->xtalk_cancel[n]) {
            if (filter_index + 4 > BF_MAXFILTERS) {
                elog(LOGDEBUG, "Too many filters");
                free(scales);
                free(filters);
                return false;
            }
            /* direct path left input channel to left speaker */
            f = &filters[filter_index];
            f->debug.type = FTYPE_XTC;
            f->debug.arg[0] = n;
            f->debug.arg[1] = 0;
            f->index = filter_index;
            f->coeff = config->xtalkcancel_coeffs[4*n+0];
            map_filter(config, virt2filter, f, type, 2*n+0, 2*n+0, rmap);
            rmap->xtc2filt[n][0] = filter_index;
            filter_index++;
            /* cross-path left input channel to right speaker */
            f = &filters[filter_index];
            f->debug.type = FTYPE_XTC;
            f->debug.arg[0] = n;
            f->debug.arg[1] = 1;
            f->index = filter_index;
            f->coeff = config->xtalkcancel_coeffs[4*n+1];
            map_filter(config, virt2filter, f, type, 2*n+0, 2*n+1, rmap);
            rmap->xtc2filt[n][1] = filter_index;
            filter_index++;
            /* direct path right input channel to right speaker */
            f = &filters[filter_index];
            f->debug.type = FTYPE_XTC;
            f->debug.arg[0] = n;
            f->debug.arg[1] = 2;
            f->index = filter_index;
            f->coeff = config->xtalkcancel_coeffs[4*n+2];
            map_filter(config, virt2filter, f, type, 2*n+1, 2*n+1, rmap);
            rmap->xtc2filt[n][2] = filter_index;
            filter_index++;
            /* cross-path right input channel to left speaker */
            f = &filters[filter_index];
            f->debug.type = FTYPE_XTC;
            f->debug.arg[0] = n;
            f->debug.arg[1] = 3;
            f->index = filter_index;
            f->coeff = config->xtalkcancel_coeffs[4*n+3];
            map_filter(config, virt2filter, f, type, 2*n+1, 2*n+0, rmap);
            rmap->xtc2filt[n][3] = filter_index;
            filter_index++;
        } else {
            for (i = 0; i < 2; i++) {
                if (!connect_io(virt2filter, filters, &filter_index, type,
                                config->chmap.logi2virt[IN][type][2*n+i],
                                config->chmap.logi2virt[OUT][type][2*n+i],
                                1.0, &scales, &scales_count, rmap))
                {
                    elog(LOGDEBUG, "Too many filters");
                    free(scales);
                    free(filters);
                    return false;
                }
            }
        }
    }
    
    /* generate reverb filters */
    elog(LOGDEBUG, "generate reverb filters");
    type = VCU_CH_REVERB;
    for (n = 0; n < config->channel_count[OUT][type]; n++) {
        if (filter_index + config->channel_count[IN][type] > BF_MAXFILTERS) {
            elog(LOGDEBUG, "Too many filters");
            free(scales);
            free(filters);
            return false;
        }
        for (i = 0; i < config->channel_count[IN][type]; i++) {
            f = &filters[filter_index];
            f->debug.type = FTYPE_REVERB;
            f->debug.arg[0] = n;
            f->debug.arg[1] = i;
            f->debug.arg[2] = config->channel_count[IN][type];
            f->index = filter_index;
            f->coeff = config->reverb_coeffs[2*n+i];
            map_filter(config, virt2filter, f, type, i, n, rmap);
            rmap->rev2filt[n][i] = filter_index;
            filter_index++;
        }
    }
    
    /* generate ambisonic filters */
    elog(LOGDEBUG, "generate ambisonic filters");
    type = VCU_CH_AMBISONIC;
    if (config->channel_count[OUT][type] > 0) {
        switch (config->ambisonic_input_format) {
        default:
        case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:
            inputs = 4; /* W X Y Z */
            break;
        case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
            inputs = 9; /* W X Y Z R S T U V */
            break;
        }
        for (i = 0; i < inputs; i++) {
            memset(vmap, 0, sizeof(vmap));            
            for (n = 0; n < config->channel_count[OUT][type]; n++) {
                if (!connect_io(virt2filter, filters, &filter_index, type,
                                config->chmap.logi2virt[IN][type][i],
                                config->chmap.logi2virt[OUT][type][n],
                                config->ambisonic_gains[n * inputs + i],
                                &scales, &scales_count, rmap))
                {
                    elog(LOGDEBUG, "Too many filters");
                    free(scales);
                    free(filters);
                    return false;
                }
            }
        }
    }

    /* generate passthru filters */
    elog(LOGDEBUG, "generate passthru filters");
    type = VCU_CH_PASSTHRU;
    for (n = 0; n < config->channel_count[OUT][type]; n++) {
        if (!connect_io(virt2filter, filters, &filter_index, type,
                        config->chmap.logi2virt[IN][type][n],
                        config->chmap.logi2virt[OUT][type][n],
                        1.0, &scales, &scales_count, rmap))
        {
            elog(LOGDEBUG, "Too many filters");
            free(scales);
            free(filters);
            return false;
        }
    }

    rmap->filter_count = filter_index;

    /*
     * Write general parameters
     */
    elog(LOGDEBUG, "write general parameters");
    if (config->dyneq_count > 0 || config->cli_pipe[0] != -1) {
        fprintf(stream, "logic: ");
    }
    if (config->dyneq_count > 0) {
#warning really use bfirdata from here?        
        filter_offset = bfirdata_coeff_index(-1, 0, 0);
        fprintf(stream, "\n\"eq\" {\n");
        for (n = 0; n < config->dyneq_count; n++) {
            fprintf(stream, "  {\n\
    coeff: %d, %d;\n\
    bands: ",
                    config->dyneq[n]->eq_id,
                    filter_offset + n);
            for (i = 0; i < config->dyneq[n]->n_bands; i++) {
                fprintf(stream, "%.1f%c ", config->dyneq[n]->band[i],
                        i < config->dyneq[n]->n_bands - 1 ?
                        ',' : ';');
            }
            fprintf(stream, "\n    magnitude: ");
            for (i = 0; i < config->dyneq[n]->n_bands; i++) {
                fprintf(stream, "%.1f/%g%c ",
                        config->dyneq[n]->band[i],
                        config->dyneq[n]->amp[i],
                        i < config->dyneq[n]->n_bands - 1 ?
                        ',' : ';');
            }
            fprintf(stream, "\n  };\n");
        }
        fprintf(stream, "},\n");
    }
    if (config->cli_pipe[0] != -1) {
        fprintf(stream, "\"cli\" { port: %d, %d; };\n",
                config->cli_pipe[0], config->cli_pipe[1]);
    }
    fprintf(stream, "sampling_rate: %d;\n\
filter_length: %d,%d;\n\
show_progress: %s;\n\
overflow_warnings: true;\n\
allow_poll_mode: %s;\n\
monitor_rate: %s;\n\
float_bits: %d;\n\
powersave: %s;\n\
modules_path: \"%s\";\n\
convolver_config: \"%s\";\n\
max_dither_table_size: 0;\n\
lock_memory: false;\n\n",
	    config->sample_rate,
	    config->block_length,
	    config->n_blocks,
	    config->show_progress ? "true" : "false",
            config->allow_poll_mode ? "true" : "false",
            config->watch_clock ? "true" : "false",
            config->double_precision ? 64 : 32,
            config->powersave ? "true" : "false",
	    config->modules_path,
            config->convolver_config);
    /*
     * Write coefficent parameters
     */
    elog(LOGDEBUG, "write coefficient parameters");
    write_coeffs_config(stream, config, coeffs, n_coeffs);

    /*
     * Write input/output devices
     */
    elog(LOGDEBUG, "write input/output devices");
    fprintf(stream, "\n\n");

    FOR_IN_AND_OUT {

        if (io == IN) {
            fprintf(stream, "# INPUT\n\n");
        } else {
            fprintf(stream, "# OUTPUT\n\n");
        }
        virt_base = 0;
        phys_base = 0;
        for (j = 0; j < config->device_count[io]; j++) {
            for (n = virt_base; n < config->chmap.virt_count[io]; n++) {
                if (config->chmap.virt2phys[io][n] >=
                    phys_base + config->open_channels[io][j])
                {
                    break;
                }
            }
            if (n == virt_base) {
                phys_base += config->open_channels[io][j];
                continue;
            }
            virt_count = n - virt_base;
            
            if (io == IN) {
                fprintf(stream, "input ");
            } else {
                fprintf(stream, "output ");
            }
            for (n = 0; n < virt_count - 1; n++) {
                fprintf(stream, "%d, ", virt_base + n);
            }
            fprintf(stream, "%d {\n", virt_base + n);
            if (io == OUT) {
                fprintf(stream, "  dither: %s;\n",
                        config->apply_dither ? "true" : "false");
            }
            fprintf(stream, "  device: \"%s\" %s;\n\
  sample: \"%s\";\n\
  maxdelay: %d;\n\
  channels: %d / ",
                    config->device[io][j],
                    config->device_params[io][j],
                    bf_strsampleformat(config->sample_format[io]),
                    config->max_delay[io],
                    config->open_channels[io][j]);
            used_channels = 0;
            virt_is_phys = true;
            for (n = 0; n < config->open_channels[io][j]; n++) {
                i = bit_count(config->chmap.phys2virt[io][phys_base+n],
                              VCU_MAX_CHANNELS);
                switch (i) {
                case 0:
                    continue;
                case 1:
                    break;
                default:
                    virt_is_phys = false;
                    break;
                }
                newmap[used_channels] = n;
                physmap[n] = used_channels;
                used_channels++;
            }
            if (virt_is_phys) {
                for (n = virt_base; n < virt_base + used_channels - 1; n++) {
                    fprintf(stream, "%d, ",
                            config->ext_channel_order[io][j]
                            [config->chmap.virt2phys[io][n]-phys_base]);
            }
            fprintf(stream, "%d;\n",
                    config->ext_channel_order[io][j]
                    [config->chmap.virt2phys[io][n]-phys_base]);
            } else {
                for (n = 0; n < used_channels - 1; n++) {
                    fprintf(stream, "%d, ",
                            config->ext_channel_order[io][j][newmap[n]]);
                }
                fprintf(stream, "%d;\n",
                        config->ext_channel_order[io][j][newmap[n]]);
            
                fprintf(stream, "  mapping: ");
                for (n = virt_base; n < virt_base + virt_count - 1; n++) {
                    fprintf(stream, "%d, ",
                            physmap[config->chmap.virt2phys[io][n]-phys_base]);
                }
                fprintf(stream, "%d;\n",
                        physmap[config->chmap.virt2phys[io][n]-phys_base]);
            }
        
            fprintf(stream, "  mute: ");
            for (n = virt_base; n < virt_base + virt_count - 1; n++) {
                fprintf(stream, "%s, ", (config->muted[io][n]) ?
                        "true" : "false");
            }
            fprintf(stream, "%s;\n", (config->muted[io][n]) ?
                    "true" : "false");
        
            fprintf(stream, "  delay: ");
            for (n = virt_base; n < virt_base + virt_count - 1; n++) {
                fprintf(stream, "%d, ", config->delays[io][n]);
            }
            fprintf(stream, "%d;\n", config->delays[io][n]);
            fprintf(stream, "};\n\n");
            
            virt_base += virt_count;
            phys_base += config->open_channels[io][j];
        }

        /* fill with dummy device if the main does not have enough physical
           channels */
        if (virt_base < config->chmap.virt_count[io]) {
            if (io == IN) {
                fprintf(stream, "input ");
            } else {
                fprintf(stream, "output ");
            }
            for (n = virt_base; n < config->chmap.virt_count[io] - 1; n++) {
                fprintf(stream, "%d, ", n);
            }
            fprintf(stream, "%d {\n", n);
            if (io == OUT) {
                fprintf(stream, "  dither: false;\n");
                fprintf(stream, "  device: \"file\" "
                        "{ path: \"/dev/null\"; };\n");
            } else {
                fprintf(stream, "  device: \"file\" "
                        "{ path: \"/dev/zero\"; };\n");
            }
            fprintf(stream, "  sample: \"S8\";\n\
  maxdelay: %d;\n\
  channels: %d;\n",
                    config->max_delay[io],
                    config->chmap.virt_count[io] - virt_base);
            fprintf(stream, "  mute: ");
            for (n = virt_base; n < config->chmap.virt_count[io] - 1; n++) {
                fprintf(stream, "%s, ", (config->muted[io][n]) ?
                        "true" : "false");
            }
            fprintf(stream, "%s;\n", (config->muted[io][n]) ? "true" : "false");
            
            fprintf(stream, "  delay: ");
            for (n = virt_base; n < config->chmap.virt_count[io] - 1; n++) {
                fprintf(stream, "%d, ", config->delays[io][n]);
            }
            fprintf(stream, "%d;\n", config->delays[io][n]);
            fprintf(stream, "};\n\n");
        }
    }

    /*
     * Write filters
     */
    elog(LOGDEBUG, "write filters");
    for (n = 0; n < rmap->filter_count; n++) {
        print_filter(stream, config, rmap, scales, scales_count,
                     &filters[n]);
    }
    free(scales);
    free(filters);
    
    elog(LOGDEBUG, "bfirconf_config_write() leave");

    return true;
}
