/*
 * (c) Copyright 2002 - 2003, 2005 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <string.h>
#include <stdio.h>
#include <malloc.h>

#include "alignalloc.h"
#include "defs.h"
#include "log2.h"
#include "elog.h"
#include "fftplan.h"
#include "convolve.h"

struct _convolve_t_ {
    void *fftplan;
    void *ifftplan;    
    void *coeffs;
    void *inbuf;
    void *outbuf;
    void *cbuf;
    int n_fft2;
    int realsize;
    bool_t timedomain;
};

static void
timedomain_convolve(void *dest,
		    void *src1,
		    int src1len,
		    void *src2,
		    int src2len,
                    int realsize)
{
    void *longsrc, *shortsrc;
    int longlen, shortlen;
    int n, i;

    if (src1len > src2len) {
	longsrc = src1;
	longlen = src1len;
	shortsrc = src2;
	shortlen = src2len;	
    } else {
	longsrc = src2;
	longlen = src2len;
	shortsrc = src1;
	shortlen = src1len;	
    }

    if (realsize == 4) {
        for (i = 0; i < longlen + shortlen - 1; i++) {
            ((float *)dest)[i] = 0.0;
            for (n = 0; n < shortlen; n++) {
                if (i - n >= 0 && i - n < longlen) {
                    ((float *)dest)[i] +=
                        ((float *)shortsrc)[n] * ((float *)longsrc)[i - n];
                }
            }
        }
    } else {
        for (i = 0; i < longlen + shortlen - 1; i++) {
            ((double *)dest)[i] = 0.0;
            for (n = 0; n < shortlen; n++) {
                if (i - n >= 0 && i - n < longlen) {
                    ((double *)dest)[i] +=
                        ((double *)shortsrc)[n] * ((double *)longsrc)[i - n];
                }
            }
        }
    }
}


static void
convolve_inplace(void *cbuf,
		 void *coeffs,
		 int n_fft,
                 int realsize)
{
    if (realsize == 4) {
        float a, *b = (float *)cbuf, *c = (float *)coeffs;
        int n, n_fft2 = n_fft >> 1;
        
        b[0] *= c[0];
        for (n = 1; n < n_fft2; n++) {
            a = b[n];
            b[n] = a * c[n] - b[n_fft - n] * c[n_fft - n];
            b[n_fft - n] = a * c[n_fft - n] + b[n_fft - n] * c[n];
        }
        b[n_fft2] *= c[n_fft2];
    } else {
        double a, *b = (double *)cbuf, *c = (double *)coeffs;
        int n, n_fft2 = n_fft >> 1;
        
        b[0] *= c[0];
        for (n = 1; n < n_fft2; n++) {
            a = b[n];
            b[n] = a * c[n] - b[n_fft - n] * c[n_fft - n];
            b[n_fft - n] = a * c[n_fft - n] + b[n_fft - n] * c[n];
        }
        b[n_fft2] *= c[n_fft2];
    }
}

convolve_t *
convolve_new(void *coeffs,
	     int coeffslen,
	     int order,
             int realsize)
{
    convolve_t *c;
    int n_fft = 1 << (order + 1), n;
    double scaled;
    float scalef;

    if ((c = alignalloc(sizeof(convolve_t))) == NULL) {
        elog(LOGERROR, "alloc failed");
	return NULL;
    }
    memset(c, 0, sizeof(convolve_t));
    if (n_fft < 128 || n_fft < 2 * ALIGNMENT || coeffs == NULL ||
        order < 0 || coeffslen < 1 || (realsize != 4 && realsize != 8))
    {
        elog(LOGERROR, "at least one invalid argument %p %d %d %d",
             coeffs, coeffslen, order, realsize);
        return NULL;
    }
    c->timedomain = false;
    c->fftplan = fftplan_get_real(order + 1, false, false, realsize);
    c->ifftplan = fftplan_get_real(order + 1, true, false, realsize);
    c->coeffs = alignalloc(n_fft * realsize);
    c->inbuf = alignalloc(n_fft * realsize);
    c->outbuf = alignalloc(n_fft * realsize);
    c->cbuf = alignalloc(n_fft * realsize);
    c->n_fft2 = n_fft / 2;
    c->realsize = realsize;

    memset(c->coeffs, 0, n_fft / 2 * realsize);
    if (coeffslen >= n_fft / 2) {
	coeffslen = n_fft / 2;
    } else {
        memset(&((uint8_t *)c->coeffs)[(n_fft / 2 + coeffslen) * realsize], 0,
               (n_fft / 2 - coeffslen) * realsize);
    }
    memcpy(&((uint8_t *)c->coeffs)[n_fft / 2 * realsize], coeffs,
           coeffslen * realsize);

    if (realsize == 4) {
        fftwf_execute_r2r(c->fftplan, c->coeffs, c->outbuf);
        scalef = 1.0 / (float)n_fft;
        for (n = 0; n < n_fft; n++) {
            ((float *)c->coeffs)[n] = ((float *)c->outbuf)[n] * scalef;
        }
    } else {
        fftw_execute_r2r(c->fftplan, c->coeffs, c->outbuf);
        scaled = 1.0 / (double)n_fft;
        for (n = 0; n < n_fft; n++) {
            ((double *)c->coeffs)[n] = ((double *)c->outbuf)[n] * scaled;
        }
    }
    memset(c->outbuf, 0, n_fft * realsize);
    memset(c->inbuf, 0, n_fft * realsize);
    memset(c->cbuf, 0, n_fft * realsize);

    return c;
}

void
convolve_delete(convolve_t *c)
{
    if (c == NULL) {
	return;
    }
    /* we use static fft plans, so we do not destroy them */
    free(c->coeffs);
    free(c->inbuf);
    free(c->outbuf);
    free(c->cbuf);
    free(c);
}

void
convolve_block(convolve_t *c,
	       void *block)
{
    if (((ptrdiff_t)block & (ALIGNMENT - 1)) != 0) {
        elog(LOGERROR, "invalid alignment");
        exit(EXIT_FAILURE);
    }
    memcpy(c->cbuf, &((uint8_t *)c->cbuf)[c->n_fft2 * c->realsize],
           c->n_fft2 * c->realsize);
    memcpy(&((uint8_t *)c->cbuf)[c->n_fft2 * c->realsize], block,
           c->n_fft2 * c->realsize);
    
    if (c->realsize == 4) {
        fftwf_execute_r2r(c->fftplan, c->cbuf, c->inbuf);
        convolve_inplace(c->inbuf, c->coeffs, c->n_fft2 << 1, c->realsize);
        fftwf_execute_r2r(c->ifftplan, c->inbuf, c->outbuf);
    } else {
        fftw_execute_r2r(c->fftplan, c->cbuf, c->inbuf);
        convolve_inplace(c->inbuf, c->coeffs, c->n_fft2 << 1, c->realsize);
        fftw_execute_r2r(c->ifftplan, c->inbuf, c->outbuf);
    }
    
    memcpy(block, c->outbuf, c->n_fft2 * c->realsize);
}

void *
convolve(void *coeffs1,
	 int coeffs1len,
	 void *coeffs2,
	 int coeffs2len,
         int realsize)
{
    int len, blocksize, n_blocks;
    convolve_t *c;
    void *dest;

    if (coeffs1 == NULL || coeffs2 == NULL) {
        elog(LOGERROR, "null input");
        return NULL;
    }
    if (realsize != 4 && realsize != 8) {
        elog(LOGERROR, "invalid realsize %d", realsize);
        return NULL;
    }
    if (((ptrdiff_t)coeffs1 & (ALIGNMENT - 1)) != 0 ||
        ((ptrdiff_t)coeffs2 & (ALIGNMENT - 1)) != 0)
    {
        elog(LOGERROR, "invalid alignment");
        return NULL;
    }
    if (coeffs1len > coeffs2len) {
        len = coeffs1len;
        coeffs1len = coeffs2len;
        coeffs2len = len;
        dest = coeffs1;
        coeffs1 = coeffs2;
        coeffs2 = dest;
    }

    if (coeffs1len < 128 || coeffs1len < 2 * ALIGNMENT) {
        /* use time-domain convolution */
        if ((dest = alignalloc((coeffs1len + coeffs2len - 1) * realsize))
            == NULL)
        {
            return NULL;
        }
        timedomain_convolve(dest, coeffs1, coeffs1len, coeffs2, coeffs2len,
                            realsize);

        return dest;
    }

    blocksize = log2_roof(coeffs1len);
    if ((c = convolve_new(coeffs1, coeffs1len, blocksize, realsize)) == NULL) {
        return NULL;
    }
    blocksize = 1 << blocksize;

    n_blocks = (coeffs1len + coeffs2len + blocksize - 1) / blocksize;
    len = n_blocks * blocksize;
    if ((dest = alignalloc(len * realsize)) == NULL) {
        convolve_delete(c);
        return NULL;
    }
    memcpy(dest, coeffs2, coeffs2len * realsize);
    memset(&((uint8_t *)dest)[coeffs2len * realsize], 0,
           (len - coeffs2len) * realsize);
    for (len = 0; len < n_blocks; len++) {
        convolve_block(c, &((uint8_t *)dest)[len * blocksize * realsize]);
    }
    convolve_delete(c);
    /*
      we want aligned memory, so we don't use realloc()
    dest = realloc(dest, (coeffs1len + coeffs2len - 1) * sizeof(real_t));
    */

    /*
    {
        static int cc = 0;
        real_t *dst, err;
        int n;
        if ((dst = malloc((coeffs1len + coeffs2len - 1) * sizeof(real_t)))
            == NULL)
        {
            return NULL;
        }
        timedomain_convolve(dst, coeffs1, coeffs1len, coeffs2, coeffs2len);
        {
            FILE *stream;
            char s[200];
            sprintf(s, "convolve_td_%d.txt", cc);
            stream = fopen(s, "wt");
            for (n = 0; n < coeffs1len + coeffs2len - 1; n++) {
                fprintf(stream, "%.16e\n", (double)dst[n]);
            }
            fclose(stream);
            sprintf(s, "convolve_fft_%d.txt", cc);
            stream = fopen(s, "wt");
            for (n = 0; n < coeffs1len + coeffs2len - 1; n++) {
                fprintf(stream, "%.16e\n", (double)dest[n]);
            }
            fclose(stream);
            sprintf(s, "coeffs1_%d.txt", cc);
            stream = fopen(s, "wt");
            for (n = 0; n < coeffs1len; n++) {
                fprintf(stream, "%.16e\n", (double)coeffs1[n]);
            }
            fclose(stream);
            sprintf(s, "coeffs2_%d.txt", cc);
            stream = fopen(s, "wt");
            for (n = 0; n < coeffs2len; n++) {
                fprintf(stream, "%.16e\n", (double)coeffs2[n]);
            }
            fclose(stream);
            cc++;
        }
        free(dst);
    }
    */
    
    return dest;
}
