/*
 * (c) Copyright 2002 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#ifndef _COPYFILE_H_
#define _COPYFILE_H_

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <zlib.h>

#include "almusvcu.h"

static inline void
_closestream(FILE *stream,
             gzFile gzfile)
{
    if (stream != NULL) {
        fclose(stream);
    }
    if (gzfile != NULL) {
        fclose(stream);
    }
}

static inline int
copystream(FILE *dest,
           gzFile gzdest,
           FILE *source,
           gzFile gzsource,
           int source_size,
           bool_t close_files)
{
    uint8_t buf[1024];
    size_t bytes, rbytes;
    
    while (true) {
        rbytes = 1024;
        if (source_size >= 0 && rbytes > source_size) {
            rbytes = source_size;
        }
        if (gzsource != NULL) {
            if ((bytes = gzread(gzsource, buf, rbytes)) < 0) {
                if (close_files) {
                    _closestream(source, gzsource);
                    _closestream(dest, gzdest);
                }
                return VCU_ERR_READ_FAILED;
            }
        } else {
            if ((bytes = fread(buf, 1, rbytes, source)) < 0) {
                if (close_files) {
                    _closestream(source, gzsource);
                    _closestream(dest, gzdest);
                }
                return VCU_ERR_READ_FAILED;
            }
        }
        if (bytes == 0) {
            break;
        }
        if (gzdest != NULL) {
            if ((bytes = gzwrite(gzdest, buf, bytes)) < 0) {
                if (close_files) {
                    _closestream(source, gzsource);
                    _closestream(dest, gzdest);
                }
                return VCU_ERR_WRITE_FAILED;
            }
        } else {
            if (fwrite(buf, 1, bytes, dest) != bytes) {
                if (close_files) {
                    _closestream(source, gzsource);
                    _closestream(dest, gzdest);
                }
                return VCU_ERR_WRITE_FAILED;
            }
        }
        source_size -= bytes;
    }
    if (close_files) {
        _closestream(source, gzsource);
        if (gzdest != NULL) {
            if (gzflush(dest, Z_FINISH) != Z_OK) {
                _closestream(dest, gzdest);
                return VCU_ERR_WRITE_FAILED;
            }
        } else {
            if (fflush(dest) != 0) {
                _closestream(dest, gzdest);
                return VCU_ERR_WRITE_FAILED;
            }
            if (fsync(fileno(dest)) == -1) {
                return VCU_ERR_WRITE_FAILED;
            }
        }
    }
    if (close_files) {
        _closestream(dest, gzdest);
    }

    return 0;    
}

static inline int
copyfile(const char dest_path[],
         const char source_path[])
{
    FILE *dest, *source;
    int err;

    if (strcmp(dest_path, source_path) == 0) {
        return 0;
    }
    if ((source = fopen(source_path, "rb")) == NULL) {
	return VCU_ERR_READ_FAILED;
    }
    if ((dest = fopen(dest_path, "wb")) == NULL) {
	return VCU_ERR_WRITE_FAILED;
    }
    if ((err = copystream(dest, NULL, source, NULL, -1, true)) != 0) {
        remove(dest_path);
    }
    return err;
} 

#endif
