/*
 * (c) Copyright 2002 - 2003 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <inttypes.h>
#include <string.h>

#include <fftw3.h>

#include "alignalloc.h"
#include "fftplan.h"
#include "elog.h"
#include "bit.h"

static void *
create_fft_plan(int length,
                bool_t invert,
                bool_t inplace,
                bool_t docomplex,
                int realsize)
{
    void *plan, *buf[2];
    int mul;

    mul = docomplex ? 2 : 1;

    buf[0] = alignalloc(length * realsize * mul);
    memset(buf[0], 0, length * realsize * mul);
    buf[1] = buf[0];
    if (!inplace) {
        buf[1] = alignalloc(length * realsize * mul);
        memset(buf[1], 0, length * realsize * mul);
    }
    if (docomplex) {
        if (realsize == 4) {
            if (invert) {
                plan = fftwf_plan_dft_r2c_1d(length, buf[0], buf[1],
                                             FFTW_ESTIMATE);
            } else {
                plan = fftwf_plan_dft_c2r_1d(length, buf[0], buf[1],
                                             FFTW_ESTIMATE);
            }
        } else {
            if (invert) {
                plan = fftw_plan_dft_r2c_1d(length, buf[0], buf[1],
                                            FFTW_ESTIMATE);
            } else {
                plan = fftw_plan_dft_c2r_1d(length, buf[0], buf[1],
                                            FFTW_ESTIMATE);
            }
        }
    } else {
        if (realsize == 4) {
            plan = fftwf_plan_r2r_1d(length, buf[0], buf[1],
                                     invert ? FFTW_HC2R : FFTW_R2HC,
                                     FFTW_ESTIMATE);
        } else {
            plan = fftw_plan_r2r_1d(length, buf[0], buf[1],
                                    invert ? FFTW_HC2R : FFTW_R2HC,
                                    FFTW_ESTIMATE);
        }
    }
    free(buf[0]);
    if (!inplace) {
        free(buf[1]);
    }
    return plan;
}

/* Get FFTW real-to-complex plan for a given order.
   FFT length == pow(2,order) */
void *
fftplan_get_real(int order,
                 bool_t invert,
                 bool_t inplace,
                 int realsize)
{
    static void *fftplans[2][2][2][32];
    static uint32_t fftplan_generated[2][2][2] =
    { { { 0, 0 }, { 0, 0 } }, { { 0, 0 }, { 0, 0 } } };
    int sz;

    if (order < 1 || order > 31) {
        return NULL;
    }
    sz = (realsize == 4) ? 0 : 1;
    if (!bit_isset(&fftplan_generated[sz][!!invert][!!inplace], order)) {
        fftplans[sz][!!invert][!!inplace][order] =
            create_fft_plan(1 << order, invert, inplace, false,
                            realsize);
    }
    return fftplans[sz][!!invert][!!inplace][order];
}

void *
fftplan_get_complex(int order,
                    bool_t invert,
                    bool_t inplace,
                    int realsize)
{
    static void *fftplans[2][2][2][32];
    static uint32_t fftplan_generated[2][2][2] =
    { { { 0, 0 }, { 0, 0 } }, { { 0, 0 }, { 0, 0 } } };
    int sz;

    if (order < 1 || order > 31) {
        return NULL;
    }
    sz = (realsize == 4) ? 0 : 1;
    if (!bit_isset(&fftplan_generated[sz][!!invert][!!inplace], order)) {
        fftplans[sz][!!invert][!!inplace][order] =
            create_fft_plan(1 << order, invert, inplace, true,
                            realsize);
    }
    return fftplans[sz][!!invert][!!inplace][order];
}


