/*
 * (c) Copyright 2002 - 2003 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <stdarg.h>
#include <inttypes.h>

#include "defs.h"
#include "textwin.h"
#include "almusvcu.h"
#include "elog.h"

#define XSIZE 20
#define YSIZE 4

#define MO_CMD 0xFE

#define MO_LINE_WRAP_ON 0x43
#define MO_LINE_WRAP_OFF 0x44
#define MO_SCROLL_ON 0x51
#define MO_SCROLL_OFF 0x52
#define MO_BACKLIGHT_ON 0x42 /* 1 byte value 1 - 100 minutes, 0 = indefinite */
#define MO_BACKLIGHT_OFF 0x46
#define MO_CLEAR 0x58
#define MO_CONTRAST 0x50 /* 1 byte value, 0 light, 255 dark */
#define MO_CURSOR_ON 0x4A
#define MO_CURSOR_OFF 0x4B
#define MO_CURSOR_LEFT 0x4C
#define MO_CURSOR_RIGHT 0x4D
#define MO_CURSOR_BLINK_ON 0x53
#define MO_CURSOR_BLINK_OFF 0x54
#define MO_CUSTOM_CHAR 0x4E /* 9 bytes, char 0 - 7, 8 byte character data */
#define MO_OUTPUT_ON 0x56 /* 1 byte value identifying the output */
#define MO_OUTPUT_OFF 0x57 /* 1 byte value identifying the output */
#define MO_CURSOR_GOTO 0x47 /* 2 bytes, column 1 - 20, row 1 - 4 */
#define MO_CURSOR_TOPLEFT 0x48
#define MO_INIT_LARGE_DIGITS 0x6E
#define MO_PLACE_LARGE_DIGIT 0x23 /* 2 bytes, column 1 - 18, digit 0 - 9 */
#define MO_INIT_HOR_GRAPH 0x68
#define MO_MAKE_HOR_GRAPH 0x7C /* 3 bytes, column 1 - 20, row 1 - 4, direction
				  (0=left, 1=right), length 0 - 100 */
#define MO_INIT_THICK_VERT_GRAPH 0x76
#define MO_INIT_THIN_VERT_GRAPH 0x73
#define MO_MAKE_VERT_GRAPH 0x3D /* 2 bytes, column 1 - 20, height 0 - 32 */

static char *lcd_device = NULL;
static bool_t nicefont = false;
static int fd = -1;

static void
putcmd(int cmd,
       ...)
{
    uint8_t data[11];
    int len;
    va_list ap;
    
    data[0] = MO_CMD;
    data[1] = (uint8_t)cmd;
    va_start(ap, cmd);
    switch (cmd) {
    case MO_BACKLIGHT_ON:
    case MO_CONTRAST:
    case MO_OUTPUT_ON:
    case MO_OUTPUT_OFF:
	data[2] = (uint8_t)va_arg(ap, int);
	len = 3;
	break;
    case MO_CURSOR_GOTO:
    case MO_PLACE_LARGE_DIGIT:
    case MO_MAKE_VERT_GRAPH:
	data[2] = (uint8_t)va_arg(ap, int);
	data[3] = (uint8_t)va_arg(ap, int);
	len = 4;
	break;
    case MO_MAKE_HOR_GRAPH:
	data[2] = (uint8_t)va_arg(ap, int);
	data[3] = (uint8_t)va_arg(ap, int);
	data[4] = (uint8_t)va_arg(ap, int);
	len = 5;
	break;
    case MO_CUSTOM_CHAR:
	data[2] = (uint8_t)va_arg(ap, int);
	data[3] = (uint8_t)va_arg(ap, int);
	data[4] = (uint8_t)va_arg(ap, int);
	data[5] = (uint8_t)va_arg(ap, int);
	data[6] = (uint8_t)va_arg(ap, int);
	data[7] = (uint8_t)va_arg(ap, int);
	data[8] = (uint8_t)va_arg(ap, int);
	data[9] = (uint8_t)va_arg(ap, int);
	data[10] = (uint8_t)va_arg(ap, int);
	len = 11;
	break;
    default:
	len = 2;
	break;
    }
    va_end(ap);
    write(fd, data, len);
}

static inline char
replacechar(char c)
{
    if (!nicefont) {
        return c;
    }
    switch (c) {
    case 'g': return 3;
    case 'j': return 4;
    case 'p': return 5;
    case 'q': return 6;
    case 'y': return 7;
    default:  return c;
    }    
}

static void
mo_properties(int xsize_range[2],
              int ysize_range[2],
              char direction_chars[4])
{
    xsize_range[0] = 0;
    xsize_range[1] = XSIZE;
    ysize_range[0] = 0;
    ysize_range[1] = YSIZE;
    direction_chars[0] = 127;
    direction_chars[1] = 126;
    direction_chars[2] = 1;
    direction_chars[3] = 2;
}

static bool_t
mo_init(int xsize,
	int ysize)
{
    struct termios termios;    

    if (xsize > XSIZE || ysize > YSIZE) {
	elog(LOGERROR, "Unsupported size requested.");
	return false;
    }

    if ((fd = open(lcd_device, O_RDWR | O_NOCTTY | O_NONBLOCK)) == -1) {
	elog(LOGERROR, "Failed to open %s: %s.\n", lcd_device, strerror(errno));
	return false;
    }
    tcgetattr(fd, &termios);
    cfmakeraw(&termios);
    cfsetospeed(&termios, B19200);
    cfsetispeed(&termios, B19200);
    tcsetattr(fd, TCSANOW, &termios);

    putcmd(MO_LINE_WRAP_OFF);
    putcmd(MO_SCROLL_OFF);
    putcmd(MO_CURSOR_OFF);
    putcmd(MO_CURSOR_BLINK_OFF);
    putcmd(MO_OUTPUT_OFF, 0);
    putcmd(MO_BACKLIGHT_ON, 0);
    putcmd(MO_CONTRAST, 140);
    putcmd(MO_CLEAR);

    /* we don't use character 0, since that would cause null termination
       problems. */
    
    /* arrow up */
    putcmd(MO_CUSTOM_CHAR, 1,  4, 14, 21,  4,  4,  0,  0,  0);
    /* arrow down */
    putcmd(MO_CUSTOM_CHAR, 2,  0,  0,  4,  4, 21, 14,  4,  0);
    /* nicer looking 'g' */
    putcmd(MO_CUSTOM_CHAR, 3,  0,  0, 15, 17, 17, 15,  1, 14);
    /* nicer looking 'j' */
    putcmd(MO_CUSTOM_CHAR, 4,  2,  0,  6,  2,  2,  2, 18, 12);
    /* nicer looking 'p' */
    putcmd(MO_CUSTOM_CHAR, 5,  0,  0, 30, 17, 17, 30, 16, 16);
    /* nicer looking 'q' */
    putcmd(MO_CUSTOM_CHAR, 6,  0,  0, 15, 17, 17, 15,  1,  1);
    /* nicer looking 'y' */
    putcmd(MO_CUSTOM_CHAR, 7,  0,  0, 17, 17, 17, 15,  1, 14);    
    
    putcmd(MO_CLEAR);

    return true;
}

static void
mo_clear(void)
{
    putcmd(MO_CLEAR);
}

static void
mo_putchar(int x,
	   int y,
	   char c)
{
    putcmd(MO_CURSOR_GOTO, x + 1, y + 1);
    c = replacechar(c);
    write(fd, &c, 1);
}

static void
mo_putstring(int x,
	     int y,
	     const char s[])
{
    int len, n;
    char *p;

    len = strlen(s);
    p = alloca(len);
    for (n = 0; n < len; n++) {
        p[n] = (char)replacechar(s[n]);
    }    
    putcmd(MO_CURSOR_GOTO, x + 1, y + 1);
    write(fd, p, len);
}

static void
mo_brightness(int value)
{
    switch (value) {
    case 1:
	putcmd(MO_BACKLIGHT_ON, 0);
	putcmd(MO_CONTRAST, 140);	
	break;
    case 2:
	putcmd(MO_BACKLIGHT_ON, 0);	
	putcmd(MO_CONTRAST, 190);
	break;	
    case 3:
	putcmd(MO_BACKLIGHT_OFF);
	putcmd(MO_CONTRAST, 140);
	break;	
    case 4:
	putcmd(MO_CONTRAST, 190);
	putcmd(MO_BACKLIGHT_OFF);
	break;	
    default:
	break;	
    }
}

void
vcu_finish(void)
{
    mo_clear();
    putcmd(MO_CURSOR_TOPLEFT);
    putcmd(MO_CURSOR_BLINK_ON);
}

const char *
vcu_name(void)
{
    return "tw_matrix_orbital";
}

static struct textwin_output tw_matrix_orbital = {
    mo_properties,
    mo_putchar,
    mo_init,
    NULL,
    mo_clear,
    mo_putstring,
    mo_brightness
};

void *
vcu_share(void)
{
    return (void *)&tw_matrix_orbital;
}

#define GET_TOKEN(token, error)                                                \
    if (get_config_token(&lval) != token) {                                    \
        elog(LOGERROR, "Parse error: " error );                                \
        return false;                                                          \
    }

bool_t
vcu_preinit(int (*get_config_token)(union vcu_lexval *lexval))
{
    union vcu_lexval lval;
    int tok;

    while ((tok = get_config_token(&lval)) > 0) {
	switch (tok) {
	case VCU_LEXVAL_FIELD:
	    if (strcmp(lval.field, "device") == 0) {
		GET_TOKEN(VCU_LEXVAL_STRING, "expected string for "
                          "device field");
		lcd_device = strdup(lval.field);
		GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
	    } else if (strcmp(lval.field, "nicefont") == 0) {
		GET_TOKEN(VCU_LEXVAL_BOOLEAN, "expected boolean for "
                          "nicechars field");
		nicefont = lval.boolean;
		GET_TOKEN(VCU_LEX_EOS, "expected end of statement (;)");
            }
	    break;            
	default:
	    elog(LOGERROR, "Parse error: expected field.");
	    return false;
	}
    }
    if (lcd_device == NULL) {
        elog(LOGERROR, "\"device\" field not set");
        return false;
    }
    return true;
}

#undef GET_TOKEN
