/*
 * (c) Copyright 2002 - 2006 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <inttypes.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <sched.h>
#include <signal.h>

#include "almusvcu.h"
#include "grammar.h"
#include "elog.h"
#include "sysmon.h"
#include "filterprg.h"
#include "equaliser.h"
#include "bfirfilter.h"
#include "ambisonics.h"
#include "copyfile.h"
#include "simpletar.h"
#include "reverb.h"
#include "idarray.h"
#include "bfir.h"
#include "conf.h"
#include "bit.h"
#include "log2.h"
#include "path.h"
#include "xstr.h"

/*
 * FIXME: get rid of translations between bool_t arrays and bit masks of bfir.h
 * FIXME: fix error handling for memory allocation realloc malloc and stuff
 * FIXME: too much copied code same code for four cases reverb, passthru,
 *        apole and ambisonic see if it is possible to make indexes in
 *        conf_user with type, so instead of having apole_channels,
 *        reverb_channels etc, one has channels[type] instead (and similar).
 * FIXME: the idarray datatype is a bit misplaced now, a list would be better,
 *        or direct editing in conf_user?
 * FIXME: reduce number of programs funcions? program type parameter
 *
 */

#define SIZEOF_CH_MASK (sizeof(uint32_t) * VCU_MAX_CHANNELS/32)

struct revprg {
    struct filterprg prg;
    struct conf_revprg conf;
    struct bfirfilter *bfirfilters;
    double orig_rt60;
    double orig_tail_volume;
    double proc_tail_volume;
    double proc_length;
    double proc_rt60;
    int bfir_id;
    bool_t regenerate;
    bool_t save;
    bool_t truncated;
};

struct xtcprg {
    struct filterprg prg;
    struct bfirfilter bfirfilters[4];
    int sample_delay[VCU_VALID_RATES_COUNT];
    int bfir_id;
};

struct equaliser {
    struct bfirfilter bfirfilter;
    int bfir_id;
    int lengths[VCU_VALID_RATES_COUNT];
    int sample_delay[VCU_VALID_RATES_COUNT];
    struct {
	bool_t updated;
	struct vcu_eqamp amp[VCU_FREQ_ISO_3OCT_BANDS];
        double peak_amp;
	double delay;
    } properties;
};

struct timer_cb {
    void *param;
    void (*callback)(void *param);
    struct timeval tvwhen;
};

struct vcu vcu;

static prgid_t zeroid;
static struct ui_module ui_modules[CONF_UIMODULES_MAX];
static fd_set readfds;
static int maxfd = -1;
static void (*fd_callbacks[FD_SETSIZE])(int);
static bool_t isruntime = false;
static int bf_fd = -1;
static bool_t save_userconf = false;
static bool_t has_run_benchmark = false;
static bool_t has_newly_run_benchmark = false;
static bool_t benchmark_dependent_setting_modified = false;
static void (*process_terminated_cb)(int);
static int runtime_sample_rate; /* valid only during runtime */

static idarray_t *rev_programs = NULL;
static idarray_t *xtc_programs = NULL;
static idarray_t *eq_programs = NULL;
static struct equaliser *equalisers[VCU_MAX_EQ_PROGRAMS];
static struct basic_sysinfo current_sysinfo;
#define MAX_TIMERS 128
static struct timer_cb callback_timers[MAX_TIMERS];
static uint32_t cb_timers_tab[MAX_TIMERS / 32];
static uint32_t cb_timers_freetab[MAX_TIMERS / 32];
static const struct filterprg *load_fprg = NULL;
static struct ui_progress_callbacks progress;
static uint64_t copyfile_size;
static const char *copyfile_dest_path;
static int uhj_bfir_id = -1;
static char static_config_path[PATH_MAX];
static char start_workdir[PATH_MAX];
static char binary_path[PATH_MAX];
static char **start_argv;
static int start_argc;
static bool_t disable_daemon_mode = false;
static int termpipe[2];

static int non_orphan_n_programs;
static struct filterprg *non_orphan_fprgs;
static FILE *save_config_stream;
static bool_t save_config_scandir_failure;
static struct bfir_chmap bfir_chmap[2][VCU_MAX_CHANNELS];

struct eq_mode_delay_alignment {
    int delay[2][VCU_CH_WP_COUNT][VCU_MAX_CHANNELS];
};

static struct {
    /*
      set trim delay in - blind add
      set trim delay out - blind add
      set reverb trim delay / set reverb program - blind add
      set ambiopole
      set eq mode
     */
    struct eq_mode_delay_alignment *eqalign;
    int xtc_maxdelay;
    int delay_buffer_size[2];
    int flowthrough_delay;
} runtime_delay_map;

static struct {
    uint32_t type[VCU_CH_WP_COUNT][VCU_MAX_CHANNEL_MODES/32];
    uint32_t ch[2][VCU_CH_WP_COUNT][VCU_MAX_CHANNELS][VCU_MAX_CHANNEL_MODES/32];
} runtime_active_map;

static struct {
    int pos;
    int cap;
    struct vcu_config *array;
} config_list;

static struct {
    int current_rate;
    int n_alternatives[VCU_VALID_RATES_COUNT];
    int block_sizes[VCU_VALID_RATES_COUNT][VCU_BENCHMARK_MAX_ALTERNATIVES];
    int n_blocks[VCU_VALID_RATES_COUNT][VCU_BENCHMARK_MAX_ALTERNATIVES];
} benchmark;

#define MUST_NOT_BE_IN_RUNTIME(retval)                                         \
    if (isruntime) {                                                           \
        elog(LOGDEBUG, "Cannot do this in runtime");                           \
        return retval;                                                         \
    }
#define MUST_NOT_NEED_TO_RUN_BENCHMARK(retval)                                 \
    if (vcu.iodelay.need_to_run_benchmark()) {                                 \
        elog(LOGDEBUG, "Cannot do this when benchmark need to be performed");  \
        return retval;                                                         \
    }
#define MUST_NOT_NEED_TO_UPDATE_HARD_SETTINGS(retval)                          \
    if (update_hard_settings(true) != 0) {                                     \
        elog(LOGDEBUG, "Cannot do this when hard settings must be updated");   \
        return retval;                                                         \
    }

/*
static void
set_realtime_priority(int prio_sub)
{
    struct sched_param schp;
    
    memset(&schp, 0, sizeof(schp));
    schp.sched_priority = sched_get_priority_max(SCHED_FIFO) - prio_sub;
    
    if (sched_setscheduler(0, SCHED_FIFO, &schp) != 0) {
	elog(LOGERROR, "Could not set realtime priority: %s.\n",
             strerror(errno));
	exit(EXIT_FAILURE);
    }
}
*/

static void
pre_exec_cleanup(void)
{
    int n;

    if (vcu.runtime.isruntime()) {
        bfir_stop();
    }
    for (n = 0; n < conf_static.n_uimodules; n++) {
        if (conf_static.uimodules[n]->finish != NULL) {
            elog(LOGDEBUG, "calling finish() for module \"%s\"",
                 conf_static.uimodules[n]->name());
            conf_static.uimodules[n]->finish();
        }
    }
    bfirmem_finish();
    /* close all file descriptors except stdin, stdout and stderr */
    for (n = 3; n < FD_SETSIZE; n++) {
        close(n);
    }
}

static void
sighandler(int sig)
{
    uint8_t b;

    b = 0;
    write(termpipe[1], &b, 1);
}

static void
periodic(void)
{
    static struct timeval lasttv = { 0, 0 };
    struct filterprg *filterprg;
    char path[PATH_MAX + 100];
    struct revprg *revprg;
    struct xtcprg *xtcprg;
    struct timeval tv;
    struct stat buf;
    int n, i;

    gettimeofday(&tv, NULL);
    tv.tv_sec -= 4;
    if (timercmp(&tv, &lasttv, <)) {
	return;
    }
    tv.tv_sec += 4;
    lasttv = tv;

    /* saving configuration */

    if (vcu.runtime.isruntime() && conf_static.suspend_in_runtime) {
        /* don't save */
        return;
    }

    if (conf_static.log_rotate_size > 0) {
        if (stat(vcu.ctl.get_log_file_path(), &buf) != -1 &&
            buf.st_size >= conf_static.log_rotate_size * 1024)
        {
            elog(LOGDEBUG, "last log before rotating log file");
            elog_close();
            sprintf(path, "%s.old", vcu.ctl.get_log_file_path());
            remove(path);
            rename(vcu.ctl.get_log_file_path(), path);
            elog(LOGDEBUG, "first log after rotating log file");
        }
    }
    
    if (save_userconf) {
	elog(LOGDEBUG, "saving user configuration");
	if (vcu.iodelay.need_to_run_benchmark()) {
	    memset(&cu.reverb.benchmark_sysinfo, 0, 
		   sizeof(cu.reverb.benchmark_sysinfo));
	}
	cu.reverb.n_loaded_programs = idarray_length(rev_programs);
	n = -1; i = 0;
	while ((revprg = idarray_next(rev_programs, &n)) != NULL) {
	    cu.reverb.loaded_programs[i++] = revprg->prg.id;
	}
	cu.ambiopole.n_loaded_programs = idarray_length(xtc_programs);
	n = -1; i = 0;
	while ((xtcprg = idarray_next(xtc_programs, &n)) != NULL) {
	    cu.ambiopole.loaded_programs[i++] = xtcprg->prg.id;
	}
	cu.eq.n_loaded_programs = idarray_length(eq_programs);
	n = -1; i = 0;
	while ((filterprg = idarray_next(eq_programs, &n)) != NULL) {
	    cu.eq.loaded_programs[i++] = filterprg->id;
	}
	conf_save_user();
	save_userconf = false;
    }
    n = -1;
    while ((revprg = idarray_next(rev_programs, &n)) != NULL) {
	if (revprg->save) {
	    revprg->save = false;
	    elog(LOGDEBUG, "saving reverb program user configuration");
	    conf_save_revprg(&revprg->conf);
	}
    }
}

static void
clean_fprog_refs(void)
{
    struct fprogref *fprogrefs[CONF_MAX_FPROGREFS];
    uint32_t bitset[CONF_MAX_FPROGREFS/32+1];
    struct stat buf;
    int n, i;

    memset(bitset, 0, sizeof(bitset));
    memset(fprogrefs, 0, sizeof(fprogrefs));
    for (n = 0; n < cu.eq.n_loaded_programs; n++) {
        for (i = 0; i < cu.ctl.n_fprogrefs; i++) {
            if (memcmp(&cu.eq.loaded_programs[n],
                       &cu.ctl.fprogrefs[i]->id, sizeof(prgid_t)) == 0)
            {
                bit_set(bitset, i);
                break;
            }
        }
    }
    for (n = 0; n < cu.ambiopole.n_loaded_programs; n++) {
        for (i = 0; i < cu.ctl.n_fprogrefs; i++) {
            if (memcmp(&cu.ambiopole.loaded_programs[n],
                       &cu.ctl.fprogrefs[i]->id, sizeof(prgid_t)) == 0)
            {
                bit_set(bitset, i);
                break;
            }
        }
    }
    for (n = 0; n < cu.reverb.n_loaded_programs; n++) {
        for (i = 0; i < cu.ctl.n_fprogrefs; i++) {
            if (memcmp(&cu.reverb.loaded_programs[n],
                       &cu.ctl.fprogrefs[i]->id, sizeof(prgid_t)) == 0)
            {
                bit_set(bitset, i);
                break;
            }
        }
    }

    for (n = 0, i = 0; n < cu.ctl.n_fprogrefs; n++) {
        if (bit_isset(bitset, n) &&
            stat(cu.ctl.fprogrefs[n]->path, &buf) != -1)
        {
            fprogrefs[i++] = cu.ctl.fprogrefs[n];
        } else {
            free(cu.ctl.fprogrefs[n]);
        }
    }
    memcpy(cu.ctl.fprogrefs, fprogrefs, sizeof(fprogrefs));
    cu.ctl.n_fprogrefs = i;
}

static int
add_fprog_ref(prgid_t id,
              const char path[])
{
    struct fprogref *fref;
    
    if (cu.ctl.n_fprogrefs == CONF_MAX_FPROGREFS) {
        return VCU_ERR_NO_MEMORY;
    }
    fref = malloc(sizeof(struct fprogref));
    memset(fref, 0, sizeof(fref));
    fref->id = id;
    strncpy(fref->path, path, sizeof(fref->path) - 1);
    cu.ctl.fprogrefs[cu.ctl.n_fprogrefs] = fref;
    cu.ctl.n_fprogrefs++;
    return 0;
}

static const char *
strtype(int type)
{
    switch (type) {
    case VCU_CH_AMBIOPOLE: return "ambiopole";
    case VCU_CH_REVERB: return "reverb";
    case VCU_CH_AMBISONIC: return "ambisonic";
    case VCU_CH_PASSTHRU: return "passthru";
    case VCU_CH_PHYSICAL: return "physical";
    }
    return "#invalid#";
}

static char *
strmask(const uint32_t mask[],
        int start,
        int end)
{
    static char *str = NULL;
    char *p;
    int n;

    str = realloc(str, (end - start + 1) * 10 + 10);
    p = str;
    p[0] = '{';
    p[1] = ' ';
    p[2] = '\0';
    p += 2;
    n = start - 1;
    while ((n = bit_find(mask, n + 1, end)) != -1) {
        sprintf(p, "%d ", n);
        p += strlen(p);
    }
    p[0] = '}';
    p[1] = '\0';
    return str;
}

static bool_t
program_in_use(const char cachename[],
               prgid_t parsed_id,
               int parsed_type)
{
    struct filterprg *filterprg;
    struct xtcprg *xtcprg;
    struct revprg *revprg;
    char strid[33];
    prgid_t id;
    int type;
    int n;

    if (cachename != NULL) {
        if (strlen(cachename) != 40) {
            return false;
        }
        memcpy(strid, &cachename[4], 32);
        strid[32] = '\0';
        id = filterprg_prgid_aton(strid);
        if (strstr(cachename, "eqr.") == cachename) {
            type = FILTERPRG_TYPE_EQUALISATION;
        } else if (strstr(cachename, "xtc.") == cachename) {
            type = FILTERPRG_TYPE_CROSSTALK_CANCELLATION;
        } else if (strstr(cachename, "rev.") == cachename) {
            /* could be other reverb types as well */
            type = FILTERPRG_TYPE_REVERB_TRADITIONAL;
        } else {
            return false;
        }
    } else {
        id = parsed_id;
        type = parsed_type;
    }        
    n = -1;
    switch (type) {
    case FILTERPRG_TYPE_EQUALISATION:
	while ((filterprg = idarray_next(eq_programs, &n)) != NULL) {
            if (memcmp(&filterprg->id, &id, sizeof(id)) == 0) {
                return true;
            }
	}
        return false;
    case FILTERPRG_TYPE_CROSSTALK_CANCELLATION:
	while ((xtcprg = idarray_next(xtc_programs, &n)) != NULL) {
            if (memcmp(&xtcprg->prg.id, &id, sizeof(id)) == 0) {
                return true;
            }
	}
        return false;
        /* all reverb types should be listed here */
    case FILTERPRG_TYPE_REVERB_TRADITIONAL:
    case FILTERPRG_TYPE_REVERB_BFORMAT:
	while ((revprg = idarray_next(rev_programs, &n)) != NULL) {
            if (memcmp(&revprg->prg.id, &id, sizeof(id)) == 0) {
                return true;
            }
	}
        return false;
    default:
        return false;
    }
}

static bool_t
lexparse_config_header(struct vcu_config *config)
{
    uint32_t isset[1] = { 0 };
    int tok, n;

    while ((tok = yylex()) > 0) {
        if (tok != FIELD) {
	    yy_perror_tok(FIELD, tok);
	    return false;
        }
        if (strcmp(yylval.field, "name") == 0) {
            TEST_AND_SET(0);
            GET_TOKEN(STRING);
            strncpy(config->name, yylval.string, VCU_CONFIG_NAME_SIZE);
            config->name[VCU_CONFIG_NAME_SIZE - 1] = '\0';
            GET_TOKEN(EOS);
        } else if (strcmp(yylval.field, "static_config_path") == 0) {
            TEST_AND_SET(1);
            GET_TOKEN(STRING);
            strncpy(config->static_config_path, yylval.string, PATH_MAX);
            config->name[PATH_MAX - 1] = '\0';
            GET_TOKEN(EOS);
        } else if (strcmp(yylval.field, "cwd") == 0) {
            TEST_AND_SET(2);
            GET_TOKEN(STRING);
            strncpy(config->cwd, yylval.string, PATH_MAX);
            config->name[PATH_MAX - 1] = '\0';
            GET_TOKEN(EOS);
        } else if (strcmp(yylval.field, "argv") == 0) {
            TEST_AND_SET(3);
            for (n = 0; n < VCU_CONFIG_ARGS_MAX; n++) {
                tok = yylex();
                if (tok == EOS) {
                    config->argc = n;
                    break;
                } else if (tok == STRING) {
                    strncpy(config->argv[n], yylval.string,
                            VCU_CONFIG_ARG_SIZE);
                    config->argv[n][VCU_CONFIG_ARG_SIZE - 1] = '\0';
                } else {
                    yy_perror_tok(EOS, tok);
                    return false;
                }
            }
            if (tok != EOS) {
                yy_perror("too many arguments");
                return false;
            }
        } else {
            yy_perror("ignoring field \"%s\"", yylval.string);
            while ((tok = yylex()) != EOS && tok > 0);
        }
    }
    if ((isset[0] & 0x1) == 0) {
        yy_perror("name not set");
        return false;
    }
    return true;
}

static int
revprg_config_scandir_select(const struct dirent *dirent)
{
    if (strstr(dirent->d_name, ".revprg") !=
        &dirent->d_name[strlen(dirent->d_name) - 7])
    {
        return 0;
    }
    return 1;
}

static int
config_scandir_select(const struct dirent *dirent)
{
    struct simple_tar_header tar_hdr;
    struct vcu_config config;
    FILE *stream;
    bool_t b;
    int slen;

    slen = strlen(dirent->d_name);
    if (strstr(dirent->d_name, ".tar") != dirent->d_name + slen - 4) {
	return 0;
    }
    if ((stream = fopen(dirent->d_name, "rb")) == NULL) {
	return 0;
    }
    if (!tar_header_read(stream, NULL, &tar_hdr) ||
        strcmp(tar_hdr.filename, "config.txt") != 0)
    {
	fclose(stream);
	return 0;
    }

    memset(&config, 0, sizeof(config));

    getcwd(config.path, PATH_MAX);
    strcat(config.path, PATH_SEPARATOR_STR);
    strcat(config.path, dirent->d_name);

    yy_restart(config.path, stream, LOGDEBUG);
    yy_input_mode(true, tar_hdr.size);
    b = lexparse_config_header(&config);
    fclose(stream);
    yy_restart(NULL, NULL, LOGDEBUG);
    
    if (!b) {
	return 0;
    }
    if (config_list.pos >= config_list.cap) {
	config_list.cap = config_list.pos + 1;
	config_list.array = realloc(config_list.array,
                                    config_list.cap *
                                    sizeof(struct vcu_config));
    }
    memcpy(&config_list.array[config_list.pos], &config,
           sizeof(struct vcu_config));
    config_list.pos++;
    
    return 1;
}

static int
save_config_scandir_select(const struct dirent *dirent)
{
    char path[PATH_MAX];
    
    if (strstr(dirent->d_name, ".revprg") !=
        &dirent->d_name[strlen(dirent->d_name) - 7])
    {
        return 0;
    }
    sprintf(path, "%s%s", conf_static.storage_path, dirent->d_name);
    if (!tar_add(path, save_config_stream, NULL)) {
        save_config_scandir_failure = true;
    }
    return 1;
}

static int
orphan_resources_scandir_select(const struct dirent *dirent)
{
    char path[PATH_MAX], *p;
    struct stat buf;
    int n;
    
    path[PATH_MAX - 1] = '\0';
    if (strstr(dirent->d_name, ".revprg") ==
        &dirent->d_name[strlen(dirent->d_name) - 7])
    {
        snprintf(path, PATH_MAX - 1, "%srev.%s", conf_static.storage_path,
                 dirent->d_name);
        p = strstr(&path[strlen(conf_static.storage_path)], ".revprg");
        strcpy(p, ".tar");
        if (stat(path, &buf) == -1) {
            /* assume file not found, thus config lacks reverb program, remove
               it. */
            return 1;
        }
        return 0;
    }
    if (strstr(dirent->d_name, ".tar") !=
        &dirent->d_name[strlen(dirent->d_name) - 4] ||
        (strstr(dirent->d_name, "rev.") != dirent->d_name &&
         strstr(dirent->d_name, "xtc.") != dirent->d_name &&
         strstr(dirent->d_name, "eqr.") != dirent->d_name))
    {
        return 0;
    }
    snprintf(path, PATH_MAX - 1, "%s%s", conf_static.storage_path,
             dirent->d_name);
    for (n = 0; n < non_orphan_n_programs; n++) {
        if (strcmp(non_orphan_fprgs[n].path, path) == 0) {
            return 0;
        }
    }
    return 1;
}

static int
copyfile_scandir_select(const struct dirent *dirent)
{
    char path[PATH_MAX];
    struct stat buf;
    
    path[PATH_MAX - 1] = '\0';
    snprintf(path, PATH_MAX - 1, "%s%s", conf_static.storage_path,
             dirent->d_name);
    if (stat(path, &buf) == -1) {
        elog(LOGERROR, "failed to stat file \"%s\": %s", path, strerror(errno));
        return 0;
    }
    if (strstr(dirent->d_name, ".tar") !=
        &dirent->d_name[strlen(dirent->d_name) - 4] ||
        strcmp(path, copyfile_dest_path) == 0)
    {
        return 0;
    }
    copyfile_size += buf.st_size;
    return 1;
}

static int
copyfile_scandir_compare(const struct dirent **a,
                         const struct dirent **b)
{
    bool_t ainuse, binuse;
    char path[PATH_MAX];
    struct stat sa, sb;
    prgid_t dummyid;

    path[PATH_MAX - 1] = '\0';
    snprintf(path, PATH_MAX - 1, "%s%s", conf_static.storage_path,
             a[0]->d_name);
    if (stat(path, &sa) == -1) {
        elog(LOGERROR, "failed to stat file \"%s\": %s", path, strerror(errno));
        return 0;
    }
    snprintf(path, PATH_MAX - 1, "%s%s", conf_static.storage_path,
             b[0]->d_name);
    if (stat(path, &sb) == -1) {
        elog(LOGERROR, "failed to stat file \"%s\": %s", path, strerror(errno));
        return 0;
    }
    memset(&dummyid, 0, sizeof(dummyid));
    ainuse = program_in_use(a[0]->d_name, dummyid, 0);
    binuse = program_in_use(b[0]->d_name, dummyid, 0);
    if (ainuse && binuse) {
        return sa.st_ctime - sb.st_ctime;
    }
    return (int)ainuse - (int)binuse;
}

static int
copyfile_storage(const char dest_path[],
                 const char source_path[])
{
    struct dirent **namelist;
    char path[PATH_MAX];
    int64_t delsize;
    struct stat buf;
    int n, n_files;
    
    if (strstr(dest_path, conf_static.storage_path) == dest_path) {
        if (stat(source_path, &buf) == -1) {
            elog(LOGERROR, "failed to stat file \"%s\": %s",
                 source_path, strerror(errno));
            return VCU_ERR_READ_FAILED;
        }
        if (buf.st_size > (off_t)conf_static.storage_size * 1024 * 1024) {
            return VCU_ERR_OBJECT_TOO_LARGE;
        }
        copyfile_size = buf.st_size;
        copyfile_dest_path = dest_path;
        if ((n_files = scandir(conf_static.storage_path, &namelist,
                               copyfile_scandir_select,
                               (int(*)(const void *, const void *))
                               copyfile_scandir_compare)) < 0)
        {
            elog(LOGDEBUG, "scandir failed: %s", strerror(errno));
            return VCU_ERR_WRITE_FAILED;
        }
        for (n = 0; n < n_files; n++) {
            elog(LOGDEBUG, "\"%s\"", namelist[n]->d_name);
        }
        delsize = (int64_t)copyfile_size -
            (int64_t)conf_static.storage_size * 1024 * 1024;
        elog(LOGDEBUG, "would be %llu bytes, must delete %lld bytes",
             copyfile_size, delsize);
        path[PATH_MAX - 1] = '\0';        
        for (n = 0; n < n_files && delsize > 0; n++) {
            snprintf(path, PATH_MAX - 1, "%s%s", conf_static.storage_path,
                     namelist[n]->d_name);
            free(namelist[n]);
            if (stat(path, &buf) == -1) {
                elog(LOGERROR, "failed to stat file \"%s\": %s",
                     path, strerror(errno));
                continue;
            }
            if (remove(path) == -1) {
                elog(LOGERROR, "failed to remove file \"%s\": %s",
                     path , strerror(errno));
                continue;
            }
            elog(LOGDEBUG, "removed \"%s\" (gained %d bytes, %d bytes left "
                 "to remove)", path, (int)buf.st_size, (int)delsize);
            delsize -= (int64_t)buf.st_size;
        }
        free(namelist);
        if (delsize > 0) {
            elog(LOGDEBUG, "could not delete enough files from cache");
            return VCU_ERR_WRITE_FAILED;
        }
    }
    return copyfile(dest_path, source_path);
}

static void
make_bfir_chmap(struct bfir_chmap *maps[2])
{
    int phys, type, logi, io, phys_count;

    memset(bfir_chmap, 0, sizeof(bfir_chmap));
    FOR_IN_AND_OUT {
        maps[io] = bfir_chmap[io];
        phys_count = cu.channel.count[io][VCU_CH_PHYSICAL];
        for (type = 0; type < VCU_CH_COUNT; type++) {
            for (logi = 0; logi < cu.channel.count[io][type]; logi++) {
                phys = -1;
                while ((phys = bit_find(cu.channel.map[io][type][logi],
                                        phys + 1, phys_count - 1)) != -1)
                {
                    /*
                    elog(LOGDEBUG, "io %d phys(%d)->logi(%d,%d)",
                         io, phys, type, logi);
                    */
                    bit_set(maps[io][phys].channels[type], logi);
                }
            }
        }
    }
}

#define RETURN_CHANNEL_ACTIVE(isactive)                                        \
    elog(LOGDEBUG, "channel [%s][%s][%d] is %s",                               \
         io == IN ? "in" : "out", strtype(type), idx,                          \
         (isactive) ? "is active" : "not active");                             \
    return (isactive);

static bool_t
channel_is_active(const struct vcu_chmode *mode,
                  int runtime_active_mode,
                  bool_t use_runtime_active_mode,
                  int io,
                  int type,
                  int idx)
{
    struct vcu_chgrp *chgrp, grp;
    struct bfir_chmap *maps[2];
    int n, i, logi_type;

    /*
     * Here we answer the question - can sound come out / in through the
     * given channel? Note that this is related to channel modes and mappings,
     * not to mute settings.
     *
     * A logical channel is active if:
     *   - channel modes are not used
     *   - the channel type and index, ignoring io, is included in a group
     *     which is included in the active channel mode
     *
     * A physical channel is active if:
     *   - there is a mapping to a logical channel, and that logical channel
     *     is active
     *
     */
    
    if ((io != IN && io != OUT) || type < 0 || type >= VCU_CH_WP_COUNT ||
        idx < 0 || idx >= cu.channel.count[io][type])
    {
        elog(LOGDEBUG, "invalid params io %d; type %d; idx %d",
             io, type, idx);
        return false;
    }
    if (vcu.runtime.isruntime() && use_runtime_active_mode) {
        n = (runtime_active_mode < 0) ? 0 : runtime_active_mode;
        return !!bit_isset(runtime_active_map.ch[io][type][idx], n);
    }
    
    if (type == VCU_CH_PHYSICAL) {
        make_bfir_chmap(maps);
        for (logi_type = 0; logi_type < VCU_CH_COUNT; logi_type++) {
            n = -1;
            while ((n = bit_find(bfir_chmap[io][idx].channels[logi_type], n + 1,
                                 cu.channel.count[io][logi_type] - 1)) != -1)
            {
                if (channel_is_active(mode, runtime_active_mode,
                                      use_runtime_active_mode, io,
                                      logi_type, n))
                {
                    RETURN_CHANNEL_ACTIVE(true);
                }
            }
        }
        RETURN_CHANNEL_ACTIVE(false);
    }
    
    if (mode == NULL) {
        RETURN_CHANNEL_ACTIVE(true);
    }
    
    /* put all channels that are in a group which is included in the active
       channel mode to 'grp' */
    memset(&grp, 0, sizeof(grp));
    n = -1;
    while ((n = bit_find(mode->groups, n + 1,
                         cu.channel.group_count - 1)) != -1)
    {
        chgrp = &cu.channel.group[n];
        for (i = 0; i < VCU_MAX_CHANNELS/32; i++) {
            grp.channels[IN][type][i] |= chgrp->channels[IN][type][i];
            grp.channels[OUT][type][i] |= chgrp->channels[OUT][type][i];
        }
    }
    
    switch (type) {
    case VCU_CH_AMBIOPOLE:
    case VCU_CH_PASSTHRU:
        if (bit_isset(grp.channels[IN][type], idx) ||
            bit_isset(grp.channels[OUT][type], idx))
        {
            RETURN_CHANNEL_ACTIVE(true);
        }
        break;
    case VCU_CH_REVERB:
    case VCU_CH_AMBISONIC:
        if (io == OUT) {
            /* check inputs, if at least one input is active, all outputs
               will produce sound. This may not be the case for ambisonics
               though (for example only Z active in a planar layout), but we
               don't care */
            if (bit_find(grp.channels[IN][type],
                         0, cu.channel.count[IN][type]) != -1)
            {
                RETURN_CHANNEL_ACTIVE(true);
            }
            if (bit_isset(grp.channels[OUT][type], idx)) {
                RETURN_CHANNEL_ACTIVE(true);
            }
        } else {
            if (bit_find(grp.channels[OUT][type],
                         0, cu.channel.count[OUT][type]) != -1)
            {
                RETURN_CHANNEL_ACTIVE(true);
            }
            if (bit_isset(grp.channels[IN][type], idx)) {
                RETURN_CHANNEL_ACTIVE(true);
            }
        }
        break;
    }
    RETURN_CHANNEL_ACTIVE(false);
}

#undef RETURN_CHANNEL_ACTIVE

static bool_t
eqmode_valid_in_chmode(const struct vcu_eqmode *eqmode,
                       const struct vcu_chmode *chmode)
{
    int n, io, type;

    FOR_IN_AND_OUT {
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            for (n = 0; n < cu.channel.count[io][type]; n++) {
                if (eqmode->prgs[io][type][n] >= 0 &&
                    channel_is_active(chmode, -1, false, io, type, n))
                {
                    return true;
                }
            }
        }
    }
    return false;
}

static void
clear_invalid_eqmodes(struct vcu_chmode *chmode)
{
    int n;
    
    n = -1;
    while ((n = bit_find(chmode->eqmodes, n + 1, VCU_MAX_EQ_MODES - 1)) != -1) {
        if (!eqmode_valid_in_chmode(&cu.eq.mode[n], chmode)) {
            bit_clr(chmode->eqmodes, n);
        }
    }
    if (chmode->active_eq_mode >= 0 &&
        !bit_isset(chmode->eqmodes, chmode->active_eq_mode))
    {
        chmode->active_eq_mode = bit_find(chmode->eqmodes, 0, cu.eq.mode_count);
    }
}

static bool_t
sanitycheck_chmode(const struct vcu_chmode *mode)
{
    int n;
    
    if (mode == NULL) {
        elog(LOGDEBUG, "Null mode");
        return false;
    }
    if (mode->name[VCU_CHMODE_NAME_SIZE - 1] != '\0') {
        elog(LOGDEBUG, "Mode name not null terminated");
        return false;
    }
    if (mode->active_eq_mode >= cu.eq.mode_count ||
        (mode->active_eq_mode > 0 && !bit_isset(mode->eqmodes,
                                                mode->active_eq_mode)))
    {
        elog(LOGDEBUG, "Invalid eq settings");
        return false;
    }

    n = -1;
    while ((n = bit_find(mode->groups, n + 1, VCU_MAX_CHANNEL_GROUPS - 1))
           != -1)
    {
        if (n >= cu.channel.group_count) {
            elog(LOGDEBUG, "too large group index");
            return false;
        }
    }
    n = -1;
    while ((n = bit_find(mode->eqmodes, n + 1, VCU_MAX_EQ_MODES - 1)) != -1) {
        if (n >= cu.eq.mode_count) {
            elog(LOGDEBUG, "too large eq mode index");
            return false;
        }
        if (!eqmode_valid_in_chmode(&cu.eq.mode[n], mode)) {
            elog(LOGDEBUG, "eqmode not valid in channel mode");
            return false;
        }
    }
    return true;
}                  

static bool_t
sanitycheck_eqmode(const struct vcu_eqmode *mode)
{
    int io, n, type;
    
    if (mode == NULL || mode->name[VCU_EQMODE_NAME_SIZE - 1] != '\0') {
        return false;
    }        
    FOR_IN_AND_OUT {
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            for (n = 0; n < cu.channel.count[io][type]; n++) {
                if (mode->prgs[io][type][n] >= 0 &&
                    mode->prgs[io][type][n] >= cu.eq.eqprg_count)
                {
                    elog(LOGDEBUG, "eqprg id %d is invalid",
                         mode->prgs[io][type][n]);
                    return false;
                }
            }
        }
    }
    return true;
}

static bool_t
is_eq_active(int idx)
{
    int io, type, n;
    
    if (cu.eq.active_mode < 0) {
        return false;
    }
    if (cu.channel.active_mode >= 0 &&
        !bit_isset(cu.channel.mode[cu.channel.active_mode].eqmodes,
                   cu.eq.active_mode))
    {
        return false;
    }
    FOR_IN_AND_OUT {
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            for (n = 0; n < cu.channel.count[io][type]; n++) {
                if (cu.eq.mode[cu.eq.active_mode].prgs[io][type][n] == idx) {
                    return true;
                }
            }
        }
    }
    return false;
}

static int
sampleformat2bits(int format) {
    switch (format) {
    case BF_SAMPLE_FORMAT_S8:
	return 8;
    case BF_SAMPLE_FORMAT_S16_LE:
    case BF_SAMPLE_FORMAT_S16_BE:
    case BF_SAMPLE_FORMAT_S16_4LE:
    case BF_SAMPLE_FORMAT_S16_4BE:
	return 16;
    case BF_SAMPLE_FORMAT_S24_LE:
    case BF_SAMPLE_FORMAT_S24_BE:
    case BF_SAMPLE_FORMAT_S24_4LE:
    case BF_SAMPLE_FORMAT_S24_4BE:
	return 24;
    case BF_SAMPLE_FORMAT_S32_LE:
    case BF_SAMPLE_FORMAT_S32_BE:
    case BF_SAMPLE_FORMAT_FLOAT_LE:
    case BF_SAMPLE_FORMAT_FLOAT_BE:
	return 32;
    case BF_SAMPLE_FORMAT_FLOAT64_LE:
    case BF_SAMPLE_FORMAT_FLOAT64_BE:
    default:	    
	return 64;
    }
}

static void
load_program_progress_callback(int channel_index,
			       int n_channels,
			       int from_rate,
			       int to_rate,
                               bool_t deriving)
{
    if (load_fprg == NULL || progress.load_program == NULL) {
	return;
    }
    progress.load_program(load_fprg, channel_index, n_channels,
			  from_rate, to_rate, deriving);
}

static struct revprg *
get_revprg(prgid_t id)
{
    struct revprg *revprg;
    int n;

    n = -1;
    while ((revprg = (struct revprg *)idarray_next(rev_programs, &n)) != NULL) {
	if (memcmp(&revprg->prg.id, &id, sizeof(id)) == 0) {
	    return revprg;
	}
    }
    return NULL;
}

static struct xtcprg *
get_xtcprg(prgid_t id)
{
    struct xtcprg *xtcprg;
    int n;

    n = -1;
    while ((xtcprg = (struct xtcprg *)idarray_next(xtc_programs, &n)) != NULL) {
	if (memcmp(&xtcprg->prg.id, &id, sizeof(id)) == 0) {
	    return xtcprg;
	}
    }
    return NULL;
}

static void
mark_stage_regenerate(void)
{
    struct revprg *revprg;
    int n;
    
    if (rev_programs != NULL) {
	n = -1;
	while ((revprg = idarray_next(rev_programs, &n)) != NULL) {
            if (revprg->conf.response_mode ==
                VCU_REVERB_RESP_MODE_STAGE_ADAPTED)
            {
                revprg->regenerate = true;
            }
        }
    }
}

static bool_t
length_seconds2samples(int lengths[],
		       double seconds,
                       bool_t truncate_if_necessary)
{
    bool_t truncated;
    int n;

    elog(LOGDEBUG, "seconds: %.6f", seconds);
    truncated = false;
    for (n = 0; n < cu.signal.n_rates; n++) {
	lengths[n] = (int)((double)cu.signal.rates[n] * seconds);
	if (lengths[n] % cu.reverb.block_sizes[n] != 0) {
	    lengths[n] = (lengths[n] / cu.reverb.block_sizes[n] + 1) *
		cu.reverb.block_sizes[n];
	}
        if (truncate_if_necessary &&
            lengths[n] > cu.reverb.n_blocks[n] * cu.reverb.block_sizes[n])
        {
            lengths[n] = cu.reverb.n_blocks[n] * cu.reverb.block_sizes[n];
            truncated = true;
        }
    }
    return truncated;
}

static int
new_equaliser(struct equaliser **eq,
	      const struct vcu_eqpart *eqparts,
	      int n_components)
{
    int err;

    *eq = NULL;
    if ((err = eq_sanitycheck(eqparts, n_components, eq_programs)) != 0) {
	return err;
    }
    *eq = malloc(sizeof(struct equaliser));
    memset(*eq, 0, sizeof(struct equaliser));
    (*eq)->bfir_id = -1;
    
    return 0;
}

static void
channel_positions(double azimuts[],
                  double elevations[],
                  double distances[],
                  int type)
{
    int phys, logi;

    if (azimuts != NULL) {
        memset(azimuts, 0, cu.channel.count[OUT][type] * sizeof(double));
    }
    if (elevations != NULL) {
        memset(elevations, 0, cu.channel.count[OUT][type] * sizeof(double));
    }
    if (distances != NULL) {
        memset(distances, 0, cu.channel.count[OUT][type] * sizeof(double));
    }
    for (logi = 0; logi < cu.channel.count[OUT][type]; logi++) {
        phys = -1;
        while ((phys = bit_find(cu.channel.map[OUT][type][logi],
                                phys + 1,
                                cu.channel.count[OUT][VCU_CH_PHYSICAL] - 1))
               != -1)
        {
            if (!cu.channel.speakerpos[phys].active) {
                continue;
            }
            if (azimuts != NULL) {
                azimuts[logi] = cu.channel.speakerpos[phys].azimut;
            }
            if (elevations != NULL) {
                elevations[logi] = cu.channel.speakerpos[phys].elevation;
            }
            if (distances != NULL) {
                distances[logi] = cu.channel.speakerpos[phys].distance;
            }
            break;
        }
    }        
}

static int
load_revprg(const struct filterprg *program,
            const struct conf_revprg *conf,
	    struct bfirfilter *filters,
            double *orig_rt60,
            double *orig_tail_volume,
            double *proc_tail_volume,
            double *proc_length,
            double *proc_rt60,
            bool_t *truncated)
{
    double elevations[VCU_MAX_CHANNELS];
    double azimuts[VCU_MAX_CHANNELS];
    double widths[VCU_MAX_STAGES];
    double heights[VCU_MAX_STAGES];
    
    int err, id, lengths[FILTERPRG_RATES_MAX], n;
    double len;

    elog(LOGDEBUG, "Load reverb program: \"%s\" (%s)",
	 filterprg_prgid_ntoa(program->id), program->longname);

    filterprg_debug_print_program(program);
    
    load_fprg = program;
    for (n = 0; n < cu.channel.stage_count; n++) {
        azimuts[n] = cu.channel.stage[n].azimut;
        elevations[n] = cu.channel.stage[n].elevation;
        widths[n] = cu.channel.stage[n].width;
        heights[n] = cu.channel.stage[n].height;
    }
    err = reverb_settings(conf->bformat_directivity,
                          VCU_REVERB_EARLY_LENGTH,
                          VCU_REVERB_DIRECT_LENGTH,
                          conf->mix_with_dry ?
                          pow(10, conf->dry_volume / 20) : 0.0,
                          conf->response_mode,
                          conf->decorrelate_mode,
                          conf->force_delay_diff,
                          conf->force_decay,
                          cu.reverb.source_layout,
                          cu.channel.stage_count,
                          azimuts,
                          elevations,
                          widths,
                          heights);
    if (err != 0) {
        elog(LOGERROR, "reverb_settings failed: %s", vcu.ctl.strerror(err));
    }
    channel_positions(azimuts, elevations, NULL, VCU_CH_REVERB);

    len = reverb_postprocess_maxlength(program->length,
                                       program->predelay < 0 ? 0.0 :
                                       program->predelay);
    length_seconds2samples(lengths, len, true);
    if ((id = bfir_new_filters(BFIR_FILTER_TYPE_REVERB,
                               filters, lengths)) == -1)
    {
	return VCU_ERR_NO_MEMORY;
    }
    switch (program->type) {
    case FILTERPRG_TYPE_REVERB_BFORMAT:
        break;
    case FILTERPRG_TYPE_REVERB_TRADITIONAL:
        break;
    default:
        elog(LOGERROR, "unkown reverb program format: %d", program->type);
        break;
    }
    err = reverb_preprocess(program,
                            cu.signal.double_precision ? sizeof(double) :
                            sizeof(float),
                            conf->scale_length,
                            cu.channel.count[OUT][VCU_CH_REVERB],
                            azimuts,
                            elevations,
                            filters,
                            orig_rt60,
                            orig_tail_volume,
                            proc_rt60,
                            proc_tail_volume,
                            proc_length,
                            truncated,
                            load_program_progress_callback);
    load_fprg = NULL;
    length_seconds2samples(lengths, *proc_length, true);
    if (err != 0) {
	bfir_delete_filters(id);
	return err;
    }
    elog(LOGDEBUG, "before shorten filter");
    bfir_shorten_filters(id, lengths);
    elog(LOGDEBUG, "after shorten filter");
    return id;
}

static int
load_xtcprg(const struct filterprg *program,
	    struct bfirfilter filters[4],
	    int sample_delay[FILTERPRG_RATES_MAX])
{
    int n, err, id, lengths[FILTERPRG_RATES_MAX];
    struct prepfilter prepf;

    elog(LOGDEBUG, "Load xtc program: \"%s\"",
         filterprg_prgid_ntoa(program->id));
    load_fprg = program;
    if ((err = filterprg_load(program, &prepf, sample_delay,
                              NULL, 0, program->n_channels,
			      cu.signal.rates,
			      cu.signal.n_rates,
                              cu.signal.double_precision ? sizeof(double) :
                              sizeof(float),
			      load_program_progress_callback, -1, -1)) != 0)
    {
	return err;
    }

    load_fprg = NULL;
    length_seconds2samples(lengths, program->length, false);
    for (n = 0; n < cu.signal.n_rates; n++) {
        if (lengths[n] > VCU_MAX_XTC_SIZE) {
            elog(LOGDEBUG, "truncation necessary");
            return VCU_ERR_TRUNCATION_NECESSARY;
        }
    }
    if ((id = bfir_new_filters(BFIR_FILTER_TYPE_XTC,
                               filters, lengths)) == -1)
    {
	filterprg_free_prepfilter_contents(&prepf);
	return VCU_ERR_NO_MEMORY;
    }
    if (prepf.n_channels == 4) {
        err = bfirfilter_from_prepf(&prepf, filters);
    } else {
        err = bfirfilter_from_prepf(&prepf, filters);
        if (err == 0) {
            err = bfirfilter_from_prepf(&prepf, &filters[2]);
        }
    }
    filterprg_free_prepfilter_contents(&prepf);
    if (err != 0) {
	bfir_delete_filters(id);
	return err;
    }
    return id;
}

static bool_t
find_filter_program(prgid_t id,
                    const struct filterprg *fprg_list,
                    int n_programs,
                    struct filterprg *fprg)
{
    int n;
    
    memset(fprg, 0, sizeof(struct filterprg));

    for (n = 0; n < n_programs; n++) {
        if (memcmp(&fprg_list[n].id, &id, sizeof(prgid_t)) == 0) {
            *fprg = fprg_list[n];
            return true;
        }
    }
    for (n = 0; n < cu.ctl.n_fprogrefs; n++) {
        if (memcmp(&cu.ctl.fprogrefs[n]->id, &id, sizeof(prgid_t)) == 0) {
            return filterprg_scan_file(cu.ctl.fprogrefs[n]->path, fprg);
        }
    }
    return false;
}

static void
init_load_custom_eq_programs(void)
{
    struct filterprg *filterprg, *fprgs;
    int n, n_programs;
    
    eq_programs = idarray_new(false);    
    if ((fprgs = filterprg_scan_directory(conf_static.storage_path,
					  &n_programs)) == NULL)
    {
	return;
    }
    
    for (n = 0; n < cu.eq.n_loaded_programs; n++) {
        filterprg = malloc(sizeof(struct filterprg));
        if (!find_filter_program(cu.eq.loaded_programs[n], fprgs, n_programs,
                                 filterprg))
        {
            free(filterprg);
        } else {
            idarray_new_id(eq_programs, filterprg);
        }
    }        
}

static void
init_load_stored_filter_programs(void)
{
    uint32_t mask[VCU_MAX_CHANNELS/32];
    struct filterprg *fprgs, fprg;
    struct xtcprg *xtcprg;
    struct revprg *revprg;
    int n, i, n_programs;
	
    xtc_programs = idarray_new(false);
    rev_programs = idarray_new(false);
    if ((fprgs = filterprg_scan_directory(conf_static.storage_path,
					  &n_programs)) == NULL)
    {
	return;
    }

    if (cu.channel.count[OUT][VCU_CH_AMBIOPOLE] > 0) {
        if (cu.ambiopole.n_loaded_programs > 0) {
            progress.init_phase(VCU_INIT_PHASE_LOAD_XTC_PROGRAM);
        }
        for (n = 0; n < cu.ambiopole.n_loaded_programs; n++) {
            if (find_filter_program(cu.ambiopole.loaded_programs[n], fprgs,
                                    n_programs, &fprg))
            {
                xtcprg = malloc(sizeof(struct xtcprg));
                memset(xtcprg, 0, sizeof(struct xtcprg));
                if ((xtcprg->bfir_id = load_xtcprg(&fprg,
                                                   xtcprg->bfirfilters,
                                                   xtcprg->sample_delay)) < 0)
                {
                    free(xtcprg);
                } else {
                    xtcprg->prg = fprg;
                    idarray_new_id(xtc_programs, xtcprg);
                }
            }
        }
    }
    
    if (cu.channel.count[OUT][VCU_CH_REVERB] > 0) {
        if (cu.reverb.n_loaded_programs > 0) {
            progress.init_phase(VCU_INIT_PHASE_LOAD_REVERB_PROGRAM);
        }
        for (n = 0; n < cu.reverb.n_loaded_programs; n++) {
            if (find_filter_program(cu.reverb.loaded_programs[n], fprgs,
                                    n_programs, &fprg))
            {
                revprg = malloc(sizeof(struct revprg));
                memset(revprg, 0, sizeof(struct revprg));
                revprg->bfirfilters =
                    malloc(cu.channel.count[IN][VCU_CH_REVERB] *
                           cu.channel.count[OUT][VCU_CH_REVERB] *
                           sizeof(struct bfirfilter));                    
                conf_load_revprg(&revprg->conf, &fprg);
                if ((revprg->bfir_id =
                     load_revprg(&fprg,
                                 &revprg->conf,
                                 revprg->bfirfilters,
                                 &revprg->orig_rt60,
                                 &revprg->orig_tail_volume,
                                 &revprg->proc_tail_volume,
                                 &revprg->proc_length,
                                 &revprg->proc_rt60,
                                 &revprg->truncated)) < 0)
                {
                    free(revprg->bfirfilters);
                    free(revprg);
                } else {
                    revprg->prg = fprg;
                    idarray_new_id(rev_programs, revprg);
                }
            }
        }
    }
    
    /* verify conf settings, and correct them if mismatching */
    if (get_revprg(cu.reverb.current) == NULL) {
	n = -1;
	if ((revprg = idarray_next(rev_programs, &n)) == NULL) {
	    cu.reverb.current = zeroid;
	} else {
	    cu.reverb.current = revprg->prg.id;
	}	
	save_userconf = true;
    }
    memset(mask, 0, SIZEOF_CH_MASK);
    for (n = 0; n < cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2; n++) {
        if ((xtcprg = get_xtcprg(cu.ambiopole.current[n])) == NULL) {
            i = -1;
            if ((xtcprg = idarray_next(xtc_programs, &i)) == NULL) {
                cu.ambiopole.current[n] = zeroid;
            } else {
                cu.ambiopole.current[n] = xtcprg->prg.id;
            }	
            save_userconf = true;
        }
        if (xtcprg == NULL && cu.ambiopole.use_xtalk_cancel[n]) {
            cu.ambiopole.use_xtalk_cancel[n] = false;
            save_userconf = true;
        }
        if (xtcprg != NULL) {
            bit_set(mask, n);
            bfir_set_xtalkcancel(mask, xtcprg->bfir_id);
            bit_clr(mask, n);
        }
    }
}

static int
generate_equalisers(int *fail_index)
{
    struct prepfilter prepf;
    struct equaliser *eq;
    bool_t too_large;
    int idx, err, n;

    for (idx = 0; idx < cu.eq.eqprg_count; idx++) {
        eq = equalisers[idx];
	if (eq->bfir_id != -1) {
            elog(LOGDEBUG, "equaliser already generated");
	    continue;
	}
	progress.generate_eq(idx, cu.eq.eqprg_count, 0);
	memset(eq->sample_delay, 0, sizeof(eq->sample_delay));
	if ((err = eq_generate(cu.eq.eqprg[idx].parts,
			       cu.eq.eqprg[idx].part_count,
			       eq_programs,
                               cu.reverb.block_sizes,
			       cu.signal.rates,
			       cu.signal.n_rates,
                               cu.signal.double_precision ? sizeof(double) :
                               sizeof(float),
			       eq->sample_delay,
			       &prepf)) != 0)
	{
	    elog(LOGDEBUG, "eq_generate() failed: %s", vcu.ctl.strerror(err));
	    *fail_index = idx;
	    progress.generate_eq(idx, cu.eq.eqprg_count, err);
	    return err;
	}
        elog(LOGDEBUG, "sample delay for equaliser %d is %d at rate index 0",
             idx, eq->sample_delay[0]);
        too_large = false;
        for (n = 0; n < cu.signal.n_rates; n++) {
            if (prepf.lengths[n] > VCU_MAX_EQ_SIZE) {
                too_large = true;
                break;
            }
        }
        if (too_large) {
	    filterprg_free_prepfilter_contents(&prepf);
	    *fail_index = idx;
	    progress.generate_eq(idx, cu.eq.eqprg_count,
                                 VCU_ERR_TRUNCATION_NECESSARY);
            elog(LOGDEBUG, "truncation necessary");
	    return VCU_ERR_TRUNCATION_NECESSARY;
	}
	if ((eq->bfir_id = bfir_new_filters(BFIR_FILTER_TYPE_EQ,
                                            &eq->bfirfilter,
                                            prepf.lengths)) == -1)
	{
	    filterprg_free_prepfilter_contents(&prepf);
	    *fail_index = idx;
	    progress.generate_eq(idx, cu.eq.eqprg_count, VCU_ERR_NO_MEMORY);
            elog(LOGDEBUG, "no memory");
	    return VCU_ERR_NO_MEMORY;
	}
	if ((err = bfirfilter_from_prepf(&prepf, &eq->bfirfilter)) != 0) {
	    elog(LOGDEBUG, "bfirfilter_from_prepf() failed: %s",
                 vcu.ctl.strerror(err));
	    filterprg_free_prepfilter_contents(&prepf);
	    bfir_delete_filters(eq->bfir_id);
	    eq->bfir_id = -1;
	    *fail_index = idx;
	    progress.generate_eq(idx, cu.eq.eqprg_count, err);
	    return err;
	}
        memset(eq->lengths, 0, sizeof(eq->lengths));
        for (n = 0; n < cu.signal.n_rates; n++) {
            eq->lengths[n] = prepf.lengths[n];
        }
	filterprg_free_prepfilter_contents(&prepf);
    }
    if (cu.eq.eqprg_count > 0) {
	progress.generate_eq(-1, cu.eq.eqprg_count, 0);
    }
    return 0;
}

static void
generate_ambisonic_filters(void)
{
    int n, lengths[VCU_VALID_RATES_COUNT];
    struct bfirfilter bfirfilter[3];

    if (cu.ambisonics.input_format != VCU_AMBISONIC_INPUT_FORMAT_UHJ ||
        uhj_bfir_id != -1)
    {
        return;
    }
    for (n = 0; n < cu.signal.n_rates; n++) {
        lengths[n] =
            ambisonics_uhj2bformat_filter_length(cu.signal.rates[n], NULL);
    }
    uhj_bfir_id = bfir_new_filters(BFIR_FILTER_TYPE_UHJ,
                                   bfirfilter, lengths);
    for (n = 0; n < 3; n++) {
        progress.generate_ambisonic(n, 3);
        ambisonics_generate_uhj2bformat_filter(n, &bfirfilter[n]);
    }
}

static double *
calculate_ambisonic_gains(void)
{
    static double gains[9 * BFIR_MAX_OUTPUT_CHANNELS];
    double elevations[BFIR_MAX_OUTPUT_CHANNELS];
    double azimuts[BFIR_MAX_OUTPUT_CHANNELS];
    int err;
    
    memset(gains, 0, sizeof(gains));
    if (cu.channel.count[OUT][VCU_CH_AMBISONIC] == 0) {
        return gains;
    }
    channel_positions(azimuts, elevations, NULL, VCU_CH_AMBISONIC);
    err = ambisonics_calculate_gains(
        cu.channel.count[IN][VCU_CH_AMBISONIC],
        cu.channel.count[OUT][VCU_CH_AMBISONIC],
        azimuts, elevations, gains);
    if (err != 0) {
        elog(LOGERROR, "ambisonic gains failed: %s", vcu.ctl.strerror(err));
        exit(EXIT_FAILURE);
    }
    return gains;
}

static int
calc_eq_length(int block_size,
               int rate,
               int eqprgidx,
               bool_t generate)
{
    int err, sample_delay, ri;
    struct prepfilter prepf;
    struct equaliser *eq;

    elog(LOGDEBUG, "block_size: %d; rate: %d; eqprgidx: %d; generate: %d",
         block_size, rate, eqprgidx, generate);
    for (ri = 0; ri < cu.signal.n_rates; ri++) {
	if (cu.signal.rates[ri] == runtime_sample_rate) {
	    break;
	}
    }
    if (eqprgidx < 0 || eqprgidx >= cu.eq.eqprg_count) {
        return 0;
    }
    eq = equalisers[eqprgidx];
    if (!generate &&
        ri != cu.signal.n_rates &&
        eq->lengths[ri] != 0)
    {
        elog(LOGDEBUG, "cached length %d", eq->lengths[ri]);
        return eq->lengths[ri];
    }
    if ((err = eq_generate(cu.eq.eqprg[eqprgidx].parts,
                           cu.eq.eqprg[eqprgidx].part_count,
                           eq_programs,
                           &block_size,
                           &rate,
                           1,
                           cu.signal.double_precision ? sizeof(double) :
                           sizeof(float),
                           &sample_delay,
                           &prepf)) != 0)
    {
        elog(LOGDEBUG, "eq generate failed: %s", vcu.ctl.strerror(err));
        return 0;
    }
    filterprg_free_prepfilter_contents(&prepf);
    elog(LOGDEBUG, "calculated length %d", prepf.lengths[0]);
    return prepf.lengths[0];
}

static void
get_filter_lengths(int block_size,
                   int rate,
                   int xtc_length[],
                   int *eq_length[2][VCU_CH_WP_COUNT],
                   bool_t generate)
{
    int n, i, io, id, type, length, eqlen[VCU_MAX_EQ_PROGRAMS];
    struct xtcprg *xtcprg;
    double rlength;

    memset(eqlen, 0, sizeof(eqlen));
    memset(xtc_length, 0, sizeof(int) * VCU_MAX_CHANNELS / 2);

    FOR_IN_AND_OUT {
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            memset(eq_length[io][type], 0, sizeof(int) * VCU_MAX_CHANNELS);
            for (n = 0; n < cu.channel.count[io][type]; n++) {
                length = 0;
                for (i = 0; i < cu.eq.mode_count; i++) {
                    id = cu.eq.mode[i].prgs[io][type][n];
                    if (id < 0) {
                        continue;
                    }
                    if (eqlen[id] == 0) {
                        eqlen[id] = calc_eq_length(block_size, rate, id,
                                                   generate);
                    }
                    if (eqlen[id] > length) {
                        length = eqlen[id];
                    }
                }
                eq_length[io][type][n] = length;
                elog(LOGDEBUG, "eq_length[%d][%d][%d]: %d", io, type, n,
                     eq_length[io][type][n]);
            }
        }
    }
    
    if (cu.channel.count[OUT][VCU_CH_AMBIOPOLE] > 0) {
	for (n = 0; n < cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2; n++) {
            if (cu.ambiopole.use_xtalk_cancel[n]) {
                /* find the longest length of the currently loaded filters */
                id = -1;
                rlength = 0.0;
                while ((xtcprg = idarray_next(xtc_programs, &id)) != NULL) {
                    if (xtcprg->prg.length > rlength) {
                        rlength = xtcprg->prg.length;
                    }
                }
                xtc_length[n] = (int)(rlength * 1.1 * (double)rate);		
                elog(LOGDEBUG, "xtc_length[%d]: %d", n, xtc_length[n]);
            }
        }
    }
}

static void
benchmark_filter_lengths(int block_size,
                         int rate,
                         int xtc_length[],
                         int *eq_length[2][VCU_CH_WP_COUNT])
{
    get_filter_lengths(block_size, rate, xtc_length, eq_length, true);
}

static void
benchmark_progress_callback(int n_blocks,
			    int block_size,
			    double load)
{
    double delay, length;

    if (progress.benchmark == NULL) {
	return;
    }
    if (load != 0.0) {
	load /= cu.reverb.max_benchmark_load;
    }
    length = (double)n_blocks * block_size  / (double)benchmark.current_rate;
    delay = (double)(2 * block_size) / (double)benchmark.current_rate;
    progress.benchmark(benchmark.current_rate, length, delay, load);
}

static void
update_bfir_soft_settings(void)
{
    uint32_t apply_eq[2][VCU_CH_WP_COUNT][VCU_MAX_CHANNELS/32];
    uint32_t phys_used[VCU_MAX_CHANNELS/32];
    struct bfir_chmap *maps[2];
    int n, i, io, type, phys;

    make_bfir_chmap(maps);
    memset(apply_eq, 0, sizeof(apply_eq));
    FOR_IN_AND_OUT {
        memset(phys_used, 0, sizeof(phys_used));
        for (type = 0; type < VCU_CH_COUNT; type++) {
            for (n = 0; n < cu.channel.count[io][type]; n++) {
                phys = -1;
                while ((phys =
                        bit_find(cu.channel.map[io][type][n],
                                 phys + 1,
                                 cu.channel.count[io][VCU_CH_PHYSICAL]- 1))
                       != -1)
                {
                    bit_set(phys_used, phys);
                }
                for (i = 0; i < cu.eq.mode_count; i++) {
                    if (cu.eq.mode[i].prgs[io][type][n] >= 0) {
                        bit_set(apply_eq[io][type], n);
                    }
                }
            }
        }
        type = VCU_CH_PHYSICAL;
        for (n = 0; n < cu.channel.count[io][type]; n++) {
            for (i = 0; i < cu.eq.mode_count; i++) {
                if (!bit_isset(phys_used, n)) {
                    elog(LOGDEBUG, "skipping unused phys io %d idx %d", io, n);
                    continue;
                }
                if (cu.eq.mode[i].prgs[io][type][n] >= 0) {
                    bit_set(apply_eq[io][type], n);
                }
            }
        }
    }
    /* find the largest resolution input format */
    for (n = i = 0; n < conf_static.input.n_sample_formats; n++) {
	if (sampleformat2bits(conf_static.input.sample_formats[n]) >
	    sampleformat2bits(conf_static.input.sample_formats[i]))
	{
	    i = n;
	}
    }
    if (!bfir_soft_settings(conf_static.input.sample_formats[i],
			    cu.signal.output_sample_format,
                            cu.signal.apply_dither,
			    apply_eq,
			    cu.ambiopole.use_xtalk_cancel,
                            calculate_ambisonic_gains(),
			    (const struct bfir_chmap **)maps))
    {
	elog(LOGERROR, "Could not update convolution engine soft settings.");
	exit(EXIT_FAILURE);
    }
}

static int
update_hard_settings(bool_t check_only)
{
    static int rates[VCU_VALID_RATES_COUNT], n_rates = -1;
    static int block_sizes[VCU_VALID_RATES_COUNT], n_reverb_channels = -1;
    static int n_blocks[VCU_VALID_RATES_COUNT];
    static struct speakerpos speakerpos[VCU_MAX_CHANNELS];
    static int ambisonic_input_format = -1;
    static int reverb_source_layout = -1;
    static bool_t double_precision = false;
    static bool_t isinit = true;
    struct revprg *revprg;
    struct xtcprg *xtcprg;
    struct equaliser *eq;
    bool_t reload_revprg = false;
    bool_t reload_xtcprg = false;
    bool_t regenerate_eq = false;
    bool_t regenerate_revprg = false;
    bool_t regenerate_ambisonic = false;
    bool_t update_bfir_settings = false;
    int n, retval;
    
    /* first we check if we actually need to do anything */
    elog(LOGDEBUG, "updating hard settings enter");
    if (isinit) {
	isinit = false;
	update_bfir_settings = true;
	elog(LOGDEBUG, "Init");
    }
    if (has_newly_run_benchmark) {
        elog(LOGDEBUG, "has newly run benchmark, update everything");
        reload_revprg = true;
        reload_xtcprg = true;
        regenerate_eq = true;
        regenerate_ambisonic = true;
	update_bfir_settings = true;
    }
    if (cu.signal.double_precision != double_precision) {
        elog(LOGDEBUG, "has changed precision, update everything");
        reload_revprg = true;
        reload_xtcprg = true;
        regenerate_eq = true;
        regenerate_ambisonic = true;
	update_bfir_settings = true;
    }
    for (n = 0; n < n_rates; n++) {
	if (cu.reverb.block_sizes[n] != block_sizes[n] ||
	    cu.signal.rates[n] != rates[n] ||
            cu.signal.n_rates != n_rates)
	{
	    reload_revprg = true;
	    reload_xtcprg = true;
	    regenerate_eq = true;
            regenerate_ambisonic = true;
	    update_bfir_settings = true;
	    elog(LOGDEBUG, "Block sizes or rates changed");
	    break;
	}
        if (cu.reverb.n_blocks[n] != n_blocks[n]) {
	    reload_revprg = true;
        }

    }
    if (rev_programs != NULL) {
	n = -1;
	while ((revprg = idarray_next(rev_programs, &n)) != NULL) {
            if (revprg->regenerate) {
                regenerate_revprg = true;
                break;
            }
        }
    }
    if (cu.channel.count[OUT][VCU_CH_REVERB] != n_reverb_channels) {
	reload_revprg = true;
	update_bfir_settings = true;
	elog(LOGDEBUG, "Reverb channel amount changed");
    } else {
	for (n = 0; n < n_reverb_channels; n++) {
	    if (cu.channel.speakerpos[n].active !=
		speakerpos[n].active ||
                cu.channel.speakerpos[n].azimut !=
		speakerpos[n].azimut ||
		cu.channel.speakerpos[n].elevation !=
		speakerpos[n].elevation ||
		cu.channel.speakerpos[n].distance !=
		speakerpos[n].distance)
	    {
		elog(LOGDEBUG, "Speaker position changed");
                reload_revprg = true;
		break;
	    }
	}
    }
    if (cu.reverb.source_layout != reverb_source_layout) {
        elog(LOGDEBUG, "Reverb source layout changed");
        reload_revprg = true;
    }
    if (cu.ambisonics.input_format != ambisonic_input_format) {
        elog(LOGDEBUG, "Ambisonics input format changed");
        regenerate_ambisonic = true;
    }
    for (n = 0; n < cu.eq.eqprg_count; n++) {
        eq = equalisers[n];
        if (eq != NULL && eq->bfir_id == -1) {
            regenerate_eq = true;
            elog(LOGDEBUG, "Not all equalisers are generated");
            break;
	}
    }
    if (!reload_revprg && !regenerate_revprg && !reload_xtcprg &&
        !regenerate_eq && !regenerate_ambisonic && !update_bfir_settings)
    {
        elog(LOGDEBUG, "No changes since last time");
	return 0;
    }
    if (check_only) {
	elog(LOGDEBUG, "Check only: need to update hard settings");
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    double_precision = cu.signal.double_precision;
    n_rates = cu.signal.n_rates;
    n_reverb_channels = cu.channel.count[OUT][VCU_CH_REVERB];
    reverb_source_layout = cu.reverb.source_layout;
    ambisonic_input_format = cu.ambisonics.input_format;    
    for (n = 0; n < n_rates; n++) {
	rates[n] = cu.signal.rates[n];
	block_sizes[n] = cu.reverb.block_sizes[n];
	n_blocks[n] = cu.reverb.n_blocks[n];
    }
    for (n = 0; n < n_reverb_channels; n++) {
	speakerpos[n] = cu.channel.speakerpos[n];
    }

    if (update_bfir_settings) {
	elog(LOGDEBUG, "calling bfir_hard_settings()");
	if (!bfir_hard_settings(cu.signal.n_rates,
				cu.signal.rates,
				cu.reverb.block_sizes,
				cu.channel.count[IN][VCU_CH_REVERB],
				cu.channel.count[OUT][VCU_CH_REVERB],
                                cu.ambisonics.input_format,
                                cu.signal.double_precision ? sizeof(double) :
                                sizeof(float)))
	{
	    elog(LOGDEBUG, "Could not update convolution "
                 "engine hard settings.");
	    exit(EXIT_FAILURE);
	}
	/* bfir_hard_settings() drops all filters, so we need to reload all */
	reload_revprg = true;
	reload_xtcprg = true;
	regenerate_eq = true;
        regenerate_ambisonic = true;
        has_newly_run_benchmark = false;
    }

    if (rev_programs == NULL) {
        elog(LOGDEBUG, "init return early");
	/* init */
	return 0;
    }

    retval = 0;
    
    /* reload xtc filters */
    if (cu.ambiopole.n_loaded_programs > 0 && reload_xtcprg) {
        elog(LOGDEBUG, "reload xtc filters");
	progress.init_phase(VCU_INIT_PHASE_LOAD_XTC_PROGRAM);
	n = -1;
	while ((xtcprg = idarray_next(xtc_programs, &n)) != NULL) {
            if (!update_bfir_settings) {
                bfir_delete_filters(xtcprg->bfir_id);
            }
	    if ((xtcprg->bfir_id = load_xtcprg(&xtcprg->prg,
					       xtcprg->bfirfilters,
					       xtcprg->sample_delay)) < 0)
	    {
		elog(LOGDEBUG, "Could not reload xtc filter: %s",
		     vcu.ctl.strerror(xtcprg->bfir_id));
                retval = VCU_ERR_NO_MEMORY; /* assume no memory */
		idarray_delete_id(xtc_programs, n);
		free(xtcprg);
	    }
	}
    }

    /* regenerate equalisers */
    if (cu.eq.eqprg_count > 0 && regenerate_eq) {
        elog(LOGDEBUG, "regenerate equalisers");
	progress.init_phase(VCU_INIT_PHASE_GENERATING_EQUALISERS);
        for (n = 0; n < cu.eq.eqprg_count; n++) {
            eq = equalisers[n];
            if (!update_bfir_settings) {
                bfir_delete_filters(eq->bfir_id);
            }
            eq->bfir_id = -1;
        }
	while (generate_equalisers(&n) != 0) {
            retval = VCU_ERR_NO_MEMORY; /* assume no memory */
	    vcu.eq.delete_equaliser(n);
	}
    }

    /* regenerate ambisonic filters */
    if (regenerate_ambisonic) {
        elog(LOGDEBUG, "regenerate ambisonic filters");
	progress.init_phase(VCU_INIT_PHASE_GENERATING_AMBISONIC);        
        if (!update_bfir_settings) {
            bfir_delete_filters(uhj_bfir_id);
            uhj_bfir_id = -1;
        }
        generate_ambisonic_filters();
    }
    
    /* reload reverb programs */
    if (cu.reverb.n_loaded_programs > 0 &&
        (reload_revprg || regenerate_revprg))
    {
        elog(LOGDEBUG, "reload reverb programs");
        if (cu.channel.count[OUT][VCU_CH_REVERB] > 0) {
            progress.init_phase(VCU_INIT_PHASE_LOAD_REVERB_PROGRAM);
        }
	n = -1;
	while ((revprg = idarray_next(rev_programs, &n)) != NULL) {
            if (!reload_revprg && !revprg->regenerate) {
                continue;
            }
            if (!update_bfir_settings) {
                bfir_delete_filters(revprg->bfir_id);
            }
            if (cu.channel.count[OUT][VCU_CH_REVERB] == 0) {
		idarray_delete_id(rev_programs, n);
		free(revprg->bfirfilters);
		free(revprg);
                continue;
            }
            revprg->regenerate = false;
	    revprg->bfirfilters =
		realloc(revprg->bfirfilters,
			cu.channel.count[IN][VCU_CH_REVERB] *
                        cu.channel.count[OUT][VCU_CH_REVERB] *
			sizeof(struct bfirfilter));
	    if ((revprg->bfir_id = load_revprg(&revprg->prg,
                                               &revprg->conf,
					       revprg->bfirfilters,
                                               &revprg->orig_rt60,
                                               &revprg->orig_tail_volume,
                                               &revprg->proc_tail_volume,
                                               &revprg->proc_length,
                                               &revprg->proc_rt60,
                                               &revprg->truncated)) < 0)
	    {
		elog(LOGDEBUG, "Could not reload rev filter: %s",
		     vcu.ctl.strerror(revprg->bfir_id));
                retval = VCU_ERR_NO_MEMORY; /* assume no memory */
		idarray_delete_id(rev_programs, n);
		free(revprg->bfirfilters);
		free(revprg);
	    }
	}
    }
    
    elog(LOGDEBUG, "updating hard settings leave (retval %d)", retval);
    return retval;
}

static int
int_compare(const void *int1,
	    const void *int2)
{
    return (*(const int *)int1) - (*(const int *)int2);
}

static int
config_compare(const void *config1,
               const void *config2)
{
    return strcmp(((const struct vcu_config *)config1)->name,
                  ((const struct vcu_config *)config2)->name);
}

static int
fprg_compare(const void *fprg1,
             const void *fprg2)
{
    return strcmp(((const struct filterprg *)fprg1)->longname,
                  ((const struct filterprg *)fprg2)->longname);
}

static void
process_terminated_callback(int reason)
{
    elog(LOGDEBUG, "BruteFIR terminated");
    isruntime = false;
    FD_CLR(bf_fd, &readfds);
    bf_fd = -1;
    periodic();
    if (process_terminated_cb != NULL) {
	process_terminated_cb(reason);
    }
}

static int
get_active_eq_mode(void)
{
    if (cu.channel.active_mode == -1) {
        return cu.eq.active_mode;
    } else {
        return cu.channel.mode[cu.channel.active_mode].active_eq_mode;
    }
}

static double
get_max_distance(void)
{
    double max;
    int n;

    max = 0.0;
    for (n = 0; n < conf_static.output.total_open_channels; n++) {
        if (!cu.channel.speakerpos[n].active) {
            continue;
        }
        if (cu.channel.speakerpos[n].distance > max) {
            max = cu.channel.speakerpos[n].distance;
        }
    }
    return max;
}

static double
distance_gain(int output)
{
#warning implement and use
    return 1.0;
}

static int
distance_delay(int ri,
               int output)
{
    if (!cu.channel.distance_compensate ||
        !cu.channel.speakerpos[output].active)
    {
        return 0;
    }
    return cu.signal.rates[ri] * SOUNDSPEED *
        (get_max_distance() - cu.channel.speakerpos[output].distance);
}

static void
apply_volume(int io,
             int type,
             int idx,
             bool_t reset)
{
    static double oldvol[VCU_CH_WP_COUNT][2][BFIR_MAX_OUTPUT_CHANNELS];
    static bool_t isinit = false;
    
    uint32_t setmask[BFIR_MAX_OUTPUT_CHANNELS / 32 + 1];
    uint32_t mask[BFIR_MAX_OUTPUT_CHANNELS / 32 + 1];
    double vol[2][BFIR_MAX_OUTPUT_CHANNELS];
    double elevations[BFIR_MAX_OUTPUT_CHANNELS];
    double azimuts[BFIR_MAX_OUTPUT_CHANNELS];
    double xyzvol[BFIR_MAX_OUTPUT_CHANNELS];
    struct revprg *revprg;
    int n, i, start, end;
    double volume;

    if (!isinit) {
        isinit = true;
        memset(oldvol, 0, sizeof(oldvol));
    }
    
    /*
      volumes handled:
      
      - master volume
      - channel group volume (in/out)
      - trim volume (in/out)
      - distance compensation volume
      - reverb trim volume
      - reverb xyz balance volume

     */
    if (idx < 0) {
        start = 0;
        end = cu.channel.count[io][type]  - 1;
    } else {
        start = idx;
        end = idx;
    }
    memset(vol, 0, sizeof(vol));
    for (n = start; n <= end; n++) {

        /* trim volume */
        vol[io][n] = cu.channel.trim.volume[io][type][n];
        
        /* master volume (only applied to physical outputs) */
        if (io == OUT && type == VCU_CH_PHYSICAL) {
            vol[io][n] += cu.volume.master;
        }
        
        if (type != VCU_CH_PHYSICAL) {
            /* channel group volume */
            for (i = 0; i < cu.channel.group_count; i++) {
                if (cu.channel.active_mode >= 0 &&
                    !bit_isset(cu.channel.mode[cu.channel.active_mode].groups,
                               i))
                {
                    continue;
                }
                if (bit_isset(cu.channel.group[i].channels[io][type], n)) {
                    vol[io][n] += cu.volume.group[i];
                }
            }
        }
    }
    if (io == OUT && type == VCU_CH_REVERB &&
        (revprg = get_revprg(cu.reverb.current)) != NULL &&
        cu.channel.count[io][type] > 0)
    {
        /* reverb xyz balance + reverb trim volume */
        channel_positions(azimuts, elevations, NULL, VCU_CH_REVERB);
        reverb_apply_balance(revprg->conf.xyz_balance,
                             cu.channel.count[io][type],
                             azimuts,
                             elevations,
                             xyzvol);
        for (n = start; n <= end; n++) {
            vol[io][n] += revprg->conf.volume + xyzvol[n];
        }
    }
    
    /* minimise number of bfir_set_volume calls */
    memset(setmask, 0xFF, sizeof(setmask));
    for (n = start; n <= end; n++) {
        if (!reset && vol[io][n] == oldvol[type][io][n]) {
            bit_clr(setmask, n);
            continue;
        }
        if (!reset && isruntime && cu.channel.active_mode >= 0 &&
            !bit_isset(runtime_active_map.ch[io][type][n],
                       cu.channel.active_mode))
        {
            bit_clr(setmask, n);
            continue;
        }
        oldvol[type][io][n] = vol[io][n];
    }
    bfir_start_buffering_commands();
    while ((n = bit_find(setmask, start, end)) != -1) {
        volume = vol[io][n];
        memset(mask, 0, sizeof(mask));
        bit_clr(setmask, n);
        bit_set(mask, n);
        i = n;
        while ((i = bit_find(setmask, i + 1,
                             cu.channel.count[io][type] - 1)) != -1)
        {
            if (vol[io][i] == volume) {
                bit_clr(setmask, i);
                bit_set(mask, i);
            }
        }
        bfir_set_volume(io, type, mask, start, end, volume);
    }
    bfir_commit_buffered_commands();
}

static void
sample_align(int io,
             int type,
             int idx)
{
    uint32_t mask[BFIR_MAX_OUTPUT_CHANNELS / 32];
    int n, i, delay, ri, start, stop, eq_mode;
    struct revprg *revprg;
    struct xtcprg *xtcprg;
    struct equaliser *eq;

    for (ri = 0; ri < cu.signal.n_rates; ri++) {
	if (cu.signal.rates[ri] == runtime_sample_rate) {
	    break;
	}
    }
    if (cu.channel.count[OUT][type] == 0) {
        return;
    }

    if (idx < 0) {
        start = 0;
        stop = cu.channel.count[io][type] - 1;
    } else if (idx < cu.channel.count[io][type]) {
        start = stop = idx;
    } else {
        return;
    }
        
    bfir_start_buffering_commands();

    eq_mode = get_active_eq_mode();

    elog(LOGDEBUG, "sample align eq mode %d, io %d, type %d, index %d",
         eq_mode, io, type, idx);
    
    memset(mask, 0, sizeof(mask));
        
    if (io == OUT) {

        revprg = get_revprg(cu.reverb.current);
        
        for (n = start; n <= stop; n++) {
            
            delay = runtime_delay_map.eqalign[eq_mode+1].delay[io][type][n];
            if (eq_mode >= 0 &&
                cu.eq.mode[eq_mode].prgs[io][type][n] >= 0 &&
                cu.eq.mode[eq_mode].prgs[io][type][n] < cu.eq.eqprg_count)
            {
                eq = equalisers[cu.eq.mode[eq_mode].prgs[io][type][n]];
                delay -= eq->sample_delay[ri];
            }
            switch (type) {
            case VCU_CH_AMBIOPOLE:
                if (cu.ambiopole.use_xtalk_cancel[n/2]) {
                    xtcprg = get_xtcprg(cu.ambiopole.current[n/2]);
                    if (xtcprg != NULL) {
                        delay -= xtcprg->sample_delay[ri];
                    }
                }
                break;
            case VCU_CH_REVERB:
                if (revprg != NULL) {
                    delay += (int)(revprg->conf.delay *
                                   (double)runtime_sample_rate);
                }
                break;
            case VCU_CH_AMBISONIC:            
            case VCU_CH_PASSTHRU:
            case VCU_CH_PHYSICAL:
                break;
            }
            
            elog(LOGDEBUG, "alignment delay: %d", delay);
            i = (int)((cu.channel.trim.delay[OUT][type][n]) *
                      (double)runtime_sample_rate);
            elog(LOGDEBUG, "trim delay: %d", i);
            delay += i;
            if (type == VCU_CH_PHYSICAL) {
                i = distance_delay(ri, n);
                elog(LOGDEBUG, "distance delay: %d", i);
                delay += i;
            }
            elog(LOGDEBUG, "resulting delay OUT [%d][%d]: %d", type, n, delay);
            bit_set(mask, n);
            bfir_set_delay(OUT, type, mask, n, n, delay);
            bit_clr(mask, n);
        }
    }

    if (io == IN) {

        for (n = start; n <= stop; n++) {
            delay = runtime_delay_map.eqalign[eq_mode+1].delay[io][type][n];
            if (eq_mode >= 0 &&
                cu.eq.mode[eq_mode].prgs[io][type][n] >= 0 &&
                cu.eq.mode[eq_mode].prgs[io][type][n] < cu.eq.eqprg_count)
            {
                eq = equalisers[cu.eq.mode[eq_mode].prgs[io][type][n]];
                delay -= eq->sample_delay[ri];
            }
            elog(LOGDEBUG, "alignment delay: %d", delay);
            delay += (int)((cu.channel.trim.delay[IN][type][n]) *
                           (double)runtime_sample_rate);
            elog(LOGDEBUG, "resulting delay IN [%d][%d]: %d", type, n, delay);
            bit_set(mask, n);
            bfir_set_delay(IN, type, mask, n, n, delay);
            bit_clr(mask, n);
        }
    }
    
    bfir_commit_buffered_commands();
}


static int
get_eq_delay(int eq_mode,
             int io,
             int type,
             int idx,
             int rate_index)
{
    struct equaliser *eq;
    
    if (eq_mode < 0 || cu.eq.mode[eq_mode].prgs[io][type][idx] < 0) {
        return 0;
    }
    eq = equalisers[cu.eq.mode[eq_mode].prgs[io][type][idx]];
    if (eq == NULL) {
        elog(LOGERROR, "eq is NULL: %d %d %d %d %d",
             eq_mode, io, type, idx, cu.eq.mode[eq_mode].prgs[io][type][idx]);
        exit(EXIT_FAILURE);
    }
    return eq->sample_delay[rate_index];
}

static int
sample_align_eqmode(struct eq_mode_delay_alignment *eqalign,
                    int eq_mode,
                    int rate_index,
                    struct bfir_chmap *phys_maps[2])
{
    int n, type, delay, i, k, maxdelay, io, delay1, delay2, phys, logi;
    int maxtpos, maxcpos, flowthrough_delay;
    int delaytab[VCU_CH_COUNT][VCU_MAX_CHANNELS];
    int delays[VCU_MAX_CHANNELS];
    uint32_t mask[VCU_CH_COUNT][VCU_MAX_CHANNELS/32];
    uint32_t assigned[VCU_MAX_CHANNELS/32];
    uint32_t group[VCU_MAX_CHANNELS/32];
    
    elog(LOGDEBUG, "start aligning eq mode %d", eq_mode);
    
    /* align physical inputs */
    elog(LOGDEBUG, "align physical inputs");
    memset(assigned, 0, sizeof(assigned));
    for (n = 0; n < cu.channel.count[IN][VCU_CH_PHYSICAL]; n++) {
        if (bit_isset(assigned, n)) {
            continue;
        }
        /* which physical inputs mix to the same logical channels as this? */
        memcpy(mask, phys_maps[IN][n].channels, sizeof(mask));
        memset(group, 0, sizeof(group));
        bit_set(assigned, n);
        bit_set(group, n);
        for (i = 0; i < cu.channel.count[IN][VCU_CH_PHYSICAL]; i++) {
            if (bit_isset(assigned, i)) {
                continue;
            }
            for (type = 0; type < VCU_CH_COUNT; type++) {
                for (k = 0; k < VCU_MAX_CHANNELS/32; k++) {
                    if ((mask[type][k] &
                         phys_maps[IN][i].channels[type][k]) != 0)
                    {
                        for (k = 0; k < VCU_MAX_CHANNELS/32; k++) {
                            mask[type][k] |= phys_maps[IN][i].channels[type][k];
                        }
                        goto align_physical_inputs_post_loop;
                    }
                }
            }
        align_physical_inputs_post_loop:
            if (type != VCU_CH_COUNT) {
                bit_set(assigned, i);
                bit_set(group, i);
                i = -1;
            }
        }

        maxdelay = 0;
        i = -1;
        while ((i = bit_find(group, i + 1,
                             cu.channel.count[IN][VCU_CH_PHYSICAL] - 1)) != -1)
        {
            delay = get_eq_delay(eq_mode, IN, VCU_CH_PHYSICAL, n, rate_index);
            if (delay > maxdelay) {
                maxdelay = delay;
            }
        }
        elog(LOGDEBUG, "maxdelay is %d for phys-in %s",
             maxdelay, strmask(group, 0, VCU_MAX_CHANNELS - 1));
        i = -1;
        while ((i = bit_find(group, i + 1,
                             cu.channel.count[IN][VCU_CH_PHYSICAL] - 1)) != -1)
        {            
            eqalign->delay[IN][VCU_CH_PHYSICAL][i] = maxdelay;
        }
    }

    /* align logical inputs */
    elog(LOGDEBUG, "align logical inputs");
    for (type = 0; type < VCU_CH_COUNT; type++) {
        switch (type) {
        case VCU_CH_AMBIOPOLE:
            /* inputs must be aligned per ambiopole */
            maxdelay = 0;
            for (n = 0; n < cu.channel.count[IN][type]; n += 2) {
                
                phys = bit_find(cu.channel.map[IN][type][n], 0,
                                cu.channel.count[IN][VCU_CH_PHYSICAL] - 1);
                delay1 = eqalign->delay[IN][VCU_CH_PHYSICAL][phys] +
                    get_eq_delay(eq_mode, IN, type, n, rate_index);
                
                phys = bit_find(cu.channel.map[IN][type][n+1], 0,
                                cu.channel.count[IN][VCU_CH_PHYSICAL] - 1);
                delay2 = eqalign->delay[IN][VCU_CH_PHYSICAL][phys] +
                    get_eq_delay(eq_mode, IN, type, n+1, rate_index);
                
                if (delay1 > delay2) {
                    eqalign->delay[IN][type][n+1] =
                        get_eq_delay(eq_mode, IN, type, n+1, rate_index) +
                        delay1 - delay2;
                    eqalign->delay[IN][type][n] =
                        get_eq_delay(eq_mode, IN, type, n, rate_index);
                } else {
                    eqalign->delay[IN][type][n] =
                        get_eq_delay(eq_mode, IN, type, n, rate_index) +
                        delay2 - delay1;
                    eqalign->delay[IN][type][n+1] =
                        get_eq_delay(eq_mode, IN, type, n+1, rate_index);
                }
            }
            break;
        case VCU_CH_REVERB:
        case VCU_CH_AMBISONIC:
            /* all inputs must be aligned */
            maxdelay = 0;
            memset(delays, 0, sizeof(delays));
            for (n = 0; n < cu.channel.count[IN][type]; n++) {
                phys = bit_find(cu.channel.map[IN][type][n], 0,
                                cu.channel.count[IN][VCU_CH_PHYSICAL] - 1);
                delays[n] = eqalign->delay[IN][VCU_CH_PHYSICAL][phys] +
                    get_eq_delay(eq_mode, IN, type, n, rate_index);
                if (delays[n] > maxdelay) {
                    maxdelay = delays[n];
                }
            }
            elog(LOGDEBUG, "maxdelay for %s-in: %d", strtype(type), maxdelay);
            for (n = 0; n < cu.channel.count[IN][type]; n++) {
                eqalign->delay[IN][type][n] =
                    get_eq_delay(eq_mode, IN, type, n, rate_index) +
                    maxdelay - delays[n];
            }
            break;
        case VCU_CH_PASSTHRU:
            /* no alignment necessary */
            for (n = 0; n < cu.channel.count[IN][type]; n++) {
                eqalign->delay[IN][type][n] =
                    get_eq_delay(eq_mode, IN, type, n, rate_index);
            }
            break;
        }
    }

    /* align logical outputs: all logical outputs that mix to the same physical
       outputs must be aligned */
    elog(LOGDEBUG, "align logical outputs");

    /* calculate accumulated delay for all logical outputs */
    memset(delaytab, 0, sizeof(delaytab));
    maxdelay = 0;
    maxcpos = 0;
    maxtpos = 0;
    for (type = 0; type < VCU_CH_COUNT; type++) {
        for (n = 0; n < cu.channel.count[OUT][type]; n++) {
            phys = bit_find(cu.channel.map[IN][type][n], 0,
                            cu.channel.count[IN][VCU_CH_PHYSICAL] - 1);
            delay = eqalign->delay[IN][VCU_CH_PHYSICAL][phys];
            switch (type) {
            case VCU_CH_AMBIOPOLE:
                delay += eqalign->delay[IN][type][n];
                delay += runtime_delay_map.xtc_maxdelay;
                break;
            case VCU_CH_REVERB:
            case VCU_CH_AMBISONIC:
                delay += eqalign->delay[IN][type][0];
                break;
            case VCU_CH_PASSTHRU:
                delay += eqalign->delay[IN][type][n];
                break;
            }
            delay += get_eq_delay(eq_mode, OUT, type, n, rate_index);
            delaytab[type][n] = delay;
            elog(LOGDEBUG, "accumulated delay to %s-out %d: %d",
                 strtype(type), n, delaytab[type][n]);
        }
    }

    memset(mask, 0, sizeof(mask));
    while (true) {
        /* start with the highest delay logical output to ensure proper
           alignment (one logical output can mix to several physical) */
        maxdelay = 0;
        maxtpos = -1;
        maxcpos = -1;
        for (type = 0; type < VCU_CH_COUNT; type++) {
            for (n = 0; n < cu.channel.count[OUT][type]; n++) {
                if (bit_isset(mask[type], n)) {
                    continue;
                }
                if (eqalign->delay[OUT][type][n] > 0) {
                    delay = eqalign->delay[OUT][type][n];
                } else {
                    delay = delaytab[type][n];
                }
                if (delay >= maxdelay) {
                    maxdelay = delay;
                    maxtpos = type;
                    maxcpos = n;
                }
            }
        }
        bit_set(mask[maxtpos], maxcpos);
        if (maxtpos == -1) {
            break;
        }
        phys = -1;
        while ((phys = bit_find(cu.channel.map[OUT][maxtpos][maxcpos], phys + 1,
                                cu.channel.count[OUT][VCU_CH_PHYSICAL] - 1))
               != -1)
        {
            for (type = 0; type < VCU_CH_COUNT; type++) {
                i = -1;
                while ((i = bit_find(phys_maps[OUT][phys].channels[type], i + 1,
                                     cu.channel.count[OUT][type] - 1)) != -1)
                {
                    if (eqalign->delay[OUT][type][i] > 0) {
                        continue;
                    }
                    eqalign->delay[OUT][type][i] =
                        get_eq_delay(eq_mode, OUT, type, i, rate_index) +
                        maxdelay - delaytab[type][i];
                    if (type == VCU_CH_AMBIOPOLE) {
                        eqalign->delay[OUT][type][i] +=
                            runtime_delay_map.xtc_maxdelay;
                    }
                    elog(LOGDEBUG, "target delay %s-out %d: %d "
                         "(max %s-out %d: %d)",
                         strtype(type), i, eqalign->delay[OUT][type][i],
                         strtype(maxtpos), maxcpos, delaytab[maxtpos][maxcpos]);
                }
            }
        }
    }

    /* align physical outputs: all physical outputs must be aligned */
    elog(LOGDEBUG, "align physical outputs");
    maxdelay = 0;
    memset(delays, 0, sizeof(delays));
    for (n = 0; n < cu.channel.count[OUT][VCU_CH_PHYSICAL]; n++) {
        for (type = 0; type < VCU_CH_COUNT; type++) {
            logi = bit_find(phys_maps[OUT][n].channels[type], 0,
                            cu.channel.count[OUT][type] - 1);
            if (logi != -1) {
                break;
            }
        }
        if (logi == -1) {
            delay = 0;
        } else {
            phys = bit_find(cu.channel.map[IN][type][logi], 0,
                            cu.channel.count[IN][VCU_CH_PHYSICAL] - 1);
            delay = eqalign->delay[OUT][type][logi] +
                eqalign->delay[IN][type][logi] +
                eqalign->delay[IN][VCU_CH_PHYSICAL][phys] +
                get_eq_delay(eq_mode, OUT, VCU_CH_PHYSICAL, n, rate_index);
        }
        delays[n] = delay;
        elog(LOGDEBUG, "accumulated delay for physical-out %d: %d", n, delay);
        if (delay > maxdelay) {
            maxdelay = delay;
        }
    }
    for (n = 0; n < cu.channel.count[OUT][VCU_CH_PHYSICAL]; n++) {
        eqalign->delay[OUT][VCU_CH_PHYSICAL][n] =
            get_eq_delay(eq_mode, OUT, VCU_CH_PHYSICAL, n, rate_index) +
            maxdelay - delays[n];
    }
    flowthrough_delay = maxdelay;

    if (elog_loglevel_get() == LOGDEBUG) {
        FOR_IN_AND_OUT {
            for (type = 0; type < VCU_CH_WP_COUNT; type++) {
                for (n = 0; n < cu.channel.count[io][type]; n++) {
                    elog(LOGDEBUG, "target delay for %s-%s %d: %d",
                         strtype(type), io == IN ? "in" : "out", n,
                         eqalign->delay[io][type][n]);
                }
            }
        }
    }
    
    elog(LOGDEBUG, "finished aligning eq mode %d (flowthrough delay %d)",
         eq_mode, flowthrough_delay);
    return flowthrough_delay;
}

static int
calculate_runtime_delay_map(int rate_index)
{
    int n, maxdelay, delay, mindelay, ri, eq_mode, type, phys;
    int flowthrough_delay[VCU_MAX_EQ_MODES + 1];
    struct eq_mode_delay_alignment *eqalign;
    struct bfir_chmap *phys_maps[2];
    bool_t all_xtc, any_xtc;
    struct xtcprg *xtcprg;
    
    make_bfir_chmap(phys_maps);
    
    runtime_delay_map.xtc_maxdelay = 0;
    runtime_delay_map.delay_buffer_size[IN] = 0;
    runtime_delay_map.delay_buffer_size[OUT] = 0;
    ri = rate_index;

    /* ambiopole delay alignment */
    elog(LOGDEBUG, "ambiopole delay alignment");
    all_xtc = true;
    any_xtc = false;
    for (n = 0; n < cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2; n++) {
        if (cu.ambiopole.use_xtalk_cancel[n]) {
            any_xtc = true;
        } else {
            all_xtc = false;
        }
    }
    if (any_xtc) {
        n = -1;
        mindelay = -1;
        while ((xtcprg = idarray_next(xtc_programs, &n)) != NULL) {
            if (runtime_delay_map.xtc_maxdelay < xtcprg->sample_delay[ri]) {
                runtime_delay_map.xtc_maxdelay = xtcprg->sample_delay[ri];
            }
            if (mindelay == -1 || mindelay > xtcprg->sample_delay[ri]) {
                mindelay = xtcprg->sample_delay[ri];
            }
                
        }
        if (mindelay == -1) {
            elog(LOGDEBUG, "no xtc program");
            return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
        }
        if (all_xtc) {
            runtime_delay_map.delay_buffer_size[OUT] +=
                runtime_delay_map.xtc_maxdelay - mindelay;
        } else {
            runtime_delay_map.delay_buffer_size[OUT] +=
                runtime_delay_map.xtc_maxdelay;
        }
    }
    elog(LOGDEBUG, "xtc_maxdelay = %d", runtime_delay_map.xtc_maxdelay);

    runtime_delay_map.eqalign = realloc(runtime_delay_map.eqalign,
                                        (cu.eq.mode_count + 1) *
                                        sizeof(struct eq_mode_delay_alignment));
    memset(runtime_delay_map.eqalign, 0,
           (cu.eq.mode_count + 1) * sizeof(struct eq_mode_delay_alignment));

    /* align all eq modes */
    elog(LOGDEBUG, "align all eq modes");
    for (eq_mode = -1; eq_mode < cu.eq.mode_count; eq_mode++) {
        flowthrough_delay[eq_mode+1] =
            sample_align_eqmode(&runtime_delay_map.eqalign[eq_mode+1], eq_mode,
                                ri, phys_maps);
    }

    /* calculate flowthrough delay for all modes, and then align them */
    elog(LOGDEBUG, "flowthrough delay alignment");
    maxdelay = 0;
    for (eq_mode = -1; eq_mode < cu.eq.mode_count; eq_mode++) {
        if (maxdelay < flowthrough_delay[eq_mode+1]) {
            maxdelay = flowthrough_delay[eq_mode+1];
        }            
    }
    runtime_delay_map.flowthrough_delay = maxdelay;
    for (eq_mode = -1; eq_mode < cu.eq.mode_count; eq_mode++) {
        for (n = 0; n < cu.channel.count[OUT][VCU_CH_PHYSICAL]; n++) {
            runtime_delay_map.eqalign[eq_mode+1].delay[OUT][VCU_CH_PHYSICAL][n]
                += maxdelay - flowthrough_delay[eq_mode+1];
        }
    }

    elog(LOGDEBUG, "calculate delay buffer size");
    /* FIXME: delay buffer sizes could be more precisely calculated */
    
    /* 2 * trim delay = both physical and logical trim delay */    
    runtime_delay_map.delay_buffer_size[IN] =
        (int)(2 * VCU_TRIM_DELAY_RANGE * (double)cu.signal.rates[ri] * 1.05);

    runtime_delay_map.delay_buffer_size[OUT] =
        (int)(2 * VCU_TRIM_DELAY_RANGE * (double)cu.signal.rates[ri] * 1.05) +
        (int)(VCU_REVERB_DELAY_RANGE * (double)cu.signal.rates[ri] * 1.05);

    maxdelay = 0;
    for (n = 0; n < cu.channel.count[OUT][VCU_CH_PHYSICAL]; n++) {
        if (maxdelay < distance_delay(ri, n)) {
            maxdelay = distance_delay(ri, n);
        }
    }
    runtime_delay_map.delay_buffer_size[OUT] += maxdelay;
    
    maxdelay = 0;
    mindelay = 0;
    for (type = 0; type < VCU_CH_COUNT; type++) {
        for (n = 0; n < cu.channel.count[IN][type]; n++) {
            for (eq_mode = -1; eq_mode < cu.eq.mode_count; eq_mode++) {
                eqalign = &runtime_delay_map.eqalign[eq_mode+1];
                phys = bit_find(cu.channel.map[IN][type][n], 0,
                                cu.channel.count[IN][VCU_CH_PHYSICAL] - 1);
                delay = eqalign->delay[IN][VCU_CH_PHYSICAL][phys] +
                    eqalign->delay[IN][type][n];
                if (maxdelay < delay) {
                    maxdelay = delay;
                }
                if (mindelay == 0 || mindelay > delay) {
                    mindelay = delay;
                }
            }
        }
    }
    runtime_delay_map.delay_buffer_size[IN] += maxdelay;

    delay = 0;
    maxdelay = 0;
    for (n = 0; n < cu.channel.count[OUT][VCU_CH_PHYSICAL]; n++) {
        for (eq_mode = -1; eq_mode < cu.eq.mode_count; eq_mode++) {
            if (maxdelay < flowthrough_delay[eq_mode+1]) {
                maxdelay = flowthrough_delay[eq_mode+1];
            }
        }
    }
    runtime_delay_map.delay_buffer_size[OUT] += maxdelay - mindelay;
    
    return 0;
}

static void
apply_mute(int io,
           int type,
           int idx)
{
    uint32_t mask[BFIR_MAX_OUTPUT_CHANNELS / 32 + 1];
    int n, i, start, end;
    bool_t mute_group;
    
    if (idx < 0) {
        start = 0;
        end = cu.channel.count[io][type]  - 1;
    } else {
        start = idx;
        end = idx;
    }
    
    memset(mask, 0xFF, sizeof(mask));
    if (type == VCU_CH_PHYSICAL) {
        for (n = start; n <= end; n++) {
            if ((io == OUT && cu.volume.master_temp_mute) ||
                cu.volume.mute[io][type][n])
            {
                bit_clr(mask, n);
            }
        }
    } else {
        for (n = start; n <= end; n++) {
            mute_group = false;
            for (i = 0; i < cu.channel.group_count; i++) {
                if (cu.channel.active_mode >= 0 &&
                    !bit_isset(cu.channel.mode[cu.channel.active_mode].groups,
                               i))
                {
                    /* group is not active */
                    continue;
                }
                if (cu.volume.group_mute[i] &&
                    bit_isset(cu.channel.group[i].channels[io][type], n))
                {
                    mute_group = true;
                    break;
                }
            }
            if (mute_group ||
                cu.volume.mute[io][type][n] ||
                !vcu.runtime.channel.get_active(io, type, n))
            {
                bit_clr(mask, n);
            }
        }
    }
    elog(LOGDEBUG, "%s-%s-(%d-%d) not muted: %s",
         strtype(type), io == IN ? "in" : "out", start, end,
         strmask(mask, start, end));
    bfir_set_active(io, type, mask, start, end);
}

static void
apply_eq(int io,
         int type,
         int idx)
{
    uint32_t mask[VCU_MAX_CHANNELS/32];
    int start, stop, eq_mode, n;
    struct equaliser *eq;
    
    bfir_start_buffering_commands();

    if (idx < 0) {
        start = 0;
        stop = cu.channel.count[io][type] - 1;
    } else if (idx < cu.channel.count[io][type]) {
        start = stop = idx;
    } else {
        return;
    }

    eq_mode = get_active_eq_mode();
    
    memset(mask, 0, sizeof(mask));
    if (cu.channel.count[io][type] > 0 && eq_mode >= 0) {
        for (n = start; n <= stop; n++) {
            eq = equalisers[cu.eq.mode[eq_mode].prgs[io][type][n]];
            bit_set(mask, n);
            bfir_set_equalisation(mask, type,
                                  eq == NULL ? -1 : eq->bfir_id, io);
            bit_clr(mask, n);
        }
    } else {
        memset(mask, 0xFF, sizeof(mask));
        bfir_set_equalisation(mask, type, -1, io);
        memset(mask, 0, sizeof(mask));
    }        
    
    bfir_commit_buffered_commands();
}

/*
 * exported functions below
 */

static bool_t
runtime_isruntime(void)
{
    return isruntime;
}

static int
runtime_start(void (*peak_meter)(double peak),
	      void (*process_terminated)(int reason))
{
    return vcu.runtime.start_file_input(peak_meter, process_terminated,
                                        0, 0, false, NULL);
}

static int
runtime_start_file_input(void (*peak_meter)(double peak),
                         void (*process_terminated)(int reason),
                         int sample_format,
                         int channel_count,
                         bool_t loop,
                         const char path[])
{
    int sample_rate, n, i, ri, type, mode, count, io;
    uint32_t mask[BFIR_MAX_OUTPUT_CHANNELS / 32 + 1];
    struct revprg *revprg = NULL;
    struct vcu_eqpart *eqpart;
    struct bfirdyneq dyneq;
    struct xtcprg *xtcprg;
    struct equaliser *eq;
    struct stat buf;

    if (channel_count != 0) {
        if (path == NULL || stat(path, &buf) == -1) {
            return VCU_ERR_NO_SIGNAL;
        }
    }
    
    elog(LOGDEBUG, "runtime_start() called");
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    MUST_NOT_NEED_TO_UPDATE_HARD_SETTINGS(VCU_ERR_PREREQUISITES_NOT_FULFILLED);

    update_bfir_soft_settings();
    
    /* verify that we have selected a reverb program, (if we need one) */    
    if (cu.channel.count[OUT][VCU_CH_REVERB] > 0 &&
	(revprg = get_revprg(cu.reverb.current)) == NULL)
    {
	elog(LOGDEBUG, "No reverb program selected");
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    /* verify that we have a valid and supported signal on the input */
    if ((sample_rate = vcu.runtime.signal.get_input_sample_rate()) < 0) {
	return sample_rate;
    }
    for (ri = 0; ri < cu.signal.n_rates; ri++) {
	if (cu.signal.rates[ri] == sample_rate) {
	    break;
	}
    }
    if (ri == cu.signal.n_rates) {
	elog(LOGDEBUG, "No valid sample rate");
        return VCU_ERR_UNSUPPORTED_SAMPLE_RATE;
    }
    runtime_sample_rate = sample_rate;
    memset(mask, 0, sizeof(mask));

    elog(LOGDEBUG, "calculate runtime delay map");
    if ((n = calculate_runtime_delay_map(ri)) < 0) {
        elog(LOGDEBUG, "failed to calculate runtime delay map");
        return n;
    }
    
    for (type = 0; type < VCU_CH_WP_COUNT; type++) {
        elog(LOGDEBUG, "SAMPLE ALIGN type %d", type);
        sample_align(IN, type, -1);
        sample_align(OUT, type, -1);
        elog(LOGDEBUG, "APPLY VOLUME type %d", type);
        apply_volume(IN, type, -1, true);
        apply_volume(OUT, type, -1, true);
        elog(LOGDEBUG, "APPLY MUTE type %d", type);
        apply_mute(IN, type, -1);
        apply_mute(OUT, type, -1);
        elog(LOGDEBUG, "APPLY EQ type %d", type);
        apply_eq(IN, type, -1);
        apply_eq(OUT, type, -1);
    }
    
    /* Assign dynamic equalisers */
    elog(LOGDEBUG, "assign dynamic equalisers");
    bfir_set_dyneq(-1, NULL);
    for (n = 0; n < cu.eq.eqprg_count; n++) {
        if (cu.eq.eqprg[n].part_count != 1 ||
            cu.eq.eqprg[n].parts[0].type != VCU_EQPRG_TYPE_FFT_DYNAMIC)
        {
            continue;
        }
        eqpart = cu.eq.eqprg[n].parts;
        eq = equalisers[n];
        memset(&dyneq, 0, sizeof(dyneq));
        dyneq.n_bands = eqpart->params.fft.n_bands;
        for (i = 0; i < dyneq.n_bands; i++) {
            dyneq.band[i] = eqpart->params.fft.band[i];
            dyneq.amp[i] = eqpart->params.fft.amp[i];
        }
        bfir_set_dyneq(eq->bfir_id, &dyneq);
    }
    
    /* Assign xtc program */
    elog(LOGDEBUG, "assign xtc program");
    for (n = 0; n < cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2; n++) {
        if (cu.ambiopole.use_xtalk_cancel[n]) {
            xtcprg = get_xtcprg(cu.ambiopole.current[n]);
	    bit_set(mask, n);
            bfir_set_xtalkcancel(mask, xtcprg->bfir_id);
	    bit_clr(mask, n);
        }
    }
    /* Assign reverb program */
    elog(LOGDEBUG, "assign reverb program");
    if (cu.channel.count[OUT][VCU_CH_REVERB] > 0) {
        bfir_set_reverbprg(revprg->bfir_id);
    }

    /* Build runtime_active_map for fast channel active lookups in runtime */
    elog(LOGDEBUG, "build runtime active map");
    mode = cu.channel.active_mode;
    if (mode < 0) {
        count = 1;
    } else {
        count = cu.channel.mode_count;
    }
    memset(&runtime_active_map, 0, sizeof(runtime_active_map));
    for (n = 0; n < count; n++) {
        if (mode >= 0) {
            cu.channel.active_mode = n;
        }
        elog(LOGDEBUG, "testing channel mode %d", cu.channel.active_mode);
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            if (vcu.runtime.channel.get_type_active(type)) {
                bit_set(runtime_active_map.type[type], n);
            }
            FOR_IN_AND_OUT {
                for (i = 0; i < cu.channel.count[io][type]; i++) {
                    if (vcu.runtime.channel.get_active(io, type, i)) {
                        bit_set(runtime_active_map.ch[io][type][i], n);
                    }
                }
            }
            
        }
    }
    cu.channel.active_mode = mode;

    /* Should we suspend a disk? */
    if (conf_static.suspend_in_runtime &&
        conf_static.suspend_command[0] != '\0')
    {
        periodic();
        elog(LOGDEBUG, "running suspend command \"%s\"",
             conf_static.suspend_command);
        system(conf_static.suspend_command);
    }
    
    /* Now we try to start the filter engine */
    process_terminated_cb = process_terminated;
    
    bfir_flush_logs();
    
    elog(LOGDEBUG, "calling bfir_start()");
    if ((bf_fd = bfir_start(sample_rate,
			    runtime_delay_map.delay_buffer_size,
                            cu.ctl.powersave,
                            cu.signal.watch_clock,
			    peak_meter,
			    process_terminated_callback,
                            sample_format,
                            channel_count,
                            loop,
                            path)) < 0)
    {
	elog(LOGDEBUG, "Could not start convolver: %s",
             vcu.ctl.strerror(bf_fd));
	return bf_fd;
    }
    if (bf_fd > maxfd) {
	maxfd = bf_fd;
    }
    FD_SET(bf_fd, &readfds);
    isruntime = true;
    elog(LOGDEBUG, "success, we are now in runtime (bf_fd: %d)", bf_fd);
    return 0;
}

static void
runtime_stop(void)
{
    bfir_stop();
}

static void
runtime_buffer_commands(bool_t do_buffer)
{
    if (do_buffer) {
        bfir_start_buffering_commands();
    } else {
        bfir_commit_buffered_commands();
    }
}

static int
runtime_signal_get_input_sample_rate(void)
{
    static struct timeval lasttv = { 0, 0 };
    static int sample_rate;
    struct timeval tv, tv1;

    if (vcu.runtime.isruntime()) {
	return runtime_sample_rate;
    }
    /* since measuring the sample rate takes some time, we don't do it on
       every function call if they are close */
    gettimeofday(&tv, NULL);
    tv1.tv_sec = 0;
    tv1.tv_usec = 500000;
    timeradd(&lasttv, &tv1, &tv1);
    if (lasttv.tv_sec != 0 && timercmp(&tv, &tv1, <)) {
        return sample_rate;
    }
    if ((sample_rate = sysmon_sampling_rate
         (conf_static.input.device[conf_static.master_input_device],
          conf_static.input.device_params[conf_static.master_input_device],
          conf_static.input.open_channels[conf_static.master_input_device],
          conf_static.input.sample_formats[0],
          cu.signal.suggested_rate))
	== SYSMON_NA)
    {
	return VCU_ERR_NO_SIGNAL;
    }
    if (sample_rate == 0) {
	/* input device has no sample clock, just pick the suggested rate */
	elog(LOGDEBUG, "no sample clock, returning suggested rate %dHz",
	     cu.signal.suggested_rate);
	return cu.signal.suggested_rate;
    }
    gettimeofday(&lasttv, NULL);
    return sample_rate;
}

static void
runtime_signal_reset_peak_meter(void)
{
    bfir_reset_peak_meter();
}

static void
runtime_volume_get_range(double range[2])
{
    if (range == NULL) {
        return;
    }
    range[0] = VCU_VOLUME_MIN;
    range[1] = VCU_VOLUME_MAX;
}

static double
runtime_volume_get_master(void)
{
    return cu.volume.master;
}

static int
runtime_volume_set_master(double volume)
{
    double range[2];
    int type;

    runtime_volume_get_range(range);
    if (volume < range[0]) {
	volume = range[0];
    }
    if (volume > range[1]) {
	volume = range[1];
    }
    if (volume == cu.volume.master) {
	return 0;
    }
    cu.volume.master = volume;
    save_userconf = true;
    if (runtime_isruntime()) {
        bfir_start_buffering_commands();
        for (type = 0; type < VCU_CH_WP_COUNT; type++) {
            apply_volume(IN, type, -1, false);
            apply_volume(OUT, type, -1, false);
        }
        bfir_commit_buffered_commands();
    }
    return 0;
}

static double
runtime_volume_get_group(int idx)
{
    if (idx < 0 || idx >= cu.channel.group_count) {
        return 0.0;
    }
    return cu.volume.group[idx];
}

static int
runtime_volume_set_group(int idx,
                         double volume)
{
    double range[2];
    int type;
    
    if (idx < 0 || idx >= cu.channel.group_count) {
        return VCU_ERR_INVALID_VALUE;
    }
    runtime_volume_get_range(range);
    if (volume < range[0]) {
	volume = range[0];
    }
    if (volume > range[1]) {
	volume = range[1];
    }
    if (volume == cu.volume.group[idx]) {
	return 0;
    }
    cu.volume.group[idx] = volume;    
    save_userconf = true;
    if (runtime_isruntime()) {
        bfir_start_buffering_commands();
        for (type = 0; type < VCU_CH_COUNT; type++) {
            apply_volume(IN, type, -1, false);
            apply_volume(OUT, type, -1, false);
        }
        bfir_commit_buffered_commands();
    }
    return 0;
}

static bool_t
runtime_volume_get_mute(int io,
                        int type,
			int idx)
{
    if (type < 0 || type >= VCU_CH_WP_COUNT ||
        (io != IN && io != OUT) ||
        idx < 0 || idx >= cu.channel.count[io][type])
    {        
	return true;
    }
    return cu.volume.mute[io][type][idx];
}

static int
runtime_volume_set_mute(int io,
                        int type,
			int idx,
			bool_t mute)
{
    if (type < 0 || type >= VCU_CH_WP_COUNT ||
        (io != IN && io != OUT) ||
        idx < 0 || idx >= cu.channel.count[io][type])
    {        
	return VCU_ERR_INVALID_VALUE;
    }
    mute = !!mute;
    if (cu.volume.mute[io][type][idx] == mute) {
        return 0;
    }
    cu.volume.mute[io][type][idx] = mute;
    save_userconf = true;
    apply_mute(io, type, idx);
    return 0;
}

static bool_t
runtime_volume_get_group_mute(int idx)
{
    if (idx < 0 || idx >= cu.channel.group_count) {
        return false;
    }
    return cu.volume.group_mute[idx];
}

static int
runtime_volume_set_group_mute(int idx,
                              bool_t mute)
{
    int type;

    elog(LOGDEBUG, "idx %d; mute %d", idx, mute);    
    if (idx < 0 || idx >= cu.channel.group_count) {
        return VCU_ERR_INVALID_VALUE;
    }
    mute = !!mute;
    if (cu.volume.group_mute[idx] == mute) {
        return 0;
    }
    cu.volume.group_mute[idx] = mute;
    save_userconf = true;
    
    if (!runtime_isruntime()) {
	return 0;
    }
    bfir_start_buffering_commands();
    for (type = 0; type < VCU_CH_COUNT; type++) {
        apply_mute(IN, type, -1);
        apply_mute(OUT, type, -1);
    }
    bfir_commit_buffered_commands();
    return 0;
}

static bool_t
runtime_volume_get_master_temp_mute(void)
{
    return cu.volume.master_temp_mute;
}

static int
runtime_volume_set_master_temp_mute(bool_t mute)
{
    mute = !!mute;
    if (cu.volume.master_temp_mute == mute) {
        return 0;
    }
    cu.volume.master_temp_mute = mute;
    save_userconf = true;
    
    if (!runtime_isruntime()) {
	return 0;
    }
    bfir_start_buffering_commands();
    apply_mute(IN, VCU_CH_PHYSICAL, -1);
    apply_mute(OUT, VCU_CH_PHYSICAL, -1);
    bfir_commit_buffered_commands();
    return 0;
}

static double
runtime_volume_get_trim(int io,
                        int type,
			int idx)
{
    if (type < 0 || type >= VCU_CH_WP_COUNT ||
        (io != IN && io != OUT) ||
        idx < 0 || idx >= cu.channel.count[io][type])
    {        
	return 0.0;
    }
    return cu.channel.trim.volume[io][type][idx];
}

static int
runtime_volume_set_trim(int io,
                        int type,
			int idx,
			double volume)
{
    double range[2];

    elog(LOGDEBUG, "type %d, idx %d, volume %f", type, idx, volume);
    runtime_volume_get_range(range);
    if (type < 0 || type >= VCU_CH_WP_COUNT ||
        (io != IN && io != OUT) ||
        idx < 0 || idx >= cu.channel.count[io][type] ||
        volume < range[0] || volume > range[1])
    {
	return VCU_ERR_INVALID_VALUE;
    }
    if (cu.channel.trim.volume[io][type][idx] == volume) {
        return 0;
    }
    cu.channel.trim.volume[io][type][idx] = volume;    
    save_userconf = true;
    if (runtime_isruntime()) {
        apply_volume(io, type, idx, false);
    }
    return 0;
}

static double
runtime_delay_get_max(void)
{
    return VCU_TRIM_DELAY_RANGE;    
}

static double
runtime_delay_get_trim(int io,
                       int type,
		       int idx)
{
    if (type < 0 || type >= VCU_CH_WP_COUNT ||
        (io != IN && io != OUT) ||
        idx < 0 || idx >= cu.channel.count[io][type])
    {        
	return 0.0;
    }
    return cu.channel.trim.delay[io][type][idx];
}

static int
runtime_delay_set_trim(int io,
                       int type,                       
		       int idx,
		       double delay)
{
    if (type < 0 || type >= VCU_CH_WP_COUNT ||
        (io != IN && io != OUT) ||
        idx < 0 || idx >= cu.channel.count[io][type] ||
        delay < 0 || delay > runtime_delay_get_max())
    {
        elog(LOGDEBUG, "invalid value: io %d; type %d; idx %d; delay %f",
             io, type, idx, delay);
	return VCU_ERR_INVALID_VALUE;
    }
    if (delay == cu.channel.trim.delay[io][type][idx]) {
        elog(LOGDEBUG, "no change");
        return 0;
    }
    cu.channel.trim.delay[io][type][idx] = delay;
    save_userconf = true;
    if (runtime_isruntime()) {
        sample_align(io, type, idx);
    }
    return 0;
}

static const int *
runtime_eq_get_dynamic_equalisers(int *count,
                                  bool_t only_active)
{
    static int indexes[VCU_MAX_EQ_PROGRAMS];
    int n;
    
    if (count == NULL) {
        return NULL;
    }
    *count = 0;
    for (n = 0; n < cu.eq.eqprg_count; n++) {
        if (cu.eq.eqprg[n].part_count != 1 ||
            cu.eq.eqprg[n].parts[0].type != VCU_EQPRG_TYPE_FFT_DYNAMIC)
        {
            continue;
        }
        if (only_active && !is_eq_active(n)) {
            continue;
        }
        indexes[*count] = n;
        (*count)++;
    }
    return indexes;
}

static int
runtime_eq_modify_dynamic_equaliser(int idx,
                                    const struct vcu_eqpart *eqpart)
{
    struct bfirdyneq dyneq;
    int err, n;
    
    if (eqpart == NULL || idx < 0 || idx >= VCU_MAX_EQ_PROGRAMS ||
        cu.eq.eqprg[idx].part_count != 1 ||
        cu.eq.eqprg[idx].parts[0].type != VCU_EQPRG_TYPE_FFT_DYNAMIC ||
        cu.eq.eqprg[idx].parts[0].params.fft.n_bands !=
        eqpart->params.fft.n_bands ||
        memcmp(cu.eq.eqprg[idx].parts[0].params.fft.band,
               eqpart->params.fft.band,
               sizeof(eqpart->params.fft.band)) != 0)
    {
        return VCU_ERR_INVALID_VALUE;
    }
    if ((err = eq_sanitycheck(eqpart, 1, NULL)) != 0) {
        return err;
    }        
    if (memcmp(cu.eq.eqprg[idx].parts[0].params.fft.amp,
               eqpart->params.fft.amp,
               sizeof(eqpart->params.fft.amp)) == 0)
    {
        /* no change */
        return 0;
    }
    memcpy(cu.eq.eqprg[idx].parts[0].params.fft.amp,
           eqpart->params.fft.amp,
           sizeof(eqpart->params.fft.amp));
    
    save_userconf = true;
    
    if (runtime_isruntime()) {
        memset(&dyneq, 0, sizeof(dyneq));
        dyneq.n_bands = eqpart->params.fft.n_bands;
        for (n = 0; n < dyneq.n_bands; n++) {
            dyneq.band[n] = eqpart->params.fft.band[n];
            dyneq.amp[n] = eqpart->params.fft.amp[n];
        }
        bfir_set_dyneq(equalisers[idx]->bfir_id, &dyneq);
    }
    return 0;
}

static void
runtime_eq_get_dynamic_equaliser_volume_range(double range[2])
{
    if (range == NULL) {
        return;
    }
    range[0] = VCU_FFTEQ_VOLUME_MIN;
    range[1] = VCU_FFTEQ_VOLUME_MAX;
}

static double
runtime_eq_get_dynamic_equaliser_smallest_band(void)
{
    return VCU_FFTEQ_SMALLEST_BAND;
}

static int
runtime_eq_get_mode(void)
{
    return get_active_eq_mode();
}

static int
runtime_eq_set_mode(int idx)
{
    int type;

    if (idx < 0) {
        idx = -1;
    }
    if (idx >= vcu.eq.get_mode_count()) {
        elog(LOGDEBUG, "eq mode index %d >= %d", idx, vcu.eq.get_mode_count());
        return VCU_ERR_INVALID_VALUE;
    }
    
    if (cu.channel.active_mode == -1) {
        if (cu.eq.active_mode == idx) {
            return 0;
        }
        cu.eq.active_mode = idx;    
    } else {
        if (cu.channel.mode[cu.channel.active_mode].active_eq_mode == idx) {
            return 0;
        }
        if (idx >= 0) {
            if (!bit_isset(cu.channel.mode[cu.channel.active_mode].eqmodes,
                           idx))
            {
                elog(LOGDEBUG, "eq mode index %d not active");
                return VCU_ERR_INVALID_VALUE;
            }
        }
        cu.channel.mode[cu.channel.active_mode].active_eq_mode = idx;
    }

    save_userconf = true;
    if (!runtime_isruntime()) {
        return 0;
    }
    bfir_start_buffering_commands();
    for (type = 0; type < VCU_CH_WP_COUNT; type++) {
        sample_align(IN, type, -1);
        sample_align(OUT, type, -1);
        apply_eq(IN, type, -1);
        apply_eq(OUT, type, -1);
    }
    bfir_commit_buffered_commands();
    return 0;
}

static bool_t
runtime_eq_get_mode_active(int idx)
{
    if (idx < 0) {
        /* check if any eq mode selection can be made */
        if (cu.eq.mode_count == 0) {
            return false;
        }
        if (cu.channel.active_mode >= 0) {
            return bit_find(cu.channel.mode[cu.channel.active_mode].eqmodes,
                            0, VCU_MAX_EQ_MODES - 1) >= 0;
        }
        return true;
    }
    if (idx >= cu.eq.mode_count) {
        return false;
    }
    if (cu.channel.active_mode >= 0) {
        return bit_isset(cu.channel.mode[cu.channel.active_mode].eqmodes, idx);
    }
    return true;
}

static prgid_t
runtime_reverb_get_id(void)
{
    if (memcmp(&cu.reverb.current, &zeroid, sizeof(zeroid)) != 0 &&
        get_revprg(cu.reverb.current) == NULL)
    {
        memset(&cu.reverb.current, 0, sizeof(cu.reverb.current));
        save_userconf = true;
    }
    return cu.reverb.current;
}

static int
runtime_reverb_set_id(prgid_t id)
{
    struct revprg *revprg;
    
    MUST_NOT_NEED_TO_UPDATE_HARD_SETTINGS(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if ((revprg = get_revprg(id)) == NULL) {
	return VCU_ERR_INVALID_VALUE;
    }
    if (memcmp(&cu.reverb.current, &id, sizeof(id)) == 0) {
	return 0;
    }
    cu.reverb.current = id;
    if (runtime_isruntime()) {
        bfir_start_buffering_commands();
	bfir_set_reverbprg(revprg->bfir_id);
        apply_volume(OUT, VCU_CH_REVERB, -1, false);
        sample_align(OUT, VCU_CH_REVERB, -1);
        bfir_commit_buffered_commands();
    }
    save_userconf = true;
    return 0;
}

static void
runtime_reverb_get_trim_volume_range(double range[2])
{
    if (range == NULL) {
	return;
    }
    range[0] = VCU_VOLUME_MIN;
    range[1] = VCU_VOLUME_MAX;
}

static double
runtime_reverb_get_trim_volume(void)
{
    struct revprg *revprg;
    
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return 0.0;
    }
    return revprg->conf.volume;
}

static int
runtime_reverb_set_trim_volume(double volume)
{
    struct revprg *revprg;
    double range[2];

    runtime_reverb_get_trim_volume_range(range);
    if (volume < range[0] || volume > range[1]) {
	return VCU_ERR_INVALID_VALUE;
    }
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    if (revprg->conf.volume == volume) {
	return 0;
    }
    revprg->conf.volume = volume;
    revprg->save = true;
    if (runtime_isruntime()) {
	apply_volume(OUT, VCU_CH_REVERB, -1, false);
    }
    return 0;
}

static void
runtime_reverb_get_xyz_balance_range(double range[2])
{
    if (range == NULL) {
        return;
    }
    range[0] = -VCU_REVERB_BALANCE_RANGE;
    range[1] = VCU_REVERB_BALANCE_RANGE;
}

static void
runtime_reverb_get_xyz_balance(double balance[3])
{
    struct revprg *revprg;

    if (balance == NULL) {
        return;
    }
    balance[0] = 0.0;
    balance[1] = 0.0;
    balance[2] = 0.0;
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return;
    }
    balance[0] = revprg->conf.xyz_balance[0];
    balance[1] = revprg->conf.xyz_balance[1];
    balance[2] = revprg->conf.xyz_balance[2];
}

static int
runtime_reverb_set_xyz_balance(const double balance[3])
{
    struct revprg *revprg;
    double range[2];
    int n;

    runtime_reverb_get_xyz_balance_range(range);
    if (balance == NULL) {
        elog(LOGDEBUG, "null balance");
        return VCU_ERR_INVALID_VALUE;
    }
    for (n = 0; n < 3; n++) {
        if (balance[n] < range[0] || balance[n] > range[1]) {
            elog(LOGDEBUG, "balance %d out of range %f (range %f - %f)",
                 n, balance[n], range[0], range[1]);
            return VCU_ERR_INVALID_VALUE;
        }
    }
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
        elog(LOGDEBUG, "no reverb program");
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    if (revprg->conf.xyz_balance[0] == balance[0] &&
        revprg->conf.xyz_balance[1] == balance[1] &&
        revprg->conf.xyz_balance[2] == balance[2])
    {
        elog(LOGDEBUG, "no change");
        return 0;
    }
    revprg->conf.xyz_balance[0] = balance[0];
    revprg->conf.xyz_balance[1] = balance[1];
    revprg->conf.xyz_balance[2] = balance[2];
    revprg->save = true;
    if (runtime_isruntime()) {
        apply_volume(OUT, VCU_CH_REVERB, -1, false);
    }
    elog(LOGDEBUG, "new balance applied");
    return 0;
}

static void
runtime_reverb_get_trim_delay_range(double range[2])
{
    if (range == NULL) {
        return;
    }
    range[0] = 0.0;
    range[1] = VCU_REVERB_DELAY_RANGE;
}

static double
runtime_reverb_get_trim_delay(void)
{
    struct revprg *revprg;
    
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return 0.0;
    }
    return revprg->conf.delay;
}

static int
runtime_reverb_set_trim_delay(double delay)
{
    struct revprg *revprg;
    double range[2];

    runtime_reverb_get_trim_delay_range(range);
    if (delay < range[0] || delay > range[1]) {
	return VCU_ERR_INVALID_VALUE;
    }
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    if (revprg->conf.delay == delay) {
        return 0;
    }
    revprg->conf.delay = delay;
    revprg->save = true;
    if (runtime_isruntime()) {
        sample_align(OUT, VCU_CH_REVERB, -1);
    }
    return 0;
}

static void
runtime_reverb_get_scale_length_range(double range[2])
{
    if (range == NULL) {
        return;
    }
    range[0] = 0.01;
    range[1] = 1.0;
}

static double
runtime_reverb_get_scale_length(void)
{
    struct revprg *revprg;
    
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return 0.0;
    }
    return revprg->conf.scale_length;
}

static int
runtime_reverb_set_scale_length(double scale_length)
{
    struct revprg *revprg;
    double range[2];

    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    runtime_reverb_get_scale_length_range(range);
    if (scale_length < range[0] || scale_length > range[1]) {
        scale_length = 0;
    }
    if (scale_length == revprg->conf.scale_length) {
        return 0;
    }
    revprg->conf.scale_length = scale_length;
    revprg->regenerate = true;
    revprg->save = true;
    return 0;
}

static int
runtime_reverb_get_response_mode(void)
{
    struct revprg *revprg;

    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    return revprg->conf.response_mode;
}

static int
runtime_reverb_set_response_mode(int mode)
{
    struct revprg *revprg;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }    
    if (mode == revprg->conf.response_mode) {
        return 0;
    }
    switch (mode) {
    case VCU_REVERB_RESP_MODE_FULL:
    case VCU_REVERB_RESP_MODE_MUTE_DIRECT:
    case VCU_REVERB_RESP_MODE_MUTE_EARLY:
    case VCU_REVERB_RESP_MODE_STAGE_ADAPTED:
        revprg->conf.response_mode = mode;
        break;
    default:
        return VCU_ERR_INVALID_VALUE;
    }
    revprg->regenerate = true;
    revprg->save = true;
    return 0;
}

static int
runtime_reverb_get_decorrelate_mode(void)
{
    struct revprg *revprg;

    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    return revprg->conf.decorrelate_mode;
}

static int
runtime_reverb_set_decorrelate_mode(int mode)
{
    struct revprg *revprg;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }    
    if (mode == revprg->conf.decorrelate_mode) {
        return 0;
    }
    switch (mode) {
    case VCU_REVERB_DECORR_MODE_NONE:
    case VCU_REVERB_DECORR_MODE_ALL:
    case VCU_REVERB_DECORR_MODE_NEEDED:
        revprg->conf.decorrelate_mode = mode;
        break;
    default:
        return VCU_ERR_INVALID_VALUE;
    }
    revprg->regenerate = true;
    revprg->save = true;
    return 0;
}

static bool_t
runtime_reverb_get_force_delay_diff(void)
{
    struct revprg *revprg;

    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return false;
    }
    return revprg->conf.force_delay_diff;
}

static int
runtime_reverb_set_force_delay_diff(bool_t activate)
{
    struct revprg *revprg;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }    
    activate = !!activate;
    if (!activate == !revprg->conf.force_delay_diff) {
        return 0;
    }
    revprg->conf.force_delay_diff = activate;
    revprg->regenerate = true;
    revprg->save = true;
    return 0;
}

static bool_t
runtime_reverb_get_force_decay(void)
{
    struct revprg *revprg;

    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return false;
    }
    return revprg->conf.force_decay;
}

static int
runtime_reverb_set_force_decay(bool_t activate)
{
    struct revprg *revprg;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }    
    activate = !!activate;
    if (!activate == !revprg->conf.force_decay) {
        return 0;
    }
    revprg->conf.force_decay = activate;
    revprg->regenerate = true;
    revprg->save = true;
    return 0;
}

static bool_t
runtime_reverb_get_dry_active(void)
{
    struct revprg *revprg;

    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return false;
    }
    return revprg->conf.mix_with_dry;
}

static int
runtime_reverb_set_dry_active(bool_t activate)
{
    struct revprg *revprg;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }    
    activate = !!activate;
    if (!activate == !revprg->conf.mix_with_dry) {
        return 0;
    }
    revprg->conf.mix_with_dry = activate;
    revprg->regenerate = true;
    revprg->save = true;
    return 0;
}

static double
runtime_reverb_get_dry_volume(void)
{
    struct revprg *revprg;

    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    return revprg->conf.dry_volume;
}

static int
runtime_reverb_set_dry_volume(double volume)
{
    struct revprg *revprg;
    double range[2];
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }    
    runtime_volume_get_range(range);
    if (volume < range[0]) {
	volume = range[0];
    }
    if (volume > range[1]) {
	volume = range[1];
    }
    if (volume == revprg->conf.dry_volume) {
        return 0;
    }
    revprg->conf.dry_volume = volume;
    revprg->regenerate = true;
    revprg->save = true;
    return 0;
}

static double
runtime_reverb_get_bformat_directivity(void)
{
    struct revprg *revprg;

    if ((revprg = get_revprg(cu.reverb.current)) == NULL ||
        revprg->prg.type != FILTERPRG_TYPE_REVERB_BFORMAT)
    {
	return -1;
    }
    return revprg->conf.bformat_directivity;
}

static int
runtime_reverb_set_bformat_directivity(double directivity)
{
    struct revprg *revprg;
    double range[2];
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);

    if ((revprg = get_revprg(cu.reverb.current)) == NULL ||
        revprg->prg.type != FILTERPRG_TYPE_REVERB_BFORMAT)
    {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }    
    vcu.reverb.get_bformat_directivity_range(range);
    if (directivity < range[0]) {
        directivity = range[0];
    }
    if (directivity > range[1]) {
        directivity = range[1];
    }
    if (directivity == revprg->conf.bformat_directivity) {
        return 0;
    }
    revprg->conf.bformat_directivity = directivity;
    revprg->regenerate = true;
    revprg->save = true;
    return 0;
}

static int
runtime_reverb_get_preprocess_properties(double *tail_volume,
                                         double *length,
                                         double *rt60)
{
    struct revprg *revprg;

    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    if (tail_volume != NULL) {
        *tail_volume = revprg->orig_tail_volume;
    }
    if (length != NULL) {
        *length = revprg->prg.length;
    }
    if (rt60 != NULL) {
        *rt60 = revprg->orig_rt60;
    }        
    return 0;
}
    
static int
runtime_reverb_get_postprocess_properties(double *tail_volume,
                                          double *length,
                                          double *rt60)
{
    struct revprg *revprg;

    MUST_NOT_NEED_TO_UPDATE_HARD_SETTINGS(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if ((revprg = get_revprg(cu.reverb.current)) == NULL) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    if (tail_volume != NULL) {
        *tail_volume = revprg->proc_tail_volume;
    }
    if (length != NULL) {
        *length = revprg->proc_length;
    }
    if (rt60 != NULL) {
        *rt60 = revprg->proc_rt60;
    }        
    return 0;
}

static prgid_t
runtime_ambiopole_get_id(int idx)
{
    
    if (idx < 0 || idx >= cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2) {
	return zeroid;
    }
    if (memcmp(&cu.ambiopole.current[idx], &zeroid, sizeof(zeroid)) != 0 &&
        get_xtcprg(cu.ambiopole.current[idx]) == NULL)
    {
        cu.ambiopole.current[idx] = zeroid;
        save_userconf = true;
    }
    return cu.ambiopole.current[idx];
}

static int
runtime_ambiopole_set_id(int idx,
                         prgid_t id)
{
    uint32_t mask[BFIR_MAX_OUTPUT_CHANNELS / 32 + 1];
    struct xtcprg *xtcprg;
    
    if (idx < 0 || idx >= cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    if ((xtcprg = get_xtcprg(id)) == NULL) {
	return VCU_ERR_INVALID_VALUE;
    }
    if (memcmp(&cu.ambiopole.current[idx], &id, sizeof(id)) == 0) {
	return 0;
    }
    cu.ambiopole.current[idx] = id;
    memset(mask, 0, sizeof(mask));
    if (runtime_isruntime() && cu.ambiopole.use_xtalk_cancel[idx]) {
        bfir_start_buffering_commands();
        bit_set(mask, idx);
	bfir_set_xtalkcancel(mask, xtcprg->bfir_id);
        bit_clr(mask, idx);
        sample_align(OUT, VCU_CH_AMBIOPOLE, -1);
        bfir_commit_buffered_commands();
    }
    save_userconf = true;
    return 0;
}


static int
runtime_channel_get_mode(void)
{
    return cu.channel.active_mode;
}

static int
runtime_channel_set_mode(int idx)
{
    int type;

    elog(LOGDEBUG, "setting mode %d", idx);
    if (idx >= vcu.channel.get_mode_count()) {
        elog(LOGDEBUG, "channel mode index %d is invalid", idx);
        return VCU_ERR_INVALID_VALUE;
    }
    if (idx < 0) {
        idx = -1;
    }
    if (cu.channel.active_mode == idx) {
        elog(LOGDEBUG, "no change");
        return 0;
    }
    if (runtime_isruntime() && cu.channel.active_mode >= 0 && idx == -1) {
        elog(LOGDEBUG, "Cannot deactivate channel modes in runtime");
        return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    cu.channel.active_mode = idx;
    save_userconf = true;
    
    if (!runtime_isruntime()) {
        elog(LOGDEBUG, "not in runtime, returning");
	return 0;
    }
    bfir_start_buffering_commands();
    elog(LOGDEBUG, "applying volumes");
    for (type = 0; type < VCU_CH_WP_COUNT; type++) {
        apply_volume(IN, type, -1, false);
        apply_volume(OUT, type, -1, false);
    }
    elog(LOGDEBUG, "applying mute");
    for (type = 0; type < VCU_CH_WP_COUNT; type++) {
        apply_mute(IN, type, -1);
        apply_mute(OUT, type, -1);
    }
    elog(LOGDEBUG, "committing bfir commands");
    bfir_commit_buffered_commands();
    return 0;
}

static bool_t
runtime_channel_get_active(int io,
                           int type,
                           int idx)
{
    return channel_is_active(cu.channel.active_mode < 0 ? NULL :
                             &cu.channel.mode[cu.channel.active_mode],
                             cu.channel.active_mode, true, io, type, idx);
}
    
static bool_t
runtime_channel_get_type_active(int type)
{
    int n;

    if (type < 0 || type >= VCU_CH_WP_COUNT) {
        elog(LOGDEBUG, "invalid type %d", type);
        return false;
    }
    if (type == VCU_CH_PHYSICAL) {
        elog(LOGDEBUG, "physical always active");
        return true;
    }
    if (vcu.channel.get_count(VCU_OUT, type) <= 0) {
        elog(LOGDEBUG, "type %d/%s not active (no channels)", type,
             strtype(type));
        return false;
    }
    if (cu.channel.active_mode < 0) {
        elog(LOGDEBUG, "type %d/%s is active", type, strtype(type));
        return true;
    }
    if (vcu.runtime.isruntime()) {
        return !!bit_isset(runtime_active_map.type[type],
                           cu.channel.active_mode);
    }

    /* check for at least one active channel with the type */
    
    /* io does not matter on logical channels, so we check on IN since they
       are fewer */
    for (n = 0; n < cu.channel.count[IN][type]; n++) {
        if (runtime_channel_get_active(IN, type, n)) {
            elog(LOGDEBUG, "type %d/%s is active", type, strtype(type));
            return true;
        }
    }
    elog(LOGDEBUG, "type %d/%s is not active", type, strtype(type));
    return false;
}

static bool_t
runtime_channel_get_group_active(int idx)
{
    if (idx < 0 || idx > cu.channel.group_count) {
        elog(LOGDEBUG, "invalid idx (%d)", idx);
        return false;
    }
    if (cu.channel.active_mode < 0) {
        elog(LOGDEBUG, "no active mode - group %d is active", idx);
        return true;
    }
    if (bit_isset(cu.channel.mode[cu.channel.active_mode].groups, idx)) {
        elog(LOGDEBUG, "group %d/\"%s\" active in mode %d/\"%s\"",
             idx, cu.channel.group[idx].name, cu.channel.active_mode,
             cu.channel.mode[cu.channel.active_mode].name);
        return true;
    }
    elog(LOGDEBUG, "group %d/\"%s\" not active", idx,
         cu.channel.group[idx].name);
    return false;
}

static bool_t
signal_get_soft_clip(void)
{
    return cu.signal.use_soft_clip;
}

static int
signal_set_soft_clip(bool_t do_soft_clip)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (cu.signal.use_soft_clip == do_soft_clip) {
	return 0;
    }
    cu.signal.use_soft_clip = do_soft_clip;
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    save_userconf = true;
    /* FIXME: this setting does nothing, soft_clip should get into bfir */
    return 0;
}

static int
signal_get_max_dither_bits(void)
{
    int bits[64], n_bits, n;

    vcu.signal.get_allowed_output_bits(bits, &n_bits);

    for (n = 0; n < n_bits; n++) {
        if (!cu.signal.double_precision && bits[n] > 16) {
            if (n == 0) {
                return 0;
            }
            return bits[n-1];
        }
        if (bits[n] >= 32) {
            if (n == 0) {
                return 0;
            }
            return bits[n-1];
        }
    }
    return bits[n-1];
}

static bool_t
signal_get_dither(void)
{
    return cu.signal.apply_dither;
}

static int
signal_set_dither(bool_t apply_dither)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    apply_dither = !!apply_dither;
    if (cu.signal.apply_dither == apply_dither) {
	return 0;
    }
    cu.signal.apply_dither = apply_dither;
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    save_userconf = true;
    return 0;
}

static bool_t
signal_get_watch_clock(void)
{
    return cu.signal.watch_clock;
}

static int
signal_set_watch_clock(bool_t watch_clock)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    watch_clock = !!watch_clock;
    if (cu.signal.watch_clock == watch_clock) {
	return 0;
    }
    cu.signal.watch_clock = watch_clock;
    save_userconf = true;
    return 0;
}

static bool_t
signal_get_double_precision(void)
{
    return cu.signal.double_precision;
}

static int
signal_set_double_precision(bool_t double_precision)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    double_precision = !!double_precision;
    if (cu.signal.double_precision == double_precision) {
	return 0;
    }
    cu.signal.double_precision = double_precision;
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    save_userconf = true;
    return 0;
}

static void
signal_get_allowed_output_bits(int bits[],
			       int *n_bits)
{
    int n;
    
    if (bits == NULL || n_bits == NULL) {
	return;
    }
    /* there are no two same formats with the same bits in the static config */
    *n_bits = conf_static.output.n_sample_formats;
    for (n = 0; n < *n_bits; n++) {
	bits[n] = sampleformat2bits(conf_static.output.sample_formats[n]);
    }
    qsort(bits, *n_bits, sizeof(int), int_compare);
}

static int
signal_get_output_bits(void)
{
    return sampleformat2bits(cu.signal.output_sample_format);
}

static int
signal_set_output_bits(int bits)
{
    int n;

    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    for (n = 0; n < conf_static.output.n_sample_formats; n++) {
	if (sampleformat2bits(conf_static.output.sample_formats[n]) == bits) {
	    if (cu.signal.output_sample_format ==
		conf_static.output.sample_formats[n])
	    {
		return 0;
	    }
	    cu.signal.output_sample_format =
		conf_static.output.sample_formats[n];
	    break;
	}
    }
    if (n == conf_static.output.n_sample_formats) {
	return VCU_ERR_INVALID_VALUE;
    }
    update_bfir_soft_settings();
    save_userconf = true;
    return 0;
}

static void
signal_get_allowed_rates(int rates[],
			 int *n_rates)
{
    if (rates == NULL || n_rates == NULL) {
	return;
    }    
    *n_rates = conf_static.n_rates;
    memcpy(rates, conf_static.rates, (*n_rates) * sizeof(int));
}

static void
signal_get_rates(int rates[],
		 int *n_rates)
{
    if (rates == NULL || n_rates == NULL) {
	return;
    }    
    *n_rates = cu.signal.n_rates;
    memcpy(rates, cu.signal.rates, (*n_rates) * sizeof(int));
}

#warning confusing with reverb.n_blocks -- there are cases without reverb

static int
signal_set_rates(const int rates[],
		 int n_rates)
{
    int _rates[FILTERPRG_RATES_MAX], n, i;
    bool_t valid_rate;

    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (rates == NULL || n_rates < 1 || n_rates > FILTERPRG_RATES_MAX) {
	return VCU_ERR_INVALID_VALUE;
    }
    memset(_rates, 0, sizeof(_rates));
    memcpy(_rates, rates, n_rates * sizeof(int));
    qsort(_rates, n_rates, sizeof(int), int_compare);
    if (memcmp(cu.signal.rates, _rates, n_rates * sizeof(int)) == 0 &&
	n_rates == cu.signal.n_rates)
    {
	return 0;
    }
    for (n = 0; n < n_rates; n++) {
	for (i = 0; i < conf_static.n_rates; i++) {
	    if (_rates[n] == conf_static.rates[i]) {
		break;
	    }
	}
	if (i == conf_static.n_rates) {
	    return VCU_ERR_UNSUPPORTED_SAMPLE_RATE;
	}
	for (i = 0; i < n_rates; i++) {
	    if (i != n && _rates[n] == _rates[i]) {
		return VCU_ERR_INVALID_VALUE;
	    }
	}
        elog(LOGDEBUG, "rates[%d]: %d", n, _rates[n]);
    }
    memcpy(cu.signal.rates, _rates, FILTERPRG_RATES_MAX * sizeof(int));
    cu.signal.n_rates = n_rates;

    /* make sure there are block sizes for all rates, and that the suggested
       rate exists */
    valid_rate = false;
    for (n = 0; n < n_rates; n++) {
	if (cu.reverb.block_sizes[n] == 0) {
	    cu.reverb.block_sizes[n] = BFIR_MIN_BLOCKSIZE_48000HZ;
	    cu.reverb.n_blocks[n] = BFIR_MIN_BLOCKS;
	}
	if (cu.signal.suggested_rate == cu.signal.rates[n]) {
	    valid_rate = true;
	}
    }
    if (!valid_rate) {
	cu.signal.suggested_rate = cu.signal.rates[n_rates - 1];
    }
    
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    save_userconf = true;
    return 0;
}

static int
signal_get_suggested_rate(void)
{
    return cu.signal.suggested_rate;
}

static int
signal_set_suggested_rate(int rate)
{
    int n;

    if (rate == cu.signal.suggested_rate) {
	return 0;
    }
    for (n = 0; n < cu.signal.n_rates; n++) {
	if (rate == cu.signal.rates[n]) {
	    cu.signal.suggested_rate = rate;
	    save_userconf = true;
	    return 0;
	}
    }
    return VCU_ERR_INVALID_VALUE;
}

static int
channel_get_max_count(int io,
                      int type)
{
    if ((io != IN && io != OUT) || type < 0 || type >= VCU_CH_WP_COUNT) {
        return 0;
    }
    if (type == VCU_CH_WP_COUNT) {
        return vcu.channel.get_count(io, type);
    }
    if (io == IN) {
        switch (type) {
        case VCU_CH_REVERB:
            return 3;
        case VCU_CH_AMBISONIC:
            return 9;
        default:
            return BFIR_MAX_OUTPUT_CHANNELS;
        }
    }    
    return BFIR_MAX_OUTPUT_CHANNELS;
}

static int
channel_get_count(int io,
                  int type)
{
    if ((io != IN && io != OUT) || type < 0 || type >= VCU_CH_WP_COUNT) {
        return 0;
    }
    return cu.channel.count[io][type];
}

static int
channel_set_count(int type,
                  int n_channels)
{
    int n, count, io, i, intype;
    bool_t reset;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (type < 0 || type >= VCU_CH_COUNT ||
        n_channels < 0 || n_channels > channel_get_max_count(VCU_OUT, type))
    {
	return VCU_ERR_INVALID_VALUE;
    }
    if (type == VCU_CH_AMBIOPOLE && n_channels % 2 != 0) {
        return VCU_ERR_INVALID_VALUE;
    }
    /* always set value even if it is the same, since channel_set_count is used
       internally to secure physical map */
    cu.channel.count[OUT][type] = n_channels;
    intype = type;

    /* the total amount of channels must not exceed the limit */
    do {
        count = 0;
        for (type = 0; type < VCU_CH_COUNT; type++) {
            count += cu.channel.count[OUT][type];
        }
        if (count > VCU_MAX_CHANNELS) {
            elog(LOGDEBUG, "exceeded max channels");
            for (type = 0; type < VCU_CH_COUNT; type++) {
                if (type != intype && cu.channel.count[OUT][type] != 0) {
                    elog(LOGDEBUG, "setting channel count for type %d to zero",
                         type);
                    cu.channel.count[IN][type] = 0;
                    cu.channel.count[OUT][type] = 0;
                    break;
                }
            }
        }
    } while (count > VCU_MAX_CHANNELS);
    
    do {

        /* correct input channel count */
        for (type = 0; type < VCU_CH_COUNT; type++) {
            if (cu.channel.count[OUT][type] == 0) {
                cu.channel.count[IN][type] = 0;
            } else {
                switch (type) {
                case VCU_CH_REVERB:
                    switch (cu.reverb.source_layout) {
                    case VCU_REVERB_SOURCE_LAYOUT_CENTER:
                        cu.channel.count[IN][type] = 1;
                        break;
                    case VCU_REVERB_SOURCE_LAYOUT_STEREO:
                        cu.channel.count[IN][type] = 2;
                        break;
                    case VCU_REVERB_SOURCE_LAYOUT_LRC:
                        cu.channel.count[IN][type] = 3;
                        break;
                    default:
                        cu.channel.count[IN][type] = 0;
                        break;
                    }
                    break;
                case VCU_CH_AMBISONIC:
                    switch (cu.ambisonics.input_format) {
                    case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:
                        cu.channel.count[IN][type] = 4;
                        break;
                    case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
                        cu.channel.count[IN][type] = 9;
                        break;
                    case VCU_AMBISONIC_INPUT_FORMAT_UHJ:
                        cu.channel.count[IN][type] = 2;
                        break;
                    default:
                        cu.channel.count[IN][type] = 0;
                        break;
                    }
                    break;
                default:
                    cu.channel.count[IN][type] = cu.channel.count[OUT][type];
                    break;
                }
            }
        
            if (type == VCU_CH_AMBISONIC && cu.channel.count[OUT][type] == 0) {
                bfir_delete_filters(uhj_bfir_id);
                uhj_bfir_id = -1;
            }
        }
        
        /* make sure we have a valid phys map */
        FOR_IN_AND_OUT {
            count = 0;
            for (type = 0; type < VCU_CH_COUNT; type++) {
                for (n = 0; n < cu.channel.count[io][type]; n++) {
                    if ((i = bit_count(cu.channel.map[io][type][n], 
                                       cu.channel.count[io][VCU_CH_PHYSICAL]))
                        == 0)
                    {
                        elog(LOGDEBUG, "io %d type %d idx %d is empty, "
                             "making link", io, type, n);
                        bit_set(cu.channel.map[io][type][n], 0);
                        i = 1;
                    }
                    count += i;
                }
            }
            elog(LOGDEBUG, "total link count: %d (io %d)", count, io);
            if (count > VCU_MAX_CHANNELS) {
                elog(LOGDEBUG, "link count exceeded max channels");
                reset = false;
                for (type = 0; type < VCU_CH_COUNT; type++) {
                    if (type != intype && cu.channel.count[OUT][type] != 0) {
                        elog(LOGDEBUG, "setting channel count for type "
                             "%d to zero", type);
                        cu.channel.count[IN][type] = 0;
                        cu.channel.count[OUT][type] = 0;
                        reset = true;
                        break;
                    }
                }
                if (!reset) {
                    /* last resort, we must drop physical mapping */
                    elog(LOGDEBUG, "dropping physical mapping");
                    memset(cu.channel.map, 0, sizeof(cu.channel.map));
                    reset = true;
                }
            }
        }
        
    } while (count > VCU_MAX_CHANNELS);
    
    update_bfir_soft_settings();
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    save_userconf = true;
    
    return 0;
}

static const uint32_t *
channel_get_phys_map(int io,
                     int type,
                     int idx)
{
    if ((io != IN && io != OUT) || type < 0 || type >= VCU_CH_COUNT ||
        idx < 0 || idx >= cu.channel.count[io][type])
    {
        return NULL;
    }
    return cu.channel.map[io][type][idx];
}

static int
channel_set_phys_map(int io,
                     int type,
                     int idx,
                     const uint32_t mask[VCU_MAX_CHANNELS/32])
{
    int err, channel_count[2][VCU_CH_COUNT];    
    uint32_t oldmask[VCU_MAX_CHANNELS/32];
    struct bfir_chmap *maps[2];
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (mask == NULL || (io != IN && io != OUT) ||
        type < 0 || type >= VCU_CH_COUNT ||
        idx < 0 || idx >= cu.channel.count[io][type])
    {
        return VCU_ERR_INVALID_VALUE;
    }
    if (memcmp(cu.channel.map[io][type][idx], mask, SIZEOF_CH_MASK) == 0) {
        return 0;
    }
    memcpy(oldmask, cu.channel.map[io][type][idx], SIZEOF_CH_MASK);
    memcpy(cu.channel.map[io][type][idx], mask, SIZEOF_CH_MASK);
    make_bfir_chmap(maps);
    if ((err = bfir_verify_phys_map((const struct bfir_chmap **)maps,
                                    cu.ambisonics.input_format,
                                    channel_count)) != 0)
    {
        return err;
    }
    if (channel_count[io][type] != cu.channel.count[io][type]) {
        elog(LOGDEBUG, "channel mismatch");
        memcpy(cu.channel.map[io][type][idx], oldmask, SIZEOF_CH_MASK);
        return VCU_ERR_INVALID_VALUE;
    }
    update_bfir_soft_settings();
    save_userconf = true;
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    return 0;
}

static double
channel_distance_max(void)
{
    return VCU_DISTANCE_MAX;
}

static int
channel_position_get(int idx,
                     bool_t *active,
                     double *azimut,
                     double *elevation,
                     double *distance)
{
    if (azimut == NULL || elevation == NULL || distance == NULL ||
        active == NULL || idx < 0 ||
        idx >= conf_static.output.total_open_channels)
    {
	return VCU_ERR_INVALID_VALUE;
    }
    *active = cu.channel.speakerpos[idx].active;
    *azimut = cu.channel.speakerpos[idx].azimut;
    *elevation = cu.channel.speakerpos[idx].elevation;
    *distance = cu.channel.speakerpos[idx].distance;
    return 0;
}

static int
channel_position_set(int idx,
                     bool_t activate,
                     double azimut,
                     double elevation,
                     double distance)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (azimut < -360.0 || azimut > 360.0 ||
        elevation > 90.0 || elevation < -90.0 ||
	distance < 0.0 || distance > VCU_DISTANCE_MAX || idx < 0)
    {
	return VCU_ERR_INVALID_VALUE;
    }
    if (azimut > 180.0) {
	azimut -= 360.0;
    }
    if (azimut < -180.0) {
	azimut += 360.0;
    }
    activate = !!activate;
    if (idx >= conf_static.output.total_open_channels) {
	return VCU_ERR_INVALID_VALUE;
    }
    if (activate == cu.channel.speakerpos[idx].active &&
        azimut == cu.channel.speakerpos[idx].azimut &&
	elevation == cu.channel.speakerpos[idx].elevation &&
	distance == cu.channel.speakerpos[idx].distance)
    {
	return 0;
    }
    elog(LOGDEBUG, "idx: %d, a: %f, e: %f, d: %f (active: %d)",
         idx, azimut, elevation, distance, activate);
    cu.channel.speakerpos[idx].active = activate;
    cu.channel.speakerpos[idx].azimut = azimut;
    cu.channel.speakerpos[idx].elevation = elevation;
    cu.channel.speakerpos[idx].distance = distance;
    update_bfir_soft_settings();
    save_userconf = true;
    return 0;
}

static int
channel_get_stage_max_count(void)
{
    return VCU_MAX_STAGES;
}

static int
channel_get_stage_count(void)
{
    return cu.channel.stage_count;
}

static int
channel_get_stage(int idx,
                  double *azimuth,
                  double *elevation,
                  double *width,
                  double *height)
{
    if (azimuth == NULL || elevation == NULL || width == NULL ||
        height == NULL || idx < 0 || idx >= cu.channel.stage_count)
    {
	return VCU_ERR_INVALID_VALUE;
    }
    *azimuth = cu.channel.stage[idx].azimut;
    *elevation = cu.channel.stage[idx].elevation;
    *width = cu.channel.stage[idx].width;
    *height = cu.channel.stage[idx].height;
    return 0;
}

static int
channel_new_stage(double azimuth,
                  double elevation,
                  double width,
                  double height)
{
    int idx;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (azimuth < -360.0 || azimuth > 360.0 ||
        elevation > 90.0 || elevation < -90.0 ||
	width < 0.0 || width > 360.0 || height < 0.0 || height > 180.0)
    {
	return VCU_ERR_INVALID_VALUE;
    }
    if (azimuth > 180.0) {
	azimuth -= 360.0;
    }
    if (azimuth < -180.0) {
	azimuth += 360.0;
    }
    if (cu.channel.stage_count >= VCU_MAX_STAGES) {
	return VCU_ERR_NO_MEMORY;
    }
    mark_stage_regenerate();
    idx = cu.channel.stage_count;
    cu.channel.stage[idx].azimut = azimuth;
    cu.channel.stage[idx].elevation = elevation;
    cu.channel.stage[idx].width = width;
    cu.channel.stage[idx].height = height;
    cu.channel.stage_count++;
    save_userconf = true;
    return 0;
}

static int
channel_modify_stage(int idx,
                     double azimuth,
                     double elevation,
                     double width,
                     double height)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (azimuth < -360.0 || azimuth > 360.0 ||
        elevation > 90.0 || elevation < -90.0 ||
	width < 0.0 || width > 360.0 || height < 0.0 || height > 180.0 ||
        idx < 0 || idx >= cu.channel.stage_count)
    {
	return VCU_ERR_INVALID_VALUE;
    }
    if (azimuth > 180.0) {
	azimuth -= 360.0;
    }
    if (azimuth < -180.0) {
	azimuth += 360.0;
    }
    if (cu.channel.stage[idx].azimut == azimuth &&
        cu.channel.stage[idx].elevation == elevation &&
        cu.channel.stage[idx].width == width &&
        cu.channel.stage[idx].height == height)
    {
        return 0;
    }
    mark_stage_regenerate();
    cu.channel.stage[idx].azimut = azimuth;
    cu.channel.stage[idx].elevation = elevation;
    cu.channel.stage[idx].width = width;
    cu.channel.stage[idx].height = height;
    save_userconf = true;
    return 0;
}

static int
channel_delete_stage(int idx)
{
    int n;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (idx < 0 || idx >= cu.channel.stage_count) {
        return VCU_ERR_INVALID_VALUE;
    }
    mark_stage_regenerate();
    for (n = idx; n < cu.channel.stage_count - 1; n++){
        cu.channel.stage[n] = cu.channel.stage[n + 1];
    }
    memset(&cu.channel.group[cu.channel.stage_count - 1], 0,
           sizeof(struct stage));
    cu.channel.stage_count--;
    save_userconf = true;
    return 0;
}

static bool_t
channel_get_distance_compensate_active(void)
{
    return cu.channel.distance_compensate;
}

static int
channel_set_distance_compensate_active(bool_t activate)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    activate = !!activate;
    if (cu.channel.distance_compensate == activate) {
        return 0;
    }
    cu.channel.distance_compensate = activate;
    save_userconf = true;
    return 0;
}

static int
channel_get_group_max_count(void)
{
    return VCU_MAX_CHANNEL_GROUPS;
}

static int
channel_get_group_count(void)
{
    return cu.channel.group_count;
}

static const struct vcu_chgrp *
channel_get_group(int idx)
{
    if (idx < 0 || idx >= cu.channel.group_count) {
        return NULL;
    }    
    return &cu.channel.group[idx];
}

static int
channel_new_group(const struct vcu_chgrp *chgrp)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (cu.channel.group_count >= VCU_MAX_CHANNEL_GROUPS) {
	return VCU_ERR_NO_MEMORY;
    }
    if (chgrp == NULL) {
        return VCU_ERR_INVALID_VALUE;
    }
    cu.channel.group[cu.channel.group_count] = *chgrp;
    cu.channel.group[cu.channel.group_count].
        name[VCU_CHGRP_NAME_SIZE - 1] = '\0';
    cu.channel.group_count++;
    save_userconf = true;
    return 0;
}

static int
channel_modify_group(int idx,
                     const struct vcu_chgrp *chgrp)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (idx < 0 || idx >= cu.channel.group_count || chgrp == NULL) {
        return VCU_ERR_INVALID_VALUE;
    }
    if (memcmp(&cu.channel.group[idx], chgrp, sizeof(struct vcu_chgrp)) == 0) {
        return 0;
    }
    cu.channel.group[idx] = *chgrp;
    cu.channel.group[idx].name[VCU_CHGRP_NAME_SIZE - 1] = '\0';
    save_userconf = true;
    return 0;
}

static int
channel_delete_group(int idx)
{
    int n, i;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (idx < 0 || idx >= cu.channel.group_count) {
        return VCU_ERR_INVALID_VALUE;
    }
    for (n = idx; n < cu.channel.group_count - 1; n++){
        /* shift all dependent settings */
        cu.volume.group[n] = cu.volume.group[n + 1];
        cu.channel.group[n] = cu.channel.group[n + 1];
    }
    /* restore defaults at free position */
    cu.volume.group_mute[n] = false;
    cu.volume.group[n] = VCU_VOLUME_MIN;
    memset(&cu.channel.group[n], 0, sizeof(struct vcu_chgrp));

    for (n = 0; n < cu.channel.mode_count; n++) {
        for (i = idx; i < cu.channel.group_count - 1; i++) {
            if (bit_isset(cu.channel.mode[n].groups, i + 1)) {
                bit_set(cu.channel.mode[n].groups, i);
            } else {
                bit_clr(cu.channel.mode[n].groups, i);
            }
        }
        bit_clr(cu.channel.mode[n].groups, i);
        clear_invalid_eqmodes(&cu.channel.mode[n]);
    }
    
    cu.channel.group_count--;
    save_userconf = true;
    return 0;
}

static int
channel_reorder_groups(const int order[])
{
    struct vcu_chgrp group[VCU_MAX_CHANNEL_GROUPS];
    uint32_t mask[VCU_MAX_CHANNEL_GROUPS/32];
    bool_t mute[VCU_MAX_CHANNEL_GROUPS];
    double vol[VCU_MAX_CHANNEL_GROUPS];
    bool_t diff;
    int n, i;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (order == NULL) {
        return VCU_ERR_INVALID_VALUE;
    }
    diff = false;
    for (n = 0; n < cu.channel.group_count; n++) {
        if (order[n] < 0 || order[n] >= cu.channel.group_count) {
            return VCU_ERR_INVALID_VALUE;
        }
        elog(LOGDEBUG, "order[%d] = %d", n, order[n]);
        if (order[n] != n) {
            diff = true;
        }        
    }
    if (!diff || cu.channel.group_count == 0) {
        elog(LOGDEBUG, "no change");
        return 0;
    }
    memcpy(vol, cu.volume.group, sizeof(cu.volume.group));
    memcpy(mute, cu.volume.group_mute, sizeof(cu.volume.group_mute));
    memcpy(group, cu.channel.group, sizeof(cu.channel.group));
    for (n = 0; n < cu.channel.group_count; n++) {
        cu.volume.group[n] = vol[order[n]];
        cu.volume.group_mute[n] = mute[order[n]];
        cu.channel.group[n] = group[order[n]];
    }
    for (n = 0; n < cu.channel.mode_count; n++) {
        memset(mask, 0, sizeof(mask));
        i = -1;
        while ((i = bit_find(cu.channel.mode[n].groups, i + 1,
                             cu.channel.group_count - 1)) != -1)
        {
            bit_set(mask, order[i]);
        }
        memcpy(cu.channel.mode[n].groups, mask, sizeof(mask));
    }
    save_userconf = true;
    return 0;
}

static int
channel_get_mode_max_count(void)
{
    return VCU_MAX_CHANNEL_MODES;
}

static int
channel_get_mode_count(void)
{
    return cu.channel.mode_count;
}

static const struct vcu_chmode *
channel_get_mode(int idx)
{
    if (idx < 0 || idx >= cu.channel.mode_count) {
        return NULL;
    }    
    return &cu.channel.mode[idx];
}

static int
channel_new_mode(const struct vcu_chmode *mode)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (cu.channel.mode_count >= VCU_MAX_CHANNEL_MODES) {
	return VCU_ERR_NO_MEMORY;
    }
    if (!sanitycheck_chmode(mode)) {
        elog(LOGDEBUG, "insane mode");
        return VCU_ERR_INVALID_VALUE;
    }
    cu.channel.mode[cu.channel.mode_count] = *mode;
    cu.channel.mode[cu.channel.mode_count].
        name[VCU_CHMODE_NAME_SIZE - 1] = '\0';
    cu.channel.mode_count++;
    save_userconf = true;
    return 0;
}

static int
channel_modify_mode(int idx,
                    const struct vcu_chmode *mode)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (idx < 0 || idx >= cu.channel.mode_count) {
        return VCU_ERR_INVALID_VALUE;
    }
    if (!sanitycheck_chmode(mode)) {
        elog(LOGDEBUG, "insane mode");
        return VCU_ERR_INVALID_VALUE;
    }
    if (memcmp(&cu.channel.mode[idx], mode, sizeof(struct vcu_chmode)) == 0) {
        return 0;
    }
    cu.channel.mode[idx] = *mode;
    if (cu.channel.mode[idx].active_eq_mode < 0) {
        cu.channel.mode[idx].active_eq_mode = -1;
    }
    cu.channel.mode[idx].name[VCU_CHMODE_NAME_SIZE - 1] = '\0';
    save_userconf = true;
    return 0;
}

static int
channel_delete_mode(int idx)
{
    int n;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (idx < 0 || idx >= cu.channel.mode_count) {
        return VCU_ERR_INVALID_VALUE;
    }
    for (n = idx; n < cu.channel.mode_count - 1; n++){
        cu.channel.mode[n] = cu.channel.mode[n + 1];
    }
    memset(&cu.channel.mode[n], 0, sizeof(struct vcu_chmode));
    if (cu.channel.active_mode > idx) {
        cu.channel.active_mode--;
    } else if (cu.channel.mode_count == 0) {
        cu.channel.active_mode = -1;
    }
    cu.channel.mode_count--;
    save_userconf = true;
    return 0;
}

static int
channel_reorder_modes(const int order[])
{
    struct vcu_chmode mode[VCU_MAX_CHANNEL_MODES];
    bool_t diff;
    int n;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (order == NULL) {
        return VCU_ERR_INVALID_VALUE;
    }
    diff = false;
    for (n = 0; n < cu.channel.mode_count; n++) {
        if (order[n] < 0 || order[n] >= cu.channel.mode_count) {
            return VCU_ERR_INVALID_VALUE;
        }
        if (order[n] != n) {
            diff = true;
        }        
    }
    if (!diff) {
        return 0;
    }
    memcpy(mode, cu.channel.mode, sizeof(cu.channel.mode));
    for (n = 0; n < cu.channel.mode_count; n++) {
        cu.channel.mode[n] = mode[order[n]];
    }
    if (cu.channel.active_mode >= 0) {
        cu.channel.active_mode = order[cu.channel.active_mode];
    }
    save_userconf = true;
    return 0;
}

static int
eq_get_equaliser_count(void)
{
    return cu.eq.eqprg_count;
}

static const struct vcu_eqprg *
eq_get_equaliser(int idx)
{
    if (idx < 0 || idx >= cu.eq.eqprg_count) {
        return NULL;
    }
    return &cu.eq.eqprg[idx];
}

static int
eq_new_equaliser(const struct vcu_eqprg *eqprg)
{
    struct equaliser *eq;
    int err, idx;

    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (eqprg == NULL || eqprg->name[VCU_EQPRG_NAME_SIZE-1] != '\0') {
        return VCU_ERR_INVALID_VALUE;
    }
    if (cu.eq.eqprg_count == VCU_MAX_EQ_PROGRAMS) {
	return VCU_ERR_NO_MEMORY;
    }
    if ((err = new_equaliser(&eq, eqprg->parts, eqprg->part_count)) != 0) {
	return err;
    }
    idx = cu.eq.eqprg_count;
    equalisers[idx] = eq;
    memcpy(&cu.eq.eqprg[idx], eqprg, sizeof(struct vcu_eqprg));
    cu.eq.eqprg[idx].parts =
	malloc(eqprg->part_count * sizeof(struct vcu_eqpart));
    memcpy(cu.eq.eqprg[idx].parts, eqprg->parts,
	   eqprg->part_count * sizeof(struct vcu_eqpart));
    cu.eq.eqprg_count++;
    save_userconf = true;

    return 0;
}

static int
eq_modify_equaliser(int idx,
		    const struct vcu_eqprg *eqprg)
{
    int err;

    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (eqprg == NULL || eqprg->name[VCU_EQPRG_NAME_SIZE-1] != '\0' ||
        idx < 0 || idx >= cu.eq.eqprg_count)
    {
        return VCU_ERR_INVALID_VALUE;
    }
    if ((err = eq_sanitycheck(eqprg->parts, eqprg->part_count,
                              eq_programs)) != 0)
    {
	return err;
    }
    if (equalisers[idx]->bfir_id != -1) {
	bfir_delete_filters(equalisers[idx]->bfir_id);
    }
    memset(equalisers[idx], 0, sizeof(struct equaliser));
    free(cu.eq.eqprg[idx].parts);
    memcpy(&cu.eq.eqprg[idx], eqprg, sizeof(struct vcu_eqprg));
    cu.eq.eqprg[idx].parts =
	malloc(eqprg->part_count * sizeof(struct vcu_eqpart));
    memcpy(cu.eq.eqprg[idx].parts, eqprg->parts,
	   eqprg->part_count * sizeof(struct vcu_eqpart));
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    save_userconf = true;

    return 0;
}

static int
eq_delete_equaliser(int idx)
{
    int n, io, type, mode;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (idx < 0 || idx >= cu.eq.eqprg_count) {
	return VCU_ERR_INVALID_VALUE;
    }
    if (equalisers[idx]->bfir_id != -1) {
	bfir_delete_filters(equalisers[idx]->bfir_id);
    }
    free(equalisers[idx]);
    equalisers[idx] = NULL;
    for (mode = 0; mode < cu.eq.mode_count; mode++) {
        FOR_IN_AND_OUT {
            for (type = 0; type < VCU_CH_WP_COUNT; type++) {
                for (n = 0; n < cu.channel.count[io][type]; n++) {
                    if (cu.eq.mode[mode].prgs[io][type][n] == idx) {
                        cu.eq.mode[mode].prgs[io][type][n] = -1;
                    }
                    if (cu.eq.mode[mode].prgs[io][type][n] > idx) {
                        cu.eq.mode[mode].prgs[io][type][n] -= 1;
                    }
                }
            }
        }
    }
    for (mode = 0; mode < cu.channel.mode_count; mode++) {
        clear_invalid_eqmodes(&cu.channel.mode[mode]);
    }
    free(cu.eq.eqprg[idx].parts);
    cu.eq.eqprg_count--;
    for (n = idx; n < cu.eq.eqprg_count; n++) {
	cu.eq.eqprg[n] = cu.eq.eqprg[n + 1];
        equalisers[n] = equalisers[n + 1];
    }
    save_userconf = true;
    return 0;
}

static int
eq_get_equaliser_properties(int idx,
                            struct vcu_eqamp amp[VCU_FREQ_ISO_3OCT_BANDS],
                            double *peak_amp,
			    double *delay)
{
    struct prepfilter prepf;
    int err, sample_delay;
    struct equaliser *eq;
    
    if (idx < 0 || idx >= cu.eq.eqprg_count) {
        elog(LOGDEBUG, "non-existent equaliser %d", idx);
	return VCU_ERR_INVALID_VALUE;
    }
    eq = equalisers[idx];
    if (!eq->properties.updated) {
        if ((err = eq_generate(cu.eq.eqprg[idx].parts,
                               cu.eq.eqprg[idx].part_count,
                               eq_programs,
                               cu.reverb.block_sizes,
                               cu.signal.rates,
                               1,
                               cu.signal.double_precision ? sizeof(double) :
                               sizeof(float),
                               &sample_delay,
                               &prepf)) != 0)
        {
            elog(LOGDEBUG, "eq generate failed: %s", vcu.ctl.strerror(err));
            return err;
        }
	if ((err = eq_properties(&prepf,
                                 eq->properties.amp,
                                 &eq->properties.peak_amp)) != 0)
        {
	    filterprg_free_prepfilter_contents(&prepf);
            elog(LOGDEBUG, "eq properties failed: %s", vcu.ctl.strerror(err));
	    return err;
	}
	filterprg_free_prepfilter_contents(&prepf);
	eq->properties.delay = (double)sample_delay / (double)prepf.rates[0];
	eq->properties.updated = true;
    }
    if (amp != NULL) {
        memcpy(amp, eq->properties.amp, VCU_FREQ_ISO_3OCT_BANDS *
               sizeof(struct vcu_eqamp));
    }
    if (delay != NULL) {
        *delay = eq->properties.delay;
    }
    if (peak_amp != NULL) {
        *peak_amp = eq->properties.peak_amp;
    }
    return 0;
}

static double
eq_get_capacity(double *max_capacity)
{
    double cap, ret;   
    
    ret = bfir_filter_capacity(BFIR_FILTER_TYPE_EQ, &cap);
    if (max_capacity != NULL) {
        *max_capacity = cap;
    }
    return ret;
}

static const struct filterprg *
eq_get_programs(int *n_programs)
{
    static struct filterprg *fprgs = NULL, *filterprg;
    static int n_prgs = -1;
    int n, i;
    
    if (n_programs == NULL) {
	return NULL;
    }
    *n_programs = idarray_length(eq_programs);
    if (*n_programs != n_prgs) {
	fprgs = realloc(fprgs, (*n_programs) * sizeof(struct filterprg));
	n_prgs = *n_programs;
    }
    n = 0;
    i = -1;
    while ((filterprg = idarray_next(eq_programs, &i)) != NULL) {
	fprgs[n++] = *filterprg;
    }
    qsort(fprgs, *n_programs, sizeof(struct filterprg), fprg_compare);
    return fprgs;
}

static int
eq_load_program(const struct filterprg *program,
                bool_t copy_to_cache)
{
    struct filterprg *filterprg;
    struct prepfilter prepf;
    int err;

    /* FIXME: check if delay is too large */
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    MUST_NOT_NEED_TO_UPDATE_HARD_SETTINGS(VCU_ERR_PREREQUISITES_NOT_FULFILLED);    
    
    if (program == NULL) {
	return VCU_ERR_INVALID_VALUE;
    }
    switch (program->type) {
    case FILTERPRG_TYPE_EQUALISATION:
	break;
    default:
	return VCU_ERR_INVALID_VALUE;
    }
    if (program_in_use(NULL, program->id, program->type)) {
        return VCU_ERR_OBJECT_EXISTS;
    }
    
    /* just try to load it once to see if it is valid, use a native rate
       to avoid resampling */
    if ((err = filterprg_load(program, &prepf, NULL,
                              NULL, 0, program->n_channels,
                              program->rates,
			      1, sizeof(float), NULL, -1, -1)) != 0)
    {
	return err;
    }
    filterprg_free_prepfilter_contents(&prepf);
    filterprg = malloc(sizeof(struct filterprg));
    memset(filterprg, 0, sizeof(struct filterprg));
    memcpy(filterprg, program, sizeof(struct filterprg));
    sprintf(filterprg->path, "%seqr.%s.tar", conf_static.storage_path,
            filterprg_prgid_ntoa(filterprg->id));
    err = 0;
    if (strcmp(filterprg->path, program->path) != 0) {
        if (copy_to_cache || strlen(program->path) >= sizeof(filterprg->path)) {
            if ((err = copyfile_storage(filterprg->path, program->path)) != 0) {
                if (err == VCU_ERR_OBJECT_TOO_LARGE) {
                    elog(LOGDEBUG, "ignoring object too large error");
                } else {
                    free(filterprg);
                    return err;
                }
            }
        } else {
            strcpy(filterprg->path, program->path);
            err = add_fprog_ref(program->id, program->path);
        }
    }
    idarray_new_id(eq_programs, filterprg);
    save_userconf = true;
    return err;
}

static int
eq_delete_program(prgid_t id)
{
    struct filterprg *filterprg;
    char path[PATH_MAX];
    int n, i, j;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);

    n = -1;
    while ((filterprg = idarray_next(eq_programs, &n)) != NULL) {
	if (memcmp(&filterprg->id, &id, sizeof(id)) == 0) {
	    break;
	}
    }
    if (filterprg == NULL) {
	return VCU_ERR_INVALID_VALUE;
    }
    idarray_delete_id(eq_programs, n);
    free(filterprg);
    
    /* remove possible references to the program among the equalisers */
    for (n = 0; n < cu.eq.eqprg_count; n++) {
	for (i = 0; i < cu.eq.eqprg[n].part_count; i++) {
	    if (cu.eq.eqprg[n].parts[i].type == VCU_EQPRG_TYPE_CUSTOM &&
		memcmp(&cu.eq.eqprg[n].parts[i].params.custom.id, &id,
		       sizeof(id)) == 0)
	    {
		/* remove the element */
		for (j = i; j < cu.eq.eqprg[n].part_count - 1; j++) {
		    cu.eq.eqprg[n].parts[j] = cu.eq.eqprg[n].parts[j + 1];
		}
		cu.eq.eqprg[n].part_count -= 1;
	    }
	}
	if (cu.eq.eqprg[n].part_count == 0) {
	    /* remove whole equaliser */
	    eq_delete_equaliser(n);
	    n -= 1;
	}
    }
    sprintf(path, "%seqr.%s.tar", conf_static.storage_path,
	    filterprg_prgid_ntoa(id));
    clean_fprog_refs();
    remove(path);

    return 0;
}

static int
eq_get_mode_max_count(void)
{
    return VCU_MAX_EQ_MODES;
}

static int
eq_get_mode_count(void)
{
    return cu.eq.mode_count;
}

static const struct vcu_eqmode *
eq_get_mode(int idx)
{
    if (idx < 0 || idx >= cu.eq.mode_count) {
        return NULL;
    }    
    return &cu.eq.mode[idx];
}

static int
eq_new_mode(const struct vcu_eqmode *mode)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (cu.eq.mode_count >= VCU_MAX_EQ_MODES) {
        elog(LOGDEBUG, "mode table full");
	return VCU_ERR_NO_MEMORY;
    }
    if (!sanitycheck_eqmode(mode)) {
        elog(LOGDEBUG, "insane mode");
        return VCU_ERR_INVALID_VALUE;
    }
    cu.eq.mode[cu.eq.mode_count] = *mode;
    cu.eq.mode[cu.eq.mode_count].name[VCU_EQMODE_NAME_SIZE - 1] = '\0';
    cu.eq.mode_count++;
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    save_userconf = true;
    return 0;
}

static int
eq_modify_mode(int idx,
               const struct vcu_eqmode *mode)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (idx < 0 || idx >= cu.eq.mode_count || mode == NULL) {
        elog(LOGDEBUG, "invalid index (%d) or mode (%p)", idx, mode);
        return VCU_ERR_INVALID_VALUE;
    }
    if (memcmp(&cu.eq.mode[idx], mode, sizeof(struct vcu_eqmode)) == 0) {
        return 0;
    }
    if (!sanitycheck_eqmode(mode)) {
        elog(LOGDEBUG, "insane mode");
        return VCU_ERR_INVALID_VALUE;
    }
    cu.eq.mode[idx] = *mode;
    cu.eq.mode[idx].name[VCU_EQMODE_NAME_SIZE - 1] = '\0';
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    save_userconf = true;
    return 0;
}

static int
eq_delete_mode(int idx)
{
    int n, i;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (idx < 0 || idx >= cu.eq.mode_count) {
        elog(LOGDEBUG, "invalid index %d", idx);
        return VCU_ERR_INVALID_VALUE;
    }
    for (n = idx; n < cu.eq.mode_count - 1; n++){
        cu.eq.mode[n] = cu.eq.mode[n + 1];
    }
    memset(&cu.eq.mode[n], 0, sizeof(struct vcu_eqmode));
    if (cu.eq.active_mode > idx) {
        cu.eq.active_mode--;
    } else if (cu.eq.mode_count == 0) {
        cu.eq.active_mode = -1;
    }
    cu.eq.mode_count--;

    for (n = 0; n < cu.channel.mode_count; n++) {
        for (i = idx; i < cu.eq.mode_count - 1; i++) {
            if (bit_isset(cu.channel.mode[n].eqmodes, i + 1)) {
                bit_set(cu.channel.mode[n].eqmodes, i);
            } else {
                bit_clr(cu.channel.mode[n].eqmodes, i);
            }
        }
        bit_clr(cu.channel.mode[n].eqmodes, i);
        if (cu.channel.mode[n].active_eq_mode == idx) {
            cu.channel.mode[n].active_eq_mode =
                bit_find(cu.channel.mode[n].eqmodes, 0, cu.eq.mode_count);
        }
    }
    
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    save_userconf = true;
    return 0;
}

static int
eq_reorder_modes(const int order[])
{
    struct vcu_eqmode mode[VCU_MAX_EQ_MODES];
    struct vcu_chmode chmode;
    bool_t diff;
    int n, i;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (order == NULL) {
        return VCU_ERR_INVALID_VALUE;
    }
    diff = false;
    for (n = 0; n < cu.eq.mode_count; n++) {
        if (order[n] < 0 || order[n] >= cu.eq.mode_count) {
            return VCU_ERR_INVALID_VALUE;
        }
        if (order[n] != n) {
            diff = true;
        }        
    }
    if (!diff) {
        return 0;
    }
    memcpy(mode, cu.eq.mode, sizeof(cu.eq.mode));
    for (n = 0; n < cu.eq.mode_count; n++) {
        cu.eq.mode[n] = mode[order[n]];
    }
    if (cu.eq.active_mode >= 0) {
        cu.eq.active_mode = order[cu.eq.active_mode];
    }

    memset(&chmode, 0, sizeof(chmode));
    for (n = 0; n < cu.channel.mode_count; n++) {
        for (i = 0; i < cu.eq.mode_count; i++) {
            if (bit_isset(cu.channel.mode[n].eqmodes, i)) {
                bit_set(chmode.eqmodes, order[n]);
            }
        }
        memcpy(cu.channel.mode[n].eqmodes, chmode.eqmodes,
               sizeof(chmode.eqmodes));
        if (cu.channel.mode[n].active_eq_mode >= 0) {
            cu.channel.mode[n].active_eq_mode =
                order[cu.channel.mode[n].active_eq_mode];
        }
    }
    
    save_userconf = true;
    return 0;
}

static bool_t
eq_mode_is_assignable(int eq_mode_idx,
                      int channel_mode_idx)
{
    if (eq_mode_idx < 0 || eq_mode_idx >= cu.eq.mode_count ||
        channel_mode_idx < 0 || channel_mode_idx >= cu.channel.mode_count)
    {
        return false;
    }
    return eqmode_valid_in_chmode(&cu.eq.mode[eq_mode_idx],
                                  &cu.channel.mode[channel_mode_idx]);
}

static bool_t
iodelay_need_to_run_benchmark(void)
{
    bool_t ans;
    
    if (has_run_benchmark) {
	/* see if we have modified settings so we need to run a benchmark
	   again */
	ans = benchmark_dependent_setting_modified;
        elog(LOGDEBUG, "need to run benchmark: %s", ans ? "true" : "false");
        return ans;
    }

    /* we have not run benchmark, see if we have have modified settings since
       startup, if so, we need to do a new benchmark */
    if (benchmark_dependent_setting_modified) {
        elog(LOGDEBUG, "need to run benchmark: true");
	return true;
    }

    /* we have not modified settings since startup, then see if we have
       other hardware than the one stored in the configuration, if so,
       we need to do a new benchmark. If no benchmark has ever been run the
       saved hardware config will be clear (0). */
    if (memcmp(&current_sysinfo, &cu.reverb.benchmark_sysinfo,
	       sizeof(struct basic_sysinfo)) != 0)
    {
        elog(LOGDEBUG, "need to run benchmark: true");
	return true;
    }
    elog(LOGDEBUG, "need to run benchmark: false");
    return false;
}

static bool_t
iodelay_has_run_benchmark(void)
{
    return has_run_benchmark;
}

static bool_t
iodelay_manually_set(void)
{
    return cu.reverb.iodelay_manually_set;
}

static int
iodelay_run_benchmark(bool_t test_low_latency,
                      int *limited_by)
{
    struct bfir_chmap *maps[2];
    int n, res, type;

    if (limited_by != NULL) {
        *limited_by = 0;
    }
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);

    /* verify that we have some channels */
    for (type = 0; type < VCU_CH_COUNT && cu.channel.count[OUT][type] == 0;
         type++);
    if (type == VCU_CH_COUNT) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    make_bfir_chmap(maps);
    memset(&benchmark, 0, sizeof(benchmark));
    for (n = 0; n < cu.signal.n_rates; n++) {
	benchmark.current_rate = cu.signal.rates[n];
	res = bfir_benchmark(cu.signal.rates[n],
                             test_low_latency,
                             cu.signal.apply_dither,
			     cu.ambisonics.input_format,
                             (const struct bfir_chmap **)maps,
			     cu.reverb.max_benchmark_load,
			     cu.reverb.max_benchmark_length,
			     &benchmark.n_alternatives[n],
			     benchmark.block_sizes[n],
			     benchmark.n_blocks[n],
                             benchmark_filter_lengths,
			     benchmark_progress_callback);
        has_newly_run_benchmark = true;
	switch (res) {
	case BFIR_BENCHMARK_STATUS_OUT_OF_RANGE:
	    memset(&benchmark, 0, sizeof(benchmark));
	    return VCU_ERR_NOT_SUPPORTED_BY_HARDWARE;
	case BFIR_BENCHMARK_STATUS_MEMORY_LIMITED:
            if (limited_by != NULL &&
                *limited_by != VCU_BENCHMARK_LIMIT_CPU)
            {
                *limited_by = VCU_BENCHMARK_LIMIT_MEMORY;
            }
	    break;
	case BFIR_BENCHMARK_STATUS_PROCESSOR_LIMITED:
            if (limited_by != NULL) {
                *limited_by = VCU_BENCHMARK_LIMIT_CPU;
            }
	    break;
	case BFIR_BENCHMARK_STATUS_HARD_LIMITED:
            if (limited_by != NULL &&
                *limited_by != VCU_BENCHMARK_LIMIT_CPU)
            {
                *limited_by = VCU_BENCHMARK_LIMIT_HARD;
            }
	    break;
	default:
	    memset(&benchmark, 0, sizeof(benchmark));
	    return VCU_ERR_UNKNOWN_ERROR;
	}
    }
    has_run_benchmark = true;
    benchmark_dependent_setting_modified = false;
    /* immediately apply a valid setting */
    for (n = 0; n < cu.signal.n_rates; n++) {
        vcu.iodelay.set_delay_length_pair(cu.signal.rates[n],
                                          benchmark.n_alternatives[n] - 1);
    }
    save_userconf = true;
    return 0;
}

static double
iodelay_get_benchmark_max_load(void)
{
    return cu.reverb.max_benchmark_load;
}

static int
iodelay_set_benchmark_max_load(double load)
{
    if (load <= 0 || load > 1.0) {
        return VCU_ERR_INVALID_VALUE;
    }
    if (load == cu.reverb.max_benchmark_load) {
        return 0;
    }
    cu.reverb.max_benchmark_load = load;
    save_userconf = true;
    return 0;
}

static void
iodelay_get_benchmark_max_length_range(double range[2])
{
    if (range == NULL) {
        return;
    }
    range[0] = VCU_BENCHMARK_MIN_SECONDS;
    range[1] = VCU_BENCHMARK_MAX_SECONDS;
}

static double
iodelay_get_benchmark_max_length(void)
{
    return cu.reverb.max_benchmark_length;
}

static int
iodelay_set_benchmark_max_length(double length)
{
    double range[2];

    iodelay_get_benchmark_max_length_range(range);
    if (length < range[0] || length > range[1]) {
        return VCU_ERR_INVALID_VALUE;
    }
    if (length == cu.reverb.max_benchmark_length) {
        return 0;
    }
    cu.reverb.max_benchmark_length = length;
    save_userconf = true;
    return 0;
}

static int
iodelay_get_dlp_alternatives(int sample_rate,
			     double delay[],
			     double max_reverb_length[],
			     int *n_alternatives)
{
    int n, ri;

    MUST_NOT_NEED_TO_RUN_BENCHMARK(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (!has_run_benchmark) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    if (delay == NULL || max_reverb_length == NULL || n_alternatives == NULL) {
	return VCU_ERR_INVALID_VALUE;
    }
    for (ri = 0; ri < cu.signal.n_rates; ri++) {
	if (sample_rate == cu.signal.rates[ri]) {
	    break;
	}
    }
    if (ri == cu.signal.n_rates) {
	return VCU_ERR_INVALID_VALUE;
    }    
    for (n = 0; n < benchmark.n_alternatives[ri]; n++) {
	delay[n] = 2.0 * (double)benchmark.block_sizes[ri][n] /
	    (double)sample_rate;
	max_reverb_length[n] = (double)benchmark.n_blocks[ri][n] *
	    (double)benchmark.block_sizes[ri][n] / (double)sample_rate;
        elog(LOGDEBUG, "alt %d: %f, %f", n, delay[n], max_reverb_length[n]);
    }
    *n_alternatives = benchmark.n_alternatives[ri];
    return 0;
}

static void
iodelay_get_delay_length_pair(int sample_rate,
			      double *delay,
			      double *max_reverb_length,
			      int *benchmark_index)
{
    int n, ri = -1;

    for (n = 0; n < cu.signal.n_rates; n++) {
	if (sample_rate == cu.signal.rates[n]) {
	    ri = n;
	    break;
	}
    }
    if (n == cu.signal.n_rates) {
	return;
    }
    if (delay != NULL) {
	*delay = 2.0 * (double)cu.reverb.block_sizes[ri] /
	    (double)sample_rate;
    }
    if (max_reverb_length != NULL) {
	*max_reverb_length = (double)cu.reverb.n_blocks[ri] *
	    (double)cu.reverb.block_sizes[ri] / (double)sample_rate;
    }
    if (benchmark_index != NULL) {
	*benchmark_index = -1;
	for (n = 0;
             n < benchmark.n_alternatives[ri] &&
                 !cu.reverb.iodelay_manually_set;
             n++)
        {
	    if (benchmark.block_sizes[ri][n] ==
		cu.reverb.block_sizes[ri] &&
		benchmark.n_blocks[ri][n] == cu.reverb.n_blocks[ri])
	    {
                elog(LOGDEBUG, "index is %d (%d / %d)", n,
                     benchmark.block_sizes[ri][n],
                     benchmark.n_blocks[ri][n]);
		*benchmark_index = n;
		return;
	    }
	}
    }
}

static int
iodelay_set_delay_length_pair(int sample_rate,
			      int idx)
{
    int n, ri = -1;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    MUST_NOT_NEED_TO_RUN_BENCHMARK(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    for (n = 0; n < cu.signal.n_rates; n++) {
	if (sample_rate == cu.signal.rates[n]) {
	    ri = n;
	    break;
	}
    }
    if (n == cu.signal.n_rates ||
	idx < 0 || idx >= benchmark.n_alternatives[ri])
    {
	elog(LOGDEBUG, "invalid alternative");
	return VCU_ERR_INVALID_VALUE;
    }
    
    if (benchmark.n_blocks[ri][idx] == cu.reverb.n_blocks[ri] &&
	benchmark.block_sizes[ri][idx] == cu.reverb.block_sizes[ri] &&        
	memcmp(&cu.reverb.benchmark_sysinfo, &current_sysinfo,
	       sizeof(struct basic_sysinfo)) == 0 &&
        !cu.reverb.iodelay_manually_set)
    {
	elog(LOGDEBUG, "no change");
	return 0;
    }
    memcpy(&cu.reverb.benchmark_sysinfo, &current_sysinfo,
	   sizeof(struct basic_sysinfo));
    cu.reverb.block_sizes[ri] = benchmark.block_sizes[ri][idx];
    cu.reverb.n_blocks[ri] = benchmark.n_blocks[ri][idx];
    cu.reverb.iodelay_manually_set = false;
    save_userconf = true;
    elog(LOGDEBUG, "set dlp idx %d (%d / %d)", idx,
         benchmark.block_sizes[ri][idx],
         benchmark.n_blocks[ri][idx]);
    return 0;
}

static int
iodelay_set_delay_length_pair_manual(int sample_rate,
                                     double delay,
                                     double max_reverb_length)
{
    int ri, rm, block_size, n_blocks;

    for (ri = 0; ri < cu.signal.n_rates; ri++) {
	if (sample_rate == cu.signal.rates[ri]) {
	    break;
	}
    }
    if (ri == cu.signal.n_rates) {
	return VCU_ERR_INVALID_VALUE;
    }
    rm = sample_rate / 48010 + 1;
    block_size =
        1 << log2_roof((uint32_t)(delay * (double)sample_rate * 0.5 * 0.95));
    elog(LOGDEBUG, "%d %d %f",
         (int)(delay * (double)sample_rate * 0.5 * 0.95),
         block_size,
         delay);
    if (block_size < BFIR_MIN_BLOCKSIZE_LOWLAT) {
        block_size = BFIR_MIN_BLOCKSIZE_LOWLAT;
    } else if (block_size > rm * BFIR_MAX_BLOCKSIZE_48000HZ) {
        block_size = rm * BFIR_MAX_BLOCKSIZE_48000HZ;
    }
    n_blocks = (int)(0.995 * max_reverb_length * (double)sample_rate /
                     (double)block_size) + 1;
    elog(LOGDEBUG, "%d %f",
         (int)(0.9999 * max_reverb_length * (double)sample_rate /
               (double)block_size) + 1,
         max_reverb_length);
    if (n_blocks < BFIR_MIN_BLOCKS) {
        n_blocks = BFIR_MIN_BLOCKS;
    } else if (n_blocks > BFIR_MAX_BLOCKS) {
        n_blocks = BFIR_MAX_BLOCKS;
    }
    if (cu.reverb.block_sizes[ri] == block_size &&
        cu.reverb.n_blocks[ri] == n_blocks &&
        cu.reverb.iodelay_manually_set)
    {
        return 0;
    }
    cu.reverb.block_sizes[ri] = block_size;
    cu.reverb.n_blocks[ri] = n_blocks;
    cu.reverb.iodelay_manually_set = true;
    elog(LOGDEBUG, "set %d / %d for rate %d",
         block_size, n_blocks, sample_rate);
    benchmark_dependent_setting_modified = false;
    save_userconf = true;
    return 0;
}

static double
iodelay_get_alignment_delay(int sample_rate)
{
    int ri;

    for (ri = 0; ri < cu.signal.n_rates; ri++) {
	if (sample_rate == cu.signal.rates[ri]) {
	    break;
	}
    }
    if (ri == cu.signal.n_rates) {
	return -1.0;
    }    
    if (!runtime_isruntime()) {
        if (calculate_runtime_delay_map(ri) < 0) {
            return -1.0;
        }
    }
    return (double)runtime_delay_map.flowthrough_delay / (double)sample_rate;
}

static bool_t
iodelay_has_truncated_filters(void)
{
    struct revprg *revprg;
    int n;

    if (rev_programs == NULL) {
        return false;
    }
    n = -1;
    while ((revprg = idarray_next(rev_programs, &n)) != NULL) {
        if (revprg->truncated) {
            return true;
        }
    }
    return false;
}

static bool_t
ambiopole_get_xtc_active(int idx)
{
    if (idx < 0 || idx >= cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2) {
        return false;
    }
    return cu.ambiopole.use_xtalk_cancel[idx];
}

static int
ambiopole_set_xtc_active(int idx,
                         bool_t activate)
{
    struct xtcprg *xtcprg;
    int n;

    elog(LOGDEBUG, "idx %d; activate %d", idx, activate);
    activate = !!activate;
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (activate && idarray_length(xtc_programs) == 0) {
	return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
    }
    
    if (idx < 0 || idx >= cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2) {
        return VCU_ERR_INVALID_VALUE;
    }
    if (activate != cu.ambiopole.use_xtalk_cancel[idx]) {
	cu.ambiopole.use_xtalk_cancel[idx] = activate;
	update_bfir_soft_settings();
        elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
        benchmark_dependent_setting_modified = true;
	save_userconf = true;
    }
    if (activate && get_xtcprg(cu.ambiopole.current[idx]) == NULL) {
        elog(LOGDEBUG, "force set xtc program");
        n = -1;
        xtcprg = idarray_next(xtc_programs, &n);
        cu.ambiopole.current[idx] = xtcprg->prg.id;
	update_bfir_soft_settings();
        elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
        benchmark_dependent_setting_modified = true;
	save_userconf = true;
    }
    return 0;
}

static const struct filterprg *
ambiopole_get_programs(int *n_programs)
{
    static struct filterprg *fprgs = NULL;
    static int n_prgs = -1;
    struct xtcprg *xtcprg;
    int id, n;
    
    if (n_programs == NULL) {
	return NULL;
    }
    *n_programs = idarray_length(xtc_programs);
    if (*n_programs != n_prgs) {
	fprgs = realloc(fprgs, (*n_programs) * sizeof(struct filterprg));
	n_prgs = *n_programs;
    }
    n = 0;
    id = -1;
    while ((xtcprg = idarray_next(xtc_programs, &id)) != NULL) {
	fprgs[n++] = xtcprg->prg;
    }
    qsort(fprgs, *n_programs, sizeof(struct filterprg), fprg_compare);
    return fprgs;
}

static double
ambiopole_get_capacity(double *max_capacity)
{
    double cap, ret;    
    
    ret = bfir_filter_capacity(BFIR_FILTER_TYPE_XTC, &cap);
    if (max_capacity != NULL) {
        *max_capacity = cap;
    }
    return ret;
}

static int
ambiopole_load_program(const struct filterprg *program,
                       bool_t copy_to_cache)
{
    struct xtcprg *xtcprg;
    int err, n;

    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    MUST_NOT_NEED_TO_UPDATE_HARD_SETTINGS(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    if (program == NULL) {
	return VCU_ERR_INVALID_VALUE;
    }
    switch (program->type) {
    case FILTERPRG_TYPE_CROSSTALK_CANCELLATION:
	break;
    default:
	return VCU_ERR_INVALID_VALUE;
    }
    if (program_in_use(NULL, program->id, program->type)) {
        return VCU_ERR_OBJECT_EXISTS;
    }

    xtcprg = malloc(sizeof(struct xtcprg));
    memset(xtcprg, 0, sizeof(struct xtcprg));
    if ((xtcprg->bfir_id = load_xtcprg(program,
				       xtcprg->bfirfilters,
				       xtcprg->sample_delay)) < 0)
    {
        elog(LOGDEBUG, "load xtcprg failed");
	err = xtcprg->bfir_id;
	free(xtcprg);
	return err;
    }
    xtcprg->prg = *program;
    sprintf(xtcprg->prg.path, "%sxtc.%s.tar", conf_static.storage_path,
            filterprg_prgid_ntoa(xtcprg->prg.id));
    err = 0;
    if (strcmp(xtcprg->prg.path, program->path) != 0) {
        if (copy_to_cache || strlen(program->path) > sizeof(xtcprg->prg.path)) {
            if ((err = copyfile_storage(xtcprg->prg.path,
                                        program->path)) != 0)
            {
                if (err == VCU_ERR_OBJECT_TOO_LARGE) {
                    elog(LOGDEBUG, "ignoring object too large error");
                } else {
                    bfir_delete_filters(xtcprg->bfir_id);
                    free(xtcprg->bfirfilters);
                    free(xtcprg);
                    return err;
                }
            }
        } else {
            strcpy(xtcprg->prg.path, program->path);
            err = add_fprog_ref(program->id, program->path);
        }
    }
    for (n = 0; n < cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2; n++) {
        if (memcmp(&cu.ambiopole.current[n], &zeroid, sizeof(zeroid)) == 0) {
            cu.ambiopole.current[n] = xtcprg->prg.id;
        }
    }
    idarray_new_id(xtc_programs, xtcprg);
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    save_userconf = true;
    return err;
}

static int
ambiopole_delete_program(prgid_t id)
{
    struct xtcprg *xtcprg;
    int n, i;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    n = -1;
    while ((xtcprg = idarray_next(xtc_programs, &n)) != NULL) {
	if (memcmp(&xtcprg->prg.id, &id, sizeof(id)) == 0) {
	    idarray_delete_id(xtc_programs, n);
	    break;
	}
    }
    if (xtcprg == NULL) {
	return VCU_ERR_INVALID_VALUE;
    }
    bfir_delete_filters(xtcprg->bfir_id);
    free(xtcprg);
    if (idarray_length(xtc_programs) == 0) {
        for (n = 0; n < cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2; n++) {
            if (cu.ambiopole.use_xtalk_cancel[n]) {
                vcu.ambiopole.set_xtc_active(n, false);
            }
        }
    }
    for (n = 0; n < cu.channel.count[OUT][VCU_CH_AMBIOPOLE] / 2; n++) {
        if (memcmp(&cu.ambiopole.current[n], &id, sizeof(id)) == 0) {
            i = -1;
            if ((xtcprg = idarray_next(xtc_programs, &i)) == NULL) {
                cu.ambiopole.current[n] = zeroid;
            } else {
                cu.ambiopole.current[n] = xtcprg->prg.id;
            }
        }
    }
    clean_fprog_refs();
    save_userconf = true;
    return 0;
}

static double
reverb_get_capacity(double *max_capacity)
{
    double ret, cap;
    
    ret = bfir_filter_capacity(BFIR_FILTER_TYPE_REVERB, &cap);
    if (max_capacity != NULL) {
        *max_capacity = cap;
    }
    return ret;
}

static const struct filterprg *
reverb_get_programs(int *n_programs)
{
    static struct filterprg *fprgs = NULL;
    static int n_prgs = -1;
    struct revprg *revprg;
    int id, n;
    
    if (n_programs == NULL) {
	return NULL;
    }
    *n_programs = idarray_length(rev_programs);
    if (*n_programs != n_prgs) {
	fprgs = realloc(fprgs, (*n_programs) * sizeof(struct filterprg));
	n_prgs = *n_programs;
    }
    n = 0;
    id = -1;
    while ((revprg = idarray_next(rev_programs, &id)) != NULL) {
	fprgs[n++] = revprg->prg;
    }
    qsort(fprgs, *n_programs, sizeof(struct filterprg), fprg_compare);
    return fprgs;
}

static int
reverb_load_program(const struct filterprg *program,
                    bool_t copy_to_cache)
{
    struct revprg *revprg;
    int err;

    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    MUST_NOT_NEED_TO_UPDATE_HARD_SETTINGS(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    if (program == NULL) {
	elog(LOGDEBUG, "program is NULL");
	return VCU_ERR_INVALID_VALUE;
    }
    switch (program->type) {
    case FILTERPRG_TYPE_REVERB_TRADITIONAL:
    case FILTERPRG_TYPE_REVERB_BFORMAT:
	break;
    default:
	elog(LOGDEBUG, "unexpected program type: %d", program->type);
	return VCU_ERR_INVALID_VALUE;
    }
    if (program_in_use(NULL, program->id, program->type)) {
        return VCU_ERR_OBJECT_EXISTS;
    }
    
    elog(LOGDEBUG, "loading program: %s", program->path);

    revprg = malloc(sizeof(struct revprg));
    memset(revprg, 0, sizeof(struct revprg));
    revprg->bfirfilters = malloc(cu.channel.count[IN][VCU_CH_REVERB] *
                                 cu.channel.count[OUT][VCU_CH_REVERB] *
				 sizeof(struct bfirfilter));
    revprg->prg = *program;
    conf_load_revprg(&revprg->conf, &revprg->prg);
    if ((revprg->bfir_id = load_revprg(program,
                                       &revprg->conf,
                                       revprg->bfirfilters,
                                       &revprg->orig_rt60,
                                       &revprg->orig_tail_volume,
                                       &revprg->proc_tail_volume,
                                       &revprg->proc_length,
                                       &revprg->proc_rt60,
                                       &revprg->truncated)) < 0)
    {
	err = revprg->bfir_id;
	free(revprg->bfirfilters);
	free(revprg);
	return err;
    }
    conf_save_revprg(&revprg->conf);
    sprintf(revprg->prg.path, "%srev.%s.tar", conf_static.storage_path,
	    filterprg_prgid_ntoa(revprg->prg.id));
    err = 0;
    if (strcmp(revprg->prg.path, program->path) != 0) {
        if (copy_to_cache || strlen(program->path) > sizeof(revprg->prg.path)) {
            if ((err = copyfile_storage(revprg->prg.path,
                                        program->path)) != 0)
            {
                if (err == VCU_ERR_OBJECT_TOO_LARGE) {
                    elog(LOGDEBUG, "ignoring object too large error");
                } else {
                    bfir_delete_filters(revprg->bfir_id);
                    free(revprg->bfirfilters);
                    free(revprg);
                    return err;
                }
            }
        } else {
            strcpy(revprg->prg.path, program->path);
            err = add_fprog_ref(program->id, program->path);
        }
    }
    if (memcmp(&cu.reverb.current, &zeroid, sizeof(zeroid)) == 0) {
	cu.reverb.current = revprg->prg.id;
    }
    idarray_new_id(rev_programs, revprg);
    save_userconf = true;
    return err;
}

static int
reverb_copy_program(const struct filterprg *program,
                    struct filterprg *destination)
{    
    char path[PATH_MAX + 100], name[FILTERPRG_NAME_MAX];
    char longname[FILTERPRG_LONGNAME_MAX], tmp[100];
    const struct filterprg *fprgs;
    int err, n, n_programs, i;
    uint64_t bytes_left;    
    struct stat buf;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (program == NULL) {
        return VCU_ERR_INVALID_VALUE;
    }
    if (stat(program->path, &buf) == -1) {
        elog(LOGDEBUG, "failed to stat file \"%s\": %s", program->path,
             strerror(errno));
        return VCU_ERR_READ_FAILED;
    }
    bytes_left = vcu.cache.get_capacity(NULL);
    if (bytes_left < buf.st_size + 1024) {
        elog(LOGDEBUG, "copy does not fit in storage");
        return VCU_ERR_OBJECT_TOO_LARGE;
    }       
    if ((fprgs = vcu.cache.get_programs(&n_programs, VCU_PROGRAM_TYPE_REVERB))
        == NULL)
    {
        return VCU_ERR_READ_FAILED;
    }
    i = 1;
    do {
        sprintf(tmp, " (%d)", i);        
        strcpy(name, program->name);
        strcpy(longname, program->longname);
        if (strlen(name) + strlen(tmp) > FILTERPRG_NAME_MAX - 1) {
            strcpy(&name[FILTERPRG_NAME_MAX - strlen(tmp) - 1], tmp);
        } else {
            strcat(name, tmp);
        }
        if (strlen(longname) + strlen(tmp) > FILTERPRG_LONGNAME_MAX - 1) {
            strcpy(&longname[FILTERPRG_LONGNAME_MAX - strlen(tmp) - 1], tmp);
        } else {
            strcat(longname, tmp);
        }
        for (n = 0; n < n_programs; n++) {
            if (strcmp(fprgs[n].name, name) == 0 ||
                strcmp(fprgs[n].longname, longname) == 0)
            {
                break;
            }
        }
        i++;
    } while (n != n_programs);
    
    sprintf(path, "%srev.%%s.tar", conf_static.storage_path);
    if ((err = filterprg_copy(program, name, longname, path)) != 0) {
        elog(LOGDEBUG, "filterprg_copy failed: %s", vcu.ctl.strerror(err));
        return err;
    }

    if (destination != NULL) {
        if ((fprgs = vcu.cache.get_programs(&n_programs,
                                            VCU_PROGRAM_TYPE_REVERB)) == NULL)
        {
            return VCU_ERR_READ_FAILED;
        }
        for (n = 0; n < n_programs; n++) {
            if (strcmp(fprgs[n].name, name) == 0 ||
                strcmp(fprgs[n].longname, longname) == 0)
            {
                *destination = fprgs[n];
                break;
            }
        }
        if (n == n_programs) {
            return VCU_ERR_READ_FAILED;
        }
    }
        
    return 0;
}

static int
reverb_delete_program(prgid_t id)
{
    struct revprg *revprg;
    int n;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    n = -1;
    while ((revprg = idarray_next(rev_programs, &n)) != NULL) {
	if (memcmp(&revprg->prg.id, &id, sizeof(id)) == 0) {
	    idarray_delete_id(rev_programs, n);
	    break;
	}
    }
    if (revprg == NULL) {
	return VCU_ERR_INVALID_VALUE;
    }
    bfir_delete_filters(revprg->bfir_id);
    free(revprg->bfirfilters);
    free(revprg);
    if (memcmp(&cu.reverb.current, &id, sizeof(id)) == 0) {
	n = -1;
	if ((revprg = idarray_next(rev_programs, &n)) == NULL) {
	    cu.reverb.current = zeroid;
	} else {
	    cu.reverb.current = revprg->prg.id;
	}
    }
    clean_fprog_refs();
    save_userconf = true;    
    return 0;
}

static int
reverb_get_response_mode(void)
{
    return cu.reverb.response_mode;
}

static int
reverb_set_response_mode(int mode)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    if (mode == cu.reverb.response_mode) {
        return 0;
    }
    switch (mode) {
    case VCU_REVERB_RESP_MODE_FULL:
    case VCU_REVERB_RESP_MODE_MUTE_DIRECT:
    case VCU_REVERB_RESP_MODE_MUTE_EARLY:
    case VCU_REVERB_RESP_MODE_STAGE_ADAPTED:
        cu.reverb.response_mode = mode;
        break;
    default:
        return VCU_ERR_INVALID_VALUE;
    }
    save_userconf = true;
    return 0;
}

static int
reverb_get_decorrelate_mode(void)
{
    return cu.reverb.decorrelate_mode;
}

static int
reverb_set_decorrelate_mode(int mode)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    if (mode == cu.reverb.decorrelate_mode) {
        return 0;
    }
    switch (mode) {
    case VCU_REVERB_DECORR_MODE_NONE:
    case VCU_REVERB_DECORR_MODE_ALL:
    case VCU_REVERB_DECORR_MODE_NEEDED:
        cu.reverb.decorrelate_mode = mode;
        break;
    default:
        return VCU_ERR_INVALID_VALUE;
    }
    save_userconf = true;
    return 0;
}

static bool_t
reverb_get_force_delay_diff(void)
{
    return cu.reverb.force_delay_diff;
}

static int
reverb_set_force_delay_diff(bool_t activate)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    activate = !!activate;
    if (!activate == !cu.reverb.force_delay_diff) {
        return 0;
    }
    cu.reverb.force_delay_diff = activate;
    save_userconf = true;
    return 0;
}

static bool_t
reverb_get_force_decay(void)
{
    return cu.reverb.force_decay;
}

static int
reverb_set_force_decay(bool_t activate)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    activate = !!activate;
    if (!activate == !cu.reverb.force_decay) {
        return 0;
    }
    cu.reverb.force_decay = activate;
    save_userconf = true;
    return 0;
}

static bool_t
reverb_get_dry_active(void)
{
    return cu.reverb.mix_with_dry;
}

static int
reverb_set_dry_active(bool_t activate)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    activate = !!activate;
    if (!activate == !cu.reverb.mix_with_dry) {
        return 0;
    }
    cu.reverb.mix_with_dry = activate;
    save_userconf = true;
    return 0;
}

static void
reverb_get_dry_volume_range(double range[2])
{
    if (range == NULL) {
	return;
    }
    range[0] = VCU_VOLUME_MIN;
    range[1] = VCU_VOLUME_MAX;
}

static double
reverb_get_dry_volume(void)
{
    return cu.reverb.dry_volume;
}

static int
reverb_set_dry_volume(double volume)
{
    double range[2];
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    runtime_volume_get_range(range);
    if (volume < range[0]) {
	volume = range[0];
    }
    if (volume > range[1]) {
	volume = range[1];
    }
    if (volume == cu.reverb.dry_volume) {
        return 0;
    }
    cu.reverb.dry_volume = volume;
    save_userconf = true;
    return 0;
}

static void
reverb_get_bformat_directivity_range(double range[2])
{
    if (range == NULL) {
        return;
    }
    range[0] = REVERB_BFORMAT_DIRECTIVITY_MIN;
    range[1] = REVERB_BFORMAT_DIRECTIVITY_MAX;
}

static double
reverb_get_bformat_directivity(void)
{
    return cu.reverb.bformat_directivity;
}

static int
reverb_set_bformat_directivity(double directivity)
{
    double range[2];
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    reverb_get_bformat_directivity_range(range);
    if (directivity < range[0]) {
        directivity = range[0];
    }
    if (directivity > range[1]) {
        directivity = range[1];
    }
    if (directivity == cu.reverb.bformat_directivity) {
        return 0;
    }
    save_userconf = true;
    cu.reverb.bformat_directivity = directivity;
    return 0;
}

static int
reverb_get_source_layout(void)
{
    return cu.reverb.source_layout;
}

static int
reverb_set_source_layout(int layout)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (layout == cu.reverb.source_layout) {
        return 0;
    }
    
    switch (layout) {
    case VCU_REVERB_SOURCE_LAYOUT_CENTER:
        cu.channel.count[IN][VCU_CH_REVERB] = 1;
        break;
    case VCU_REVERB_SOURCE_LAYOUT_STEREO:
        cu.channel.count[IN][VCU_CH_REVERB] = 2;
        break;
    case VCU_REVERB_SOURCE_LAYOUT_LRC:
        cu.channel.count[IN][VCU_CH_REVERB] = 3;
        break;
    default:
        return VCU_ERR_INVALID_VALUE;
    }

    cu.reverb.source_layout = layout;
    channel_set_count(VCU_CH_REVERB, cu.channel.count[OUT][VCU_CH_REVERB]);
    elog(LOGDEBUG, "benchmark_dependent_setting_modified = true;");
    benchmark_dependent_setting_modified = true;
    save_userconf = true;
    return 0;
}

static int
ambisonics_get_input_format(void)
{
    return cu.ambisonics.input_format;
}

static int
ambisonics_set_input_format(int format)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    if (format == cu.ambisonics.input_format) {
        return 0;
    }
    
    switch (format) {
    case VCU_AMBISONIC_INPUT_FORMAT_FIRST_ORDER:
        cu.channel.count[IN][VCU_CH_AMBISONIC] = 4;
        break;
    case VCU_AMBISONIC_INPUT_FORMAT_SECOND_ORDER:
        cu.channel.count[IN][VCU_CH_AMBISONIC] = 9;
        break;
    case VCU_AMBISONIC_INPUT_FORMAT_UHJ:
        //cu.channel.count[IN][type] = 2;
        return VCU_ERR_NOT_IMPLEMENTED;
    default:
        return VCU_ERR_INVALID_VALUE;
    }

    cu.ambisonics.input_format = format;
    channel_set_count(VCU_CH_AMBISONIC,
                      cu.channel.count[OUT][VCU_CH_AMBISONIC]);
    save_userconf = true;
    return 0;
}

static uint64_t
cache_get_capacity(uint64_t *max_capacity)
{
    struct dirent **namelist;
    int n;
    
    if (max_capacity != NULL) {
        *max_capacity = (uint64_t)conf_static.storage_size * 1024 * 1024;
    }    
    copyfile_size = 0;
    copyfile_dest_path = "";
    if ((n = scandir(conf_static.storage_path, &namelist,
                     copyfile_scandir_select, NULL)) < 0)
    {
        elog(LOGDEBUG, "scandir failed: %s", strerror(errno));
        return 0;
    }
    while (n--) {
        free(namelist[n]);
    }
    free(namelist);

    return (uint64_t)conf_static.storage_size * 1024 * 1024 - copyfile_size;
}

static const struct filterprg *
cache_get_programs(int *n_programs,
		   int type)
{
    static struct filterprg *fprgs = NULL;
    static int capacity = 0;
    char fmt[PATH_MAX + 100], dummy[100];
    struct filterprg *_fprgs;
    int _n_programs, n;
    bool_t add;

    if (n_programs == NULL) {
	return NULL;
    }
    if ((_fprgs = filterprg_scan_directory(conf_static.storage_path,
					   &_n_programs)) == NULL)
    {
	*n_programs = 0;
	return NULL;
    }
    if (_n_programs > capacity) {
	fprgs = realloc(fprgs, _n_programs * sizeof(struct filterprg));
	capacity = _n_programs;
    }
    for (n = (*n_programs) = 0; n < _n_programs; n++) {
	add = false;
	switch (_fprgs[n].type) {
	case FILTERPRG_TYPE_EQUALISATION:
	    if (type != VCU_PROGRAM_TYPE_EQ) {
		break;
	    }
	    /* we just check that the name fits the format */
	    sprintf(fmt, "%seqr.%%32s.tar", conf_static.storage_path);
	    if (sscanf(_fprgs[n].path, fmt, dummy) != 1) {
		break;
	    }
	    add = true;
	    break;
	case FILTERPRG_TYPE_REVERB_TRADITIONAL:
	case FILTERPRG_TYPE_REVERB_BFORMAT:
	    if (type != VCU_PROGRAM_TYPE_REVERB) {
		break;
	    }
	    sprintf(fmt, "%srev.%%32s.tar", conf_static.storage_path);
	    if (sscanf(_fprgs[n].path, fmt, dummy) != 1) {
		break;
	    }
	    add = true;
	    break;
	case FILTERPRG_TYPE_CROSSTALK_CANCELLATION:
	    if (type != VCU_PROGRAM_TYPE_XTC) {
		break;
	    }
	    sprintf(fmt, "%sxtc.%%32s.tar", conf_static.storage_path);
	    if (sscanf(_fprgs[n].path, fmt, dummy) != 1) {
		break;
	    }
	    add = true;
	    break;
	default:
	    break;
	}
	if (add) {
	    memcpy(&fprgs[*n_programs], &_fprgs[n], sizeof(struct filterprg));
	    (*n_programs)++;
	}
    }
    if (*n_programs == 0) {
	return NULL;
    }
    qsort(fprgs, *n_programs, sizeof(struct filterprg), fprg_compare);
    return fprgs;
}

static const char *
cache_get_path(void)
{
    return conf_static.storage_path;
}

static int
cache_delete_program(prgid_t id)
{
    char path[PATH_MAX + 100];

    /* lazy implementation, just try to delete one of every type (only one
       will match) */
    sprintf(path, "%srev.%s.tar", conf_static.storage_path,
	    filterprg_prgid_ntoa(id));
    remove(path);
    sprintf(path, "%sxtc.%s.tar", conf_static.storage_path,
	    filterprg_prgid_ntoa(id));
    remove(path);
    sprintf(path, "%seqr.%s.tar", conf_static.storage_path,
	    filterprg_prgid_ntoa(id));
    remove(path);    
    clean_fprog_refs();
    return 0;
}

static int
cache_flush_orphan_resources(void)
{
    struct dirent **namelist;
    char path[PATH_MAX];
    bool_t removed;
    int n;

    if ((non_orphan_fprgs =
         filterprg_scan_directory(conf_static.storage_path,
                                  &non_orphan_n_programs)) == NULL)
    {
	return 0;
    }
    if ((n = scandir(conf_static.storage_path, &namelist,
                     orphan_resources_scandir_select, NULL)) < 0)
    {
        elog(LOGDEBUG, "scandir failed: %s", strerror(errno));
        return 0;
    }
    path[PATH_MAX - 1] = '\0';
    removed = false;
    while (n--) {
        snprintf(path, PATH_MAX - 1, "%s%s", conf_static.storage_path,
                 namelist[n]->d_name);
        elog(LOGDEBUG, "removing orphan resource: \"%s\"", namelist[n]->d_name);
        remove(path);
        free(namelist[n]);
        removed = true;
    }
    free(namelist);

    if (removed) {
        /* a removed program may yeild new orphan resources */
        cache_flush_orphan_resources();
    }    
    return 0;
}

static int
sysinfo_get_ram_megabytes(void)
{    
    return current_sysinfo.mem_total / 1024;
}

static double
sysinfo_get_ram_used(void)
{
    return sysmon_memused();
}

static int
sysinfo_get_cpu_megahertz(void)
{
    return current_sysinfo.cpu_speed;
}

static int
sysinfo_get_cpu_count(void)
{
    return current_sysinfo.cpu_count;
}

static const char *
sysinfo_get_cpu_name(void)
{
    return current_sysinfo.cpu_modelname;
}

static double
sysinfo_get_cpu_maxload(void)
{
    return sysmon_cpuload();
}

static void
sysinfo_get_uptime(int *days,
		   int *hours)
{
    int _days, _hours;
    
    sysmon_uptime(&_days, &_hours);
    if (days != NULL) {
	*days = _days;
    }
    if (hours != NULL) {
	*hours = _hours;
    }
}

static int
sysinfo_get_filter_ram_megabytes(void)
{
    return conf_static.allocate_ram;
}

static const char *
sysinfo_get_version(void)
{
    return "0.87";
}

static double
sysinfo_get_temp_cpu(void)
{
    return sysmon_temp(conf_static.cpu_temp_proc_path);
}

static double
sysinfo_get_temp_motherboard(void)
{
    return sysmon_temp(conf_static.motherboard_temp_proc_path);
}

static bool_t
ctl_get_debug(void)
{
    return cu.ctl.debug;
}

static void
ctl_set_debug(bool_t activate)
{
    activate = !!activate;
    elog_loglevel_set(activate ? LOGDEBUG : LOGINFO);
    bfir_debug = activate;
    equaliser_debug = activate;
    if (cu.ctl.debug == activate) {
        return;
    }
    cu.ctl.debug = activate;
    save_userconf = true;
}

static bool_t
ctl_get_powersave(void)
{
    return cu.ctl.powersave;
}

static void
ctl_set_powersave(bool_t activate)
{
    activate = !!activate;
    if (cu.ctl.powersave == activate) {
        return;
    }
    cu.ctl.powersave = activate;
    save_userconf = true;
}

static bool_t
ctl_need_to_update_hard_settings(void)
{
    return update_hard_settings(true);
}

static int
ctl_update_hard_settings(void)
{
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    return update_hard_settings(false);
}

static void
ctl_schedule_save_config(void)
{
    save_userconf = true;
}

static bool_t
ctl_register_read_fd(int fd,
		     void (*callback)(int fd))
{
    if (fd < 0 || fd >= FD_SETSIZE) {
	elog(LOGDEBUG, "Tried to register fd out of range: %d.", fd);
	return false;
    }
    if (FD_ISSET(fd, &readfds)) {
	elog(LOGDEBUG, "Tried to register an already registered fd (%d).", fd);
	return false;
    }
    fd_callbacks[fd] = callback;
    FD_SET(fd, &readfds);
    if (fd > maxfd) {
	maxfd = fd;
    }    
    elog(LOGDEBUG, "File descriptor %d was registered", fd);
    return true;
}

static void
ctl_unregister_read_fd(int fd)
{
    if (fd < 0 || fd >= FD_SETSIZE) {
	elog(LOGDEBUG, "Tried to unregister fd out of range: %d.", fd);
	return;
    }
    if (!FD_ISSET(fd, &readfds)) {
	elog(LOGDEBUG, "Tried to unregister an unregistered fd (%d).", fd);
	return;
    }
    fd_callbacks[fd] = NULL;
    FD_CLR(fd, &readfds);
    elog(LOGDEBUG, "File descriptor %d was unregistered", fd);
    maxfd = 0;
    for (fd = 0; fd < FD_SETSIZE; fd++) {
        if (FD_ISSET(fd, &readfds)) {
            maxfd = fd;
        }                                        
    }
}

static int
ctl_register_timer(void *param,
		   void (*callback)(void *param),
		   struct timeval *timeval)
{
    int n;
    
    if (callback == NULL || timeval == NULL) {
	return -1;
    }
    if ((n = bit_find(cb_timers_freetab, 0, MAX_TIMERS - 1)) == -1) {
	elog(LOGDEBUG, "timer table is full");
	return -1;
    }    
    callback_timers[n].param = param;
    callback_timers[n].callback = callback;
    gettimeofday(&callback_timers[n].tvwhen, NULL);
    timeradd(&callback_timers[n].tvwhen, timeval, &callback_timers[n].tvwhen);
    bit_set(cb_timers_tab, n);
    bit_clr(cb_timers_freetab, n);
    return n;
}

static void
ctl_unregister_timer(int id)
{
    if (id < 0 || id >= MAX_TIMERS || bit_isset(cb_timers_freetab, id)) {
	return;
    }
    callback_timers[id].callback = NULL;
    bit_set(cb_timers_freetab, id);
    bit_clr(cb_timers_tab, id);
}

static const struct filterprg *
ctl_scan_fprg_dir(const char path[],
                  int type,
                  int *n_programs)
{
    static struct filterprg *fprgs = NULL;
    static int capacity = 0;

    struct filterprg *_fprgs;
    int _n_programs, n;
    bool_t add;
    
    if (path == NULL || n_programs == NULL) {
	return NULL;
    }    
    if ((_fprgs = filterprg_scan_directory(path, &_n_programs)) == NULL) {
	*n_programs = 0;
	return NULL;
    }
    if (_n_programs > capacity) {
	fprgs = realloc(fprgs, _n_programs * sizeof(struct filterprg));
	capacity = _n_programs;
    }
    for (n = (*n_programs) = 0; n < _n_programs; n++) {
	/* in the future there can be multiple reverb program formats (and
	   other), not just a simple 1-1 mapping like now */
	add = false;
	switch (type) {
	case VCU_PROGRAM_TYPE_REVERB:
	    if (_fprgs[n].type == FILTERPRG_TYPE_REVERB_TRADITIONAL ||
                _fprgs[n].type == FILTERPRG_TYPE_REVERB_BFORMAT)
            {
		add = true;
	    }
	    break;
	case VCU_PROGRAM_TYPE_XTC:
	    if (_fprgs[n].type == FILTERPRG_TYPE_CROSSTALK_CANCELLATION) {
		add = true;
	    }
	    break;
	case VCU_PROGRAM_TYPE_EQ:
	    if (_fprgs[n].type == FILTERPRG_TYPE_EQUALISATION) {
		add = true;
	    }
	    break;
	default:
	    add = true;
	    break;
	}
	if (add) {
	    fprgs[*n_programs] = _fprgs[n];
	    (*n_programs)++;
	}
    }
    if (*n_programs == 0) {
	return NULL;
    }
    qsort(fprgs, *n_programs, sizeof(struct filterprg), fprg_compare);
    return fprgs;
}

static const struct vcu_config *
ctl_scan_config_dir(const char path[],
                    int *n_configs)
{
    struct dirent **namelist;
    char cwd[PATH_MAX];
    int n;

    if (path == NULL || n_configs == NULL) {
        return NULL;
    }
    getcwd(cwd, PATH_MAX);
    if (chdir(path) == -1) {
	elog(LOGDEBUG, "could not change directory to \"%s\": %s", path,
	     strerror(errno));
	return NULL;
    }
    config_list.pos = 0;
    /* Sometimes scandir fails by some strange reason, although there have been
       lots of successful reads of entries. This takes care of this problem */
    n = scandir(".", &namelist, config_scandir_select, NULL);
    *n_configs = config_list.pos;
    if (n > 0) {
        while(n--) {
            free(namelist[n]);
        }
        free(namelist);
    }
    chdir(cwd);
    elog(LOGDEBUG, "found %d configs in \"%s\"", *n_configs, path);
    if (*n_configs > 0) {
        qsort(config_list.array, *n_configs, sizeof(struct vcu_config),
              config_compare);
        return config_list.array;
    }
    return NULL;
}

static void
ctl_get_load_config_settings(bool_t *load_user_config,
                             bool_t *load_revprg_config,
                             bool_t *load_static_config)
{
    if (load_user_config != NULL) {
        *load_user_config = cu.ctl.load_user_config;
    }
    if (load_revprg_config != NULL) {
        *load_revprg_config = cu.ctl.load_revprg_config;
    }
    if (load_static_config != NULL) {
        *load_static_config = cu.ctl.load_static_config;
    }
}

static void
ctl_set_load_config_settings(bool_t load_user_config,
                             bool_t load_revprg_config,
                             bool_t load_static_config)
{
    if (cu.ctl.load_user_config == !!load_user_config &&
        cu.ctl.load_revprg_config == !!load_revprg_config &&
        cu.ctl.load_static_config == !!load_static_config)
    {
        return;
    }
    cu.ctl.load_user_config = !!load_user_config;
    cu.ctl.load_revprg_config = !!load_revprg_config;
    cu.ctl.load_static_config = !!load_static_config;
    save_userconf = true;
}

static int
ctl_load_config(const struct vcu_config *config)
{
    char static_config_filename[PATH_MAX], *p, path[PATH_MAX], path2[PATH_MAX];
    char *argv[VCU_CONFIG_ARGS_MAX + 10];
    struct simple_tar_header tar_hdr;
    FILE *instream, *stream;
    struct stat buf;
    int err, n;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);

    if (config == NULL) {
        return VCU_ERR_INVALID_VALUE;
    }

    if (cu.ctl.load_static_config) {
        if (stat(config->cwd, &buf) != 0 || !S_ISDIR(buf.st_mode)) {
            return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
        }
    }

    strcpy(path, config->static_config_path);
    p = strrchr(path, PATH_SEPARATOR_CHR);
    if (p != NULL) {
        *p = '\0';
        strcpy(static_config_filename, &p[1]);
        if (cu.ctl.load_static_config) {
            if (stat(path, &buf) != 0 || !S_ISDIR(buf.st_mode)) {
                return VCU_ERR_PREREQUISITES_NOT_FULFILLED;
            }
        }
    } else {
        strcpy(static_config_filename, path);
    }
    
    if ((instream = fopen(config->path, "rb")) == NULL) {
        elog(LOGDEBUG, "could not open \"%s\" for reading: %s",
             config->path, strerror(errno));
        return VCU_ERR_READ_FAILED;
    }
    if (!tar_header_read(instream, NULL, &tar_hdr) ||
        strcmp(tar_hdr.filename, "config.txt") != 0)
    {
        fclose(instream);
        return VCU_ERR_READ_FAILED;
    }
    fseek(instream, tar_hdr.size + TAR_PAD_SIZE(tar_hdr.size), SEEK_CUR);
    while (tar_header_read(instream, NULL, &tar_hdr)) {
        if (strcmp(tar_hdr.filename, CONF_FILE_USER) == 0 &&
            cu.ctl.load_user_config)
        {
            sprintf(path, "%s%s.new", conf_static.storage_path,
                    CONF_FILE_USER);
            if ((stream = fopen(path, "wb")) == NULL) {
                fclose(instream);
                return VCU_ERR_WRITE_FAILED;
            }
            err = copystream(stream, NULL, instream, NULL, tar_hdr.size, false);
            fclose(stream);
            if (err != 0) {
                fclose(instream);
                remove(path);
                return err;
            }
        } else if (strcmp(tar_hdr.filename, static_config_filename) == 0 &&
                   cu.ctl.load_static_config)
        {
            sprintf(path, "%s.new", config->static_config_path);
            if ((stream = fopen(path, "wb")) == NULL) {
                fclose(instream);
                return VCU_ERR_WRITE_FAILED;
            }
            err = copystream(stream, NULL, instream, NULL, tar_hdr.size, false);
            fclose(stream);
            if (err != 0) {
                fclose(instream);
                remove(path);
                return err;
            }
        } else if (strstr(tar_hdr.filename, ".revprg") ==
                   &tar_hdr.filename[strlen(tar_hdr.filename) - 7])
        {
            sprintf(path, "%s%s", conf_static.storage_path, tar_hdr.filename);
            if ((stream = fopen(path, "wb")) == NULL) {
                fclose(instream);
                return VCU_ERR_WRITE_FAILED;
            }
            err = copystream(stream, NULL, instream, NULL, tar_hdr.size, false);
            fclose(stream);
            if (err != 0) {
                fclose(instream);
                remove(path);
                return err;
            }
        } else {
            fseek(instream, tar_hdr.size, SEEK_CUR);
        }
        fseek(instream, TAR_PAD_SIZE(tar_hdr.size), SEEK_CUR);
    }
    fclose(instream);

    if (cu.ctl.load_user_config) {
        sprintf(path, "%s%s.new", conf_static.storage_path, CONF_FILE_USER);
        sprintf(path2, "%s%s", conf_static.storage_path, CONF_FILE_USER);
        if ((err = copyfile(path2, path)) != 0) {
            return err;
        }
        remove(path);
    }
    if (cu.ctl.load_static_config) {
        elog(LOGDEBUG, "load static config");
        sprintf(path, "%s.new", config->static_config_path);
        if ((err = copyfile(config->static_config_path, path)) != 0) {
            save_userconf = true;
            periodic();
            return err;
        }
        remove(path);
    }    

    memset(argv, 0, sizeof(argv));
    if (cu.ctl.load_static_config) {
        argv[0] = start_argv[0];
        for (n = 0; n < config->argc; n++) {
            argv[n+1] = strdup(config->argv[n]);
        }
        n++;
    } else {
        for (n = 0; n < start_argc; n++) {
            argv[n] = start_argv[n];
        }
    }
    argv[n] = NULL;
    if (disable_daemon_mode) {
        for (n = 0; argv[n] != NULL; n++) {
            if (strcmp(argv[n], "-d") == 0) {
                break;
            }
        }
        if (argv[n] == NULL) {
            argv[n] = "-d";
        }
    }
    for (n = 0; argv[n] != NULL; n++) {
        elog(LOGDEBUG, "about to exec with argv[%d]: \"%s\"", n, argv[n]);
    }

    pre_exec_cleanup();
    execv(binary_path, argv);
    fprintf(stderr, "exec failed: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
    
    return VCU_ERR_UNKNOWN_ERROR;
}

static int
ctl_save_config(const char path[],
                const char name[],
                bool_t full_path)
{
    char dstpath[PATH_MAX + 100], srcpath[PATH_MAX + 100];
    char fname[VCU_CONFIG_NAME_SIZE + 20], numstr[20];
    int n, i, wsize, pathlen;
    struct dirent **namelist;
    struct stat buf;
    FILE *stream;

    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);

    pathlen = strlen(path);
    if (pathlen >= PATH_MAX || strlen(name) >= VCU_CONFIG_NAME_SIZE) {
        elog(LOGDEBUG, "too long path or name");
        return VCU_ERR_INVALID_VALUE;
    }
    
    sprintf(srcpath, "%sconfig.txt", conf_static.log_path);
    if ((stream = fopen(srcpath, "wt")) == NULL) {        
        elog(LOGDEBUG, "could not open \"%s\" for writing: %s",
             srcpath, strerror(errno));
        return VCU_ERR_WRITE_FAILED;
    }
    wsize = fprintf(stream, "\
name: \"%s\";\n\
static_config_path: \"%s\";\n\
cwd: \"%s\";\n\
argv: ",
                escape_string(NULL, name),
                static_config_path,
                start_workdir);
    if (wsize < strlen(escape_string(NULL, name)) + 10 +
        strlen(static_config_path) + 24 +
        strlen(start_workdir) + 9 + 6)
    {
        elog(LOGDEBUG, "could not write config.txt");
        fclose(stream);
        remove(srcpath);
        return VCU_ERR_WRITE_FAILED;
    }
    for (n = 1; n < start_argc; n++) {
        if (strcmp(start_argv[n], "-d") == 0) {
            continue;
        }
        wsize = fprintf(stream, "\"%s\"%s", escape_string(NULL, start_argv[n]),
                        (n == start_argc - 1) ? "" : ", ");
        if (wsize < strlen(start_argv[n]) + 2 +
            ((n == start_argc - 1) ? 0 : 2))
        {
            elog(LOGDEBUG, "could not write config.txt");
            fclose(stream);
            remove(srcpath);
            return VCU_ERR_WRITE_FAILED;
        }
    }
    if (fprintf(stream, ";\n") != 2) {
        elog(LOGDEBUG, "could not write config.txt");
        fclose(stream);
        remove(srcpath);
        return VCU_ERR_WRITE_FAILED;
    }
    fclose(stream);

    if (full_path) {
        strcpy(dstpath, path);
    } else {
        strcpy(fname, name);
        for (i = 0; fname[i] != '\0'; i++) {
            if (!((fname[i] >= '0' && fname[i] <= '9') ||
                  (fname[i] >= 'A' && fname[i] <= 'Z') ||
                  (fname[i] >= 'a' && fname[i] <= 'z') ||
                  fname[i] == '+' || fname[i] == '-' || fname[i] == '_'))
            {
                fname[i] = '_';
            }
        }
        if (i == 0) {
            strcpy(fname, "unnamed");
        }
        if (path[strlen(path)-1] != PATH_SEPARATOR_CHR) {
            sprintf(dstpath, "%s%c%s.tar", path, PATH_SEPARATOR_CHR, fname);
        } else {
            sprintf(dstpath, "%s%s.tar", path, fname);
        }
        n = 0;
        if (strlen(fname) > 8 || stat(dstpath, &buf) == 0) {
            for (n = 0; n < 100; n++) {
                sprintf(numstr, "%d", n);
                fname[8 - strlen(numstr)] = '\0';
                if (path[strlen(path)-1] != PATH_SEPARATOR_CHR) {
                    sprintf(dstpath, "%s%c%s%d.tar", path, PATH_SEPARATOR_CHR,
                            fname, n);
                } else {
                    sprintf(dstpath, "%s%s%d.tar", path, fname, n);
                }
                if (stat(dstpath, &buf) != 0) {
                    break;
                }
            }
        }
        if (n == 100) {
            elog(LOGDEBUG, "too many configurations");
            remove(srcpath);
            return VCU_ERR_OBJECT_EXISTS;
        }
    }
    if ((stream = fopen(dstpath, "wb")) == NULL) {
        elog(LOGDEBUG, "could not open \"%s\" for writing: %s",
             dstpath, strerror(errno));
        remove(srcpath);
        return VCU_ERR_WRITE_FAILED;
    }
    elog(LOGDEBUG, "Saving config to \"%s\"", dstpath);

    if (!tar_add(srcpath, stream, NULL)) {
        goto save_config_failure;
    }
    remove(srcpath);
    if (!tar_add(static_config_path, stream, NULL)) {
        goto save_config_failure;
    }
    sprintf(srcpath, "%s%s", conf_static.storage_path, CONF_FILE_USER);
    if (!tar_add(srcpath, stream, NULL)) {
        goto save_config_failure;
    }
    save_config_stream = stream;
    save_config_scandir_failure = false;
    if ((n = scandir(conf_static.storage_path, &namelist,
                     save_config_scandir_select, NULL)) < 0)
    {
        elog(LOGDEBUG, "scandir failed: %s", strerror(errno));
        goto save_config_failure;
    }
    while (n--) {
        free(namelist[n]);
    }
    free(namelist);
    if (save_config_scandir_failure) {
        goto save_config_failure;
    }
    fclose(stream);
    
    return 0;

 save_config_failure:
    fclose(stream);
    remove(dstpath);
    return VCU_ERR_WRITE_FAILED;
}

static const struct ui_module *
ctl_get_modules(int *n_modules)
{
    if (n_modules == NULL) {
	return NULL;
    }
    *n_modules = conf_static.n_uimodules;
    return ui_modules;
}

static const char *
ctl_strerror(int error)
{
    switch (error) {
    case VCU_ERR_UNKNOWN_ERROR:
	return "Unknown error";
    case VCU_ERR_INVALID_VALUE:
	return "Invalid value";
    case VCU_ERR_PREREQUISITES_NOT_FULFILLED:
	return "Prerequisites not fulfilled";
    case VCU_ERR_NOT_IMPLEMENTED:
	return "Not implemented";
    case VCU_ERR_NOT_SUPPORTED_BY_HARDWARE:
	return "Not supported by hardware";
    case VCU_ERR_INCORRECT_STATE:
	return "Incorrect state";
    case VCU_ERR_NO_SIGNAL:
	return "No signal";
    case VCU_ERR_UNSUPPORTED_SAMPLE_RATE:
	return "Unsupported sample rate";
    case VCU_ERR_NO_MEMORY:
	return "No memory";
    case VCU_ERR_LOAD_FAILED:
	return "Load failed";
    case VCU_ERR_RESAMPLING_FAILED:
	return "Resampling failed";
    case VCU_ERR_WRITE_FAILED:
	return "Write failed";
    case VCU_ERR_READ_FAILED:
	return "Read failed";
    case VCU_ERR_OBJECT_BUSY:
	return "Object busy";
    case VCU_ERR_OBJECT_TOO_LARGE:
	return "Object too large";
    case VCU_ERR_INVALID_FORMAT:
	return "Invalid format";
    case VCU_ERR_CONVOLVER_FAILED_UNKNOWN:
	return "Convolver failed - unknown";
    case VCU_ERR_CONVOLVER_FAILED_UNDERFLOW:
	return "Convolver failed - underflow";
    case VCU_ERR_CONVOLVER_FAILED_NO_MEMORY:
	return "Convolver failed - no memory";
    case VCU_ERR_CONVOLVER_FAILED_SIGNAL:
	return "Convolver failed - signal";
    case VCU_ERR_OBJECT_EXISTS:
        return "Object exists";
    case VCU_ERR_CONVOLVER_FAILED_TOO_COMPLEX:
	return "Convolver failed - too complex";
    default:
	return "Unknown error code";
    }
}

static const char *
ctl_get_log_path(void)
{
    return conf_static.log_path;
}

static const char *
ctl_get_log_file_path(void)
{
    static char path[PATH_MAX + 100];
    
    snprintf(path, PATH_MAX, "%s%s", conf_static.log_path, ALMUSVCU_LOG);
    path[PATH_MAX - 1] = '\0';
    return path;
}

static void
ctl_flush_logs(void)
{
    MUST_NOT_BE_IN_RUNTIME( );
    
    elog(LOGDEBUG, "flushing logs");
    bfir_flush_logs();
    elog_truncate();
}

static int
ctl_save_logs(const char path[])
{
    char srcpath[PATH_MAX + 100];
    char dstpath[PATH_MAX + 100];
    struct stat buf;
    gzFile gzfile;
    int n, err;

    n = strlen(path);
    if (n >= PATH_MAX) {
        return VCU_ERR_INVALID_VALUE;
    }
    if (path[n-1] != PATH_SEPARATOR_CHR) {
        sprintf(dstpath, "%s%c%s", path, PATH_SEPARATOR_CHR, VCU_LOG_ARCHIVE);
    } else {
        sprintf(dstpath, "%s%s", path, VCU_LOG_ARCHIVE);
    }
    
    if ((gzfile = gzopen(dstpath, "wb9")) == NULL) {
        elog(LOGDEBUG, "failed to open file \"%s\" for writing: %s",
             dstpath, strerror(errno));
        return VCU_ERR_WRITE_FAILED;
    }
    sprintf(srcpath, "%s%s", conf_static.log_path, ALMUSVCU_LOG);
    if (!tar_add(srcpath, NULL, gzfile)) {
        goto save_logs_failure;
    }
    sprintf(srcpath, "%s%s.old", conf_static.log_path, ALMUSVCU_LOG);
    if (stat(srcpath, &buf) == 0 && !tar_add(srcpath, NULL, gzfile)) {
        goto save_logs_failure;
    }
    sprintf(srcpath, "%s%s", conf_static.log_path,
            BFIR_DEBUG_LAST_BRUTEFIR_CONFIG);
    if (stat(srcpath, &buf) == 0 && !tar_add(srcpath, NULL, gzfile)) {
        goto save_logs_failure;
    }
    sprintf(srcpath, "%s%s", conf_static.log_path,
            BFIR_DEBUG_BRUTEFIR_CONSOLE_OUTPUT);
    if (stat(srcpath, &buf) == 0 && !tar_add(srcpath, NULL, gzfile)) {
        goto save_logs_failure;
    }
    sprintf(srcpath, "%s%s", conf_static.log_path,
            BFIR_DEBUG_BRUTEFIR_CLI_OUTPUT);
    if (stat(srcpath, &buf) == 0 && !tar_add(srcpath, NULL, gzfile)) {
        goto save_logs_failure;
    }
    sprintf(srcpath, "%s%s", conf_static.log_path,
            BFIR_DEBUG_BRUTEFIR_CLI_INPUT);
    if (stat(srcpath, &buf) == 0 && !tar_add(srcpath, NULL, gzfile)) {
        goto save_logs_failure;
    }
    if (!tar_add(static_config_path, NULL, gzfile)) {
        goto save_logs_failure;
    }
    sprintf(srcpath, "%s%s", conf_static.storage_path, CONF_FILE_USER);
    if (!tar_add(srcpath, NULL, gzfile)) {
        goto save_logs_failure;
    }
    sprintf(srcpath, "%s%s.new", conf_static.storage_path, CONF_FILE_USER);
    if (stat(srcpath, &buf) == 0 && !tar_add(srcpath, NULL, gzfile)) {
        goto save_logs_failure;
    }
    sprintf(srcpath, "%s%s.failed", conf_static.storage_path, CONF_FILE_USER);
    if (stat(srcpath, &buf) == 0 && !tar_add(srcpath, NULL, gzfile)) {
        goto save_logs_failure;
    }
    sprintf(srcpath, "%ssysinfo.txt", conf_static.log_path);
    if (!sysmon_dumpinfo(srcpath) || !tar_add(srcpath, NULL, gzfile)) {
        goto save_logs_failure;
    }
    remove(srcpath);
    if (gzclose(gzfile) != 0) {
        elog(LOGDEBUG, "gzclose failed: %s", gzerror(gzfile, &err));
        return VCU_ERR_WRITE_FAILED;
    }
    
    return 0;

 save_logs_failure:
    gzclose(gzfile);
    remove(dstpath);
    return VCU_ERR_WRITE_FAILED;
}

static int
ctl_update_software(const char path[])
{
    bool_t almusvcu_updated, brutefir_updated;
    struct simple_tar_header tar_hdr;
    char destpath[PATH_MAX], **argv;
    gzFile gzfile;
    FILE *stream;
    int err, n;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);
    
    if (path == NULL) {
        return VCU_ERR_INVALID_VALUE;
    }

    elog(LOGDEBUG, "update from \"%s\"", path);
    
    /* we don't care to check format, gzopen handles both compressed and
       uncompressed files */
    if ((gzfile = gzopen(path, "rb")) == NULL) {
        elog(LOGDEBUG, "gzopen failed: %s", strerror(errno));
        return VCU_ERR_READ_FAILED;
    }

    almusvcu_updated = false;
    brutefir_updated = false;
    while (tar_header_read(NULL, gzfile, &tar_hdr)) {
        if (strcmp(tar_hdr.filename, "almusvcu") == 0) {
            almusvcu_updated = true;
            /* move the current binary to avoid overwriting it in runtime,
               we do no error checking, we just proceed if it would fail */
            snprintf(destpath, PATH_MAX, "%s.old", binary_path);
            destpath[PATH_MAX - 1] = '\0';
            remove(destpath);
            rename(binary_path, destpath);
            strcpy(destpath, binary_path);
        } else if (strcmp(tar_hdr.filename, "brutefir") == 0) {
            brutefir_updated = true;
            strcpy(destpath, conf_static.brutefir_path);
        } else if (strstr(tar_hdr.filename, ".vcu") ==
                   &tar_hdr.filename[strlen(tar_hdr.filename) - 4])
        {
            almusvcu_updated = true;
            snprintf(destpath, PATH_MAX, "%s%s", conf_static.uimodules_path,
                     tar_hdr.filename);
        } else if (strstr(tar_hdr.filename, ".bfio") ==
                   &tar_hdr.filename[strlen(tar_hdr.filename) - 5] ||
                   strstr(tar_hdr.filename, ".bflogic") ==
                   &tar_hdr.filename[strlen(tar_hdr.filename) - 8])
        {
            brutefir_updated = true;
            snprintf(destpath, PATH_MAX, "%s%s",
                     conf_static.brutefir_modules_path, tar_hdr.filename);
        } else {
            break;
        }
        destpath[PATH_MAX-1] = '\0';
        elog(LOGDEBUG, "copy \"%s\" in \"%s\" to \"%s\"",
             tar_hdr.filename, path, destpath);
        remove(destpath);
        n = open(destpath, O_CREAT | O_EXCL | O_TRUNC | O_WRONLY,
                 S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
        if (n == -1) {
            elog(LOGDEBUG, "could not open file \"%s\" for writing: %s",
                 destpath, strerror(errno));
            gzclose(gzfile);
            return VCU_ERR_WRITE_FAILED;
        }
        if ((stream = fdopen(n, "wb")) == NULL) {
            elog(LOGDEBUG, "fdopen failed: %s", strerror(errno));
            close(n);
            gzclose(gzfile);
            return VCU_ERR_WRITE_FAILED;
        }
        err = copystream(stream, NULL, NULL, gzfile, tar_hdr.size, false);
        fclose(stream);
        if (err != 0) {
            gzclose(gzfile);
            return err;
        }
        gzseek(gzfile, TAR_PAD_SIZE(tar_hdr.size), SEEK_CUR);
    }
    gzclose(gzfile);

    if (!brutefir_updated && !almusvcu_updated) {
        elog(LOGDEBUG, "invalid format");
        return VCU_ERR_INVALID_FORMAT;
    }
    if (almusvcu_updated) {
        /* we need to restart */
        elog(LOGDEBUG, "update finished, about to restart (argc: %d)",
             start_argc);
        argv = alloca((start_argc + 1) * sizeof(char *));
        for (n = 0; n < start_argc; n++) {
            argv[n] = start_argv[n];
        }
        argv[n] = NULL;
        pre_exec_cleanup();
        execv(binary_path, argv);
        fprintf(stderr, "exec failed: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }    
    elog(LOGDEBUG, "update finished, no need to restart");
    return 0;
}

static int
ctl_restart(bool_t reset_revprg_config,
            bool_t reset_user_config)
{
    char *argv[VCU_CONFIG_ARGS_MAX + 10], path[PATH_MAX];
    struct dirent **namelist;
    int n;
    
    MUST_NOT_BE_IN_RUNTIME(VCU_ERR_PREREQUISITES_NOT_FULFILLED);

    if (reset_revprg_config) {
        if ((n = scandir(conf_static.storage_path, &namelist,
                         revprg_config_scandir_select, NULL)) < 0)
        {
            elog(LOGDEBUG, "scandir failed: %s", strerror(errno));
            return 0;
        }
        path[PATH_MAX - 1] = '\0';
        while (n--) {
            snprintf(path, PATH_MAX - 1, "%s%s", conf_static.storage_path,
                     namelist[n]->d_name);
            elog(LOGDEBUG, "removing reverb config: \"%s\"",
                 namelist[n]->d_name);
            remove(path);
            free(namelist[n]);
        }
        free(namelist);
    }
    
    if (reset_user_config) {
        sprintf(path, "%s%s", conf_static.storage_path, CONF_FILE_USER);
        remove(path);
    }
    
    memset(argv, 0, sizeof(argv));
    for (n = 0; n < start_argc; n++) {
        argv[n] = start_argv[n];
    }
    argv[n] = NULL;
    if (disable_daemon_mode) {
        for (n = 0; argv[n] != NULL; n++) {
            if (strcmp(argv[n], "-d") == 0) {
                break;
            }
        }
        if (argv[n] == NULL) {
            argv[n] = "-d";
        }
    }

    pre_exec_cleanup();
    execv(binary_path, argv);
    fprintf(stderr, "exec failed: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
    
    return VCU_ERR_UNKNOWN_ERROR;
}

static void
ctl_exit(int exitcode)
{
    exit(exitcode);
}

int
main(int argc,
     char *argv[])
{
    int (*select_function)(int, fd_set *, fd_set *, fd_set *, struct timeval *);
    char path[PATH_MAX], *config_file;
    int ret, n, i, err, min_blocksize;
    struct timeval tv, tvnow;
    bool_t progress_is_set;
    struct equaliser *eq;
    fd_set rfds;

    if (sizeof(float) != 4 || sizeof(double) != 8) {
        fprintf(stderr, "unexpected floating point size.\n");
        exit(EXIT_FAILURE);
    }
    
    if (getenv("VG_ARGS") != NULL) {
        /* redirect stderr output when running with valgrind */
        n = open("/tmp/almusvcu_stderr", O_WRONLY | O_CREAT | O_TRUNC, 0666);
        dup2(n, STDERR_FILENO);
    }

    config_file = NULL;
    disable_daemon_mode = false;
    for (n = 1; n < argc; n++) {
        if (strcmp(argv[n], "-d") == 0) {
            disable_daemon_mode = true;
        } else {
            config_file = argv[n];
        }
    }
    
    start_argc = argc;
    start_argv = argv;
    getcwd(start_workdir, sizeof(start_workdir));
    path_make_absolute(start_workdir, true, true);

    if (getenv("VG_ARGS") != NULL) {
        elog(LOGINFO, "assuming we are being run with valgrind");
        strcpy(binary_path, "/tmp/dummy");
    } else {
        if (getenv("_") == NULL) {
            elog(LOGERROR, "Failed to get \"_\" environment variable");
            exit(EXIT_FAILURE);
        }
        strncpy(path, getenv("_"), PATH_MAX);
        path[PATH_MAX - 1] = '\0';
        if (path[0] == '.') {
            snprintf(binary_path, PATH_MAX, "%s%s", start_workdir, path);
            binary_path[PATH_MAX - 1] = '\0';
        } else {
            strcpy(binary_path, path);
        }
    }
        
    /*
     * Initialise the vcu struct.
     */
    memset(&vcu, 0, sizeof(vcu));    
    vcu.runtime.isruntime = runtime_isruntime;
    vcu.runtime.start = runtime_start;
    vcu.runtime.start_file_input = runtime_start_file_input;
    vcu.runtime.stop = runtime_stop;
    vcu.runtime.buffer_commands = runtime_buffer_commands;
    vcu.runtime.signal.get_input_sample_rate =
	runtime_signal_get_input_sample_rate;
    vcu.runtime.signal.reset_peak_meter = runtime_signal_reset_peak_meter;
    vcu.runtime.volume.get_range = runtime_volume_get_range;
    vcu.runtime.volume.get_master = runtime_volume_get_master;
    vcu.runtime.volume.set_master = runtime_volume_set_master;
    vcu.runtime.volume.get_group = runtime_volume_get_group;
    vcu.runtime.volume.set_group = runtime_volume_set_group;
    vcu.runtime.volume.get_mute = runtime_volume_get_mute;
    vcu.runtime.volume.set_mute = runtime_volume_set_mute;
    vcu.runtime.volume.get_group_mute = runtime_volume_get_group_mute;
    vcu.runtime.volume.set_group_mute = runtime_volume_set_group_mute;
    vcu.runtime.volume.get_master_temp_mute =
        runtime_volume_get_master_temp_mute;
    vcu.runtime.volume.set_master_temp_mute =
        runtime_volume_set_master_temp_mute;
    vcu.runtime.volume.get_trim = runtime_volume_get_trim;
    vcu.runtime.volume.set_trim = runtime_volume_set_trim;
    vcu.runtime.delay.get_max = runtime_delay_get_max;
    vcu.runtime.delay.get_trim = runtime_delay_get_trim;
    vcu.runtime.delay.set_trim = runtime_delay_set_trim;
    vcu.runtime.eq.get_dynamic_equalisers = runtime_eq_get_dynamic_equalisers;
    vcu.runtime.eq.modify_dynamic_equaliser =
        runtime_eq_modify_dynamic_equaliser;
    vcu.runtime.eq.get_dynamic_equaliser_volume_range =
        runtime_eq_get_dynamic_equaliser_volume_range;
    vcu.runtime.eq.get_dynamic_equaliser_smallest_band =
        runtime_eq_get_dynamic_equaliser_smallest_band;
    vcu.runtime.eq.get_mode = runtime_eq_get_mode;
    vcu.runtime.eq.set_mode = runtime_eq_set_mode;
    vcu.runtime.eq.get_mode_active = runtime_eq_get_mode_active;
    vcu.runtime.reverb.get_id = runtime_reverb_get_id;
    vcu.runtime.reverb.set_id = runtime_reverb_set_id;
    vcu.runtime.reverb.get_trim_volume_range =
	runtime_reverb_get_trim_volume_range;
    vcu.runtime.reverb.get_trim_volume = runtime_reverb_get_trim_volume;
    vcu.runtime.reverb.set_trim_volume = runtime_reverb_set_trim_volume;
    vcu.runtime.reverb.get_xyz_balance_range =
        runtime_reverb_get_xyz_balance_range;
    vcu.runtime.reverb.get_xyz_balance = runtime_reverb_get_xyz_balance;
    vcu.runtime.reverb.set_xyz_balance = runtime_reverb_set_xyz_balance;
    vcu.runtime.reverb.get_trim_delay_range =
        runtime_reverb_get_trim_delay_range;
    vcu.runtime.reverb.get_trim_delay = runtime_reverb_get_trim_delay;
    vcu.runtime.reverb.set_trim_delay = runtime_reverb_set_trim_delay;
    vcu.runtime.reverb.get_scale_length_range =
        runtime_reverb_get_scale_length_range;
    vcu.runtime.reverb.get_scale_length = runtime_reverb_get_scale_length;
    vcu.runtime.reverb.set_scale_length = runtime_reverb_set_scale_length;
    vcu.runtime.reverb.get_response_mode = runtime_reverb_get_response_mode;
    vcu.runtime.reverb.set_response_mode = runtime_reverb_set_response_mode;
    vcu.runtime.reverb.get_decorrelate_mode =
        runtime_reverb_get_decorrelate_mode;
    vcu.runtime.reverb.set_decorrelate_mode =
        runtime_reverb_set_decorrelate_mode;
    vcu.runtime.reverb.get_force_delay_diff =
        runtime_reverb_get_force_delay_diff;
    vcu.runtime.reverb.set_force_delay_diff =
        runtime_reverb_set_force_delay_diff;
    vcu.runtime.reverb.get_force_decay = runtime_reverb_get_force_decay;
    vcu.runtime.reverb.set_force_decay = runtime_reverb_set_force_decay;
    vcu.runtime.reverb.get_dry_active = runtime_reverb_get_dry_active;
    vcu.runtime.reverb.set_dry_active = runtime_reverb_set_dry_active;
    vcu.runtime.reverb.get_dry_volume = runtime_reverb_get_dry_volume;
    vcu.runtime.reverb.set_dry_volume = runtime_reverb_set_dry_volume;
    vcu.runtime.reverb.get_bformat_directivity =
        runtime_reverb_get_bformat_directivity;
    vcu.runtime.reverb.set_bformat_directivity =
        runtime_reverb_set_bformat_directivity;
    vcu.runtime.reverb.get_preprocess_properties =
        runtime_reverb_get_preprocess_properties;
    vcu.runtime.reverb.get_postprocess_properties =
        runtime_reverb_get_postprocess_properties;
    vcu.runtime.ambiopole.get_id = runtime_ambiopole_get_id;
    vcu.runtime.ambiopole.set_id = runtime_ambiopole_set_id;
    vcu.runtime.channel.get_mode = runtime_channel_get_mode;
    vcu.runtime.channel.set_mode = runtime_channel_set_mode;
    vcu.runtime.channel.get_active = runtime_channel_get_active;
    vcu.runtime.channel.get_type_active = runtime_channel_get_type_active;
    vcu.runtime.channel.get_group_active = runtime_channel_get_group_active;
    vcu.signal.get_soft_clip = signal_get_soft_clip;
    vcu.signal.set_soft_clip = signal_set_soft_clip;
    vcu.signal.get_max_dither_bits = signal_get_max_dither_bits;
    vcu.signal.get_dither = signal_get_dither;
    vcu.signal.set_dither = signal_set_dither;
    vcu.signal.get_watch_clock = signal_get_watch_clock;
    vcu.signal.set_watch_clock = signal_set_watch_clock;
    vcu.signal.get_double_precision = signal_get_double_precision;
    vcu.signal.set_double_precision = signal_set_double_precision;
    vcu.signal.get_allowed_output_bits = signal_get_allowed_output_bits;
    vcu.signal.get_output_bits = signal_get_output_bits;
    vcu.signal.set_output_bits = signal_set_output_bits;
    vcu.signal.get_allowed_rates = signal_get_allowed_rates;
    vcu.signal.get_rates = signal_get_rates;
    vcu.signal.set_rates = signal_set_rates;
    vcu.signal.get_suggested_rate = signal_get_suggested_rate;
    vcu.signal.set_suggested_rate = signal_set_suggested_rate;
    vcu.channel.get_max_count = channel_get_max_count;
    vcu.channel.get_count = channel_get_count;
    vcu.channel.set_count = channel_set_count;
    vcu.channel.get_phys_map = channel_get_phys_map;
    vcu.channel.set_phys_map = channel_set_phys_map;
    vcu.channel.distance_max = channel_distance_max;
    vcu.channel.position_get = channel_position_get;
    vcu.channel.position_set = channel_position_set;
    vcu.channel.get_stage_max_count = channel_get_stage_max_count;
    vcu.channel.get_stage_count = channel_get_stage_count;
    vcu.channel.get_stage = channel_get_stage;
    vcu.channel.new_stage = channel_new_stage;
    vcu.channel.modify_stage = channel_modify_stage;
    vcu.channel.delete_stage = channel_delete_stage;
    vcu.channel.get_distance_compensate_active =
        channel_get_distance_compensate_active;
    vcu.channel.set_distance_compensate_active =
        channel_set_distance_compensate_active;
    vcu.channel.get_group_max_count = channel_get_group_max_count;
    vcu.channel.get_group_count = channel_get_group_count;
    vcu.channel.get_group = channel_get_group;
    vcu.channel.new_group = channel_new_group;
    vcu.channel.modify_group = channel_modify_group;
    vcu.channel.delete_group = channel_delete_group;
    vcu.channel.reorder_groups = channel_reorder_groups;
    vcu.channel.get_mode_max_count = channel_get_mode_max_count;
    vcu.channel.get_mode_count = channel_get_mode_count;
    vcu.channel.get_mode = channel_get_mode;
    vcu.channel.new_mode = channel_new_mode;
    vcu.channel.modify_mode = channel_modify_mode;
    vcu.channel.delete_mode = channel_delete_mode;
    vcu.channel.reorder_modes = channel_reorder_modes;
    vcu.eq.get_equaliser_count = eq_get_equaliser_count;
    vcu.eq.get_equaliser = eq_get_equaliser;
    vcu.eq.new_equaliser = eq_new_equaliser;
    vcu.eq.modify_equaliser = eq_modify_equaliser;
    vcu.eq.delete_equaliser = eq_delete_equaliser;
    vcu.eq.get_equaliser_properties = eq_get_equaliser_properties;
    vcu.eq.get_capacity = eq_get_capacity;
    vcu.eq.get_programs = eq_get_programs;
    vcu.eq.load_program = eq_load_program;
    vcu.eq.delete_program = eq_delete_program;
    vcu.eq.get_mode_max_count = eq_get_mode_max_count;
    vcu.eq.get_mode_count = eq_get_mode_count;
    vcu.eq.get_mode = eq_get_mode;
    vcu.eq.new_mode = eq_new_mode;
    vcu.eq.modify_mode = eq_modify_mode;
    vcu.eq.delete_mode = eq_delete_mode;
    vcu.eq.reorder_modes = eq_reorder_modes;
    vcu.eq.mode_is_assignable = eq_mode_is_assignable;
    vcu.iodelay.need_to_run_benchmark = iodelay_need_to_run_benchmark;
    vcu.iodelay.has_run_benchmark = iodelay_has_run_benchmark;
    vcu.iodelay.manually_set = iodelay_manually_set;
    vcu.iodelay.run_benchmark = iodelay_run_benchmark;
    vcu.iodelay.get_benchmark_max_load = iodelay_get_benchmark_max_load;
    vcu.iodelay.set_benchmark_max_load = iodelay_set_benchmark_max_load;
    vcu.iodelay.get_benchmark_max_length_range =
        iodelay_get_benchmark_max_length_range;
    vcu.iodelay.get_benchmark_max_length = iodelay_get_benchmark_max_length;
    vcu.iodelay.set_benchmark_max_length = iodelay_set_benchmark_max_length;
    vcu.iodelay.get_dlp_alternatives = iodelay_get_dlp_alternatives;
    vcu.iodelay.get_delay_length_pair = iodelay_get_delay_length_pair;
    vcu.iodelay.set_delay_length_pair = iodelay_set_delay_length_pair;
    vcu.iodelay.set_delay_length_pair_manual =
        iodelay_set_delay_length_pair_manual;
    vcu.iodelay.get_alignment_delay = iodelay_get_alignment_delay;
    vcu.iodelay.has_truncated_filters = iodelay_has_truncated_filters;
    vcu.ambiopole.get_xtc_active = ambiopole_get_xtc_active;
    vcu.ambiopole.set_xtc_active = ambiopole_set_xtc_active;
    vcu.ambiopole.get_programs = ambiopole_get_programs;
    vcu.ambiopole.get_capacity = ambiopole_get_capacity;
    vcu.ambiopole.load_program = ambiopole_load_program;
    vcu.ambiopole.delete_program = ambiopole_delete_program;
    vcu.reverb.get_capacity = reverb_get_capacity;
    vcu.reverb.get_programs = reverb_get_programs;
    vcu.reverb.load_program = reverb_load_program;
    vcu.reverb.copy_program = reverb_copy_program;
    vcu.reverb.delete_program = reverb_delete_program;
    vcu.reverb.get_response_mode = reverb_get_response_mode;
    vcu.reverb.set_response_mode = reverb_set_response_mode;
    vcu.reverb.get_decorrelate_mode = reverb_get_decorrelate_mode;
    vcu.reverb.set_decorrelate_mode = reverb_set_decorrelate_mode;
    vcu.reverb.get_force_delay_diff = reverb_get_force_delay_diff;
    vcu.reverb.set_force_delay_diff = reverb_set_force_delay_diff;
    vcu.reverb.get_force_decay = reverb_get_force_decay;
    vcu.reverb.set_force_decay = reverb_set_force_decay;
    vcu.reverb.get_dry_active = reverb_get_dry_active;
    vcu.reverb.set_dry_active = reverb_set_dry_active;
    vcu.reverb.get_dry_volume_range = reverb_get_dry_volume_range;
    vcu.reverb.get_dry_volume = reverb_get_dry_volume;
    vcu.reverb.set_dry_volume = reverb_set_dry_volume;
    vcu.reverb.get_bformat_directivity_range =
        reverb_get_bformat_directivity_range;
    vcu.reverb.get_bformat_directivity = reverb_get_bformat_directivity;
    vcu.reverb.set_bformat_directivity = reverb_set_bformat_directivity;
    vcu.reverb.get_source_layout = reverb_get_source_layout;
    vcu.reverb.set_source_layout = reverb_set_source_layout;
    vcu.ambisonics.get_input_format = ambisonics_get_input_format;
    vcu.ambisonics.set_input_format = ambisonics_set_input_format;
    vcu.cache.get_capacity = cache_get_capacity;
    vcu.cache.get_programs = cache_get_programs;
    vcu.cache.get_path = cache_get_path;
    vcu.cache.delete_program = cache_delete_program;
    vcu.cache.flush_orphan_resources = cache_flush_orphan_resources;
    vcu.sysinfo.get_ram_megabytes = sysinfo_get_ram_megabytes;
    vcu.sysinfo.get_ram_used = sysinfo_get_ram_used;
    vcu.sysinfo.get_cpu_megahertz = sysinfo_get_cpu_megahertz;
    vcu.sysinfo.get_cpu_count = sysinfo_get_cpu_count;
    vcu.sysinfo.get_cpu_name = sysinfo_get_cpu_name;
    vcu.sysinfo.get_cpu_maxload = sysinfo_get_cpu_maxload;
    vcu.sysinfo.get_uptime = sysinfo_get_uptime;    
    vcu.sysinfo.get_filter_ram_megabytes = sysinfo_get_filter_ram_megabytes;
    vcu.sysinfo.get_version = sysinfo_get_version;
    vcu.sysinfo.get_temp_cpu = sysinfo_get_temp_cpu;
    vcu.sysinfo.get_temp_motherboard = sysinfo_get_temp_motherboard;
    vcu.ctl.get_debug = ctl_get_debug;
    vcu.ctl.set_debug = ctl_set_debug;
    vcu.ctl.get_powersave = ctl_get_powersave;
    vcu.ctl.set_powersave = ctl_set_powersave;
    vcu.ctl.need_to_update_hard_settings = ctl_need_to_update_hard_settings;
    vcu.ctl.update_hard_settings = ctl_update_hard_settings;
    vcu.ctl.schedule_save_config = ctl_schedule_save_config;
    vcu.ctl.register_read_fd = ctl_register_read_fd;
    vcu.ctl.unregister_read_fd = ctl_unregister_read_fd;
    vcu.ctl.register_timer = ctl_register_timer;
    vcu.ctl.unregister_timer = ctl_unregister_timer;
    vcu.ctl.scan_fprg_dir = ctl_scan_fprg_dir;
    vcu.ctl.scan_config_dir = ctl_scan_config_dir;
    vcu.ctl.get_load_config_settings = ctl_get_load_config_settings;
    vcu.ctl.set_load_config_settings = ctl_set_load_config_settings;
    vcu.ctl.load_config = ctl_load_config;
    vcu.ctl.save_config = ctl_save_config;
    vcu.ctl.get_modules = ctl_get_modules;
    vcu.ctl.strerror = ctl_strerror;
    vcu.ctl.get_log_path = ctl_get_log_path;
    vcu.ctl.get_log_file_path = ctl_get_log_file_path;
    vcu.ctl.flush_logs = ctl_flush_logs;
    vcu.ctl.save_logs = ctl_save_logs;
    vcu.ctl.update_software = ctl_update_software;
    vcu.ctl.restart = ctl_restart;
    vcu.ctl.exit = ctl_exit;
    
    /*
     * Initialise static variables
     */
    FD_ZERO(&readfds);
    memset(&runtime_delay_map, 0, sizeof(runtime_delay_map));
    memset(&benchmark, 0, sizeof(benchmark));
    memset(fd_callbacks, 0, sizeof(fd_callbacks));
    memset(ui_modules, 0, sizeof(ui_modules));
    memset(cb_timers_tab, 0, sizeof(cb_timers_tab));
    memset(cb_timers_freetab, 0xFF, sizeof(cb_timers_freetab));
    memset(callback_timers, 0, sizeof(callback_timers));
    memset(&config_list, 0, sizeof(config_list));
    memset(&zeroid, 0, sizeof(zeroid));
    sysmon_sysinfo(&current_sysinfo);

    /*
     * Load static configuration from disk
     */
    if (config_file == NULL) {
        strcpy(static_config_path, "/etc/" CONF_FILE_STATIC);
    } else {        
        strncpy(static_config_path, config_file, PATH_MAX);
        static_config_path[PATH_MAX - 1] = '\0';
        path_make_absolute(static_config_path, false, true);
    }
    conf_load_static(static_config_path);
    
    if (disable_daemon_mode) {
        conf_static.run_as_daemon = false;
    }
    
    if (conf_static.run_as_daemon) {
        /* deamon initialisation stuff */
        switch (fork()) {
        case 0:
            break;
        case -1:
            elog(LOGERROR, "fork failed: %s", strerror(errno));
            exit(EXIT_FAILURE);
        default:
            exit(EXIT_SUCCESS);
        }
        if (setsid() == -1) {
            elog(LOGERROR, "setsid failed: %s", strerror(errno));
            exit(EXIT_FAILURE);
        }
        if (chdir("/") == -1) {
            elog(LOGERROR, "chdir failed: %s", strerror(errno));
            exit(EXIT_FAILURE);
        }
        umask(0);
    }

    conf_create_pidfile();
    
    eq_init(ctl_get_log_path());
    elog_init(ctl_get_log_file_path());

    min_blocksize = 0;
    for (n = 0; n < conf_static.input.n_devices; n++) {
        i = sysmon_verify_bfir_device(conf_static.input.device[n],
                                      conf_static.input.device_params[n], true);
        if (i > min_blocksize) {
            min_blocksize = i;
        }
    }
    for (n = 0; n < conf_static.input.n_devices; n++) {
        i = sysmon_verify_bfir_device(conf_static.output.device[n],
                                      conf_static.output.device_params[n],
                                      false);
        if (i > min_blocksize) {
            min_blocksize = i;
        }
    }
    if (conf_static.min_device_blocksize > 0) {
        min_blocksize = conf_static.min_device_blocksize;
    }
    
    /*
     * Load user configuration from disk
     */
    conf_load_user();
    ctl_set_debug(cu.ctl.debug);

    elog(LOGDEBUG, "Binary path: \"%s\"", binary_path);

    /*
     * Initialise modules
     */
    select_function = NULL;
    progress_is_set = false;
    for (n = 0; n < conf_static.n_uimodules; n++) {
	ui_modules[n].name = strdup(conf_static.uimodules[n]->name());
	if (conf_static.uimodules[n]->share == NULL) {
	    ui_modules[n].shared_data = NULL;
	} else {
	    ui_modules[n].shared_data = conf_static.uimodules[n]->share();
	}
        if (conf_static.run_as_daemon &&
            conf_static.uimodules[n]->needterm != NULL &&
            conf_static.uimodules[n]->needterm())
        {
            elog(LOGINFO, "Module \"%s\" requires terminal, cannot be run in "
                 "daemon mode. Ignoring module", ui_modules[n].name);
            conf_static.n_uimodules--;
            for (i = n; i < conf_static.n_uimodules; i++) {
                conf_static.uimodules[i] =
                    conf_static.uimodules[i+1];
            }
            n--;
            continue;
        }
	if (conf_static.uimodules[n]->select_wrapper != NULL) {
            if (select_function != NULL) {
                elog(LOGERROR, "There can onle be one module with a "
                     "select wrapper.\n");
                exit(EXIT_FAILURE);
            }
            select_function = conf_static.uimodules[n]->select_wrapper;
        }
	if (conf_static.uimodules[n]->progress.init_phase != NULL &&
	    conf_static.uimodules[n]->progress.benchmark != NULL &&
	    conf_static.uimodules[n]->progress.load_program != NULL &&
	    conf_static.uimodules[n]->progress.generate_eq != NULL)
	{
	    if (progress_is_set) {
		elog(LOGERROR, "There can only be one master module (one with "
                     "progress callbacks).");
		exit(EXIT_FAILURE);
	    } else {
		progress_is_set = true;
		progress = conf_static.uimodules[n]->progress;		
	    }
	}
    }
    if (select_function == NULL) {
        select_function = select;
    }
    if (!progress_is_set) {
	elog(LOGERROR, "There is no master module (one with progress "
             "callbacks).");
	exit(EXIT_FAILURE);
    }
    for (n = 0; n < conf_static.n_uimodules; n++) {
	if (conf_static.uimodules[n]->init != NULL) {
	    if (!conf_static.uimodules[n]->init(&vcu)) {
		elog(LOGERROR, "Initialisation of module \"%s\" failed.",
		     ui_modules[n].name);
		exit(EXIT_FAILURE);
	    }
	}
    }

    /*
     * Install signal handler
     */
    if (signal(SIGINT, sighandler) == SIG_ERR ||
	signal(SIGTERM, sighandler) == SIG_ERR)
    {
	elog(LOGERROR, "Failed to install signal handlers.");
        exit(EXIT_FAILURE);
    }
    if (pipe(termpipe) != 0) {
	elog(LOGERROR, "Failed to create pipe: %s", strerror(errno));
        exit(EXIT_FAILURE);
    }

    
    /*
     * Initialise BruteFIR subsystem handler
     */
    snprintf(path, PATH_MAX, "%s.bfconv", conf_static.log_path);
    path[PATH_MAX - 1] = '\0';
    progress.init_phase(VCU_INIT_PHASE_CONVOLVER_SUBSYSTEM);
    if (!bfir_init(conf_static.allocate_ram * 1024 * 1024,
                   conf_static.log_path,
		   conf_static.brutefir_path,
		   conf_static.brutefir_modules_path,
                   path,
                   conf_static.input.n_devices,
		   conf_static.input.device,
		   conf_static.input.device_params,
		   conf_static.input.open_channels,
		   conf_static.input.channel_index,
                   conf_static.output.n_devices,
		   conf_static.output.device,
		   conf_static.output.device_params,
		   conf_static.output.open_channels,
		   conf_static.output.channel_index,
                   min_blocksize,
                   conf_static.allow_poll_mode))
    {
	exit(EXIT_FAILURE);
    }
    
    init_load_custom_eq_programs();
    
    update_hard_settings(false);
    update_bfir_soft_settings();

    /*
     * Initialise equalisers
     */
    memset(equalisers, 0, sizeof(equalisers));
    for (n = 0; n < cu.eq.eqprg_count; n++) {
        if ((err = new_equaliser(&eq, cu.eq.eqprg[n].parts,
                                 cu.eq.eqprg[n].part_count)) < 0)
        {
            elog(LOGDEBUG, "Failed to initialise equaliser: %s.",
                 vcu.ctl.strerror(err));
            cu.eq.eqprg_count = n;
            break;
        }
        equalisers[n] = eq;
    }

    if (eq_get_equaliser_count() > 0) {
	progress.init_phase(VCU_INIT_PHASE_GENERATING_EQUALISERS);
    }

    while (generate_equalisers(&n) != 0) {
	vcu.eq.delete_equaliser(n);
    }
    
    if (cu.channel.count[OUT][VCU_CH_AMBISONIC] > 0) {
	progress.init_phase(VCU_INIT_PHASE_GENERATING_AMBISONIC);
        generate_ambisonic_filters();
    }

    /*
     * Load other filter programs
     */
    init_load_stored_filter_programs();
    clean_fprog_refs();
    
    progress.init_phase(VCU_INIT_PHASE_FINISHED);
    /*
     * Enter the select loop
     */
    if (maxfd == -1) {
	elog(LOGERROR, "No fd registered.");
	exit(EXIT_FAILURE);
    }
    if (maxfd < termpipe[0]); {
        maxfd = termpipe[0];
    }
    while (true) {
	memcpy(&rfds, &readfds, sizeof(rfds));
	tv.tv_sec = 5;
	tv.tv_usec = 0;	
	gettimeofday(&tvnow, NULL);
	timeradd(&tvnow, &tv, &tv);
	n = -1;
	while ((n = bit_find(cb_timers_tab, n + 1, MAX_TIMERS - 1)) != -1) {
	    if (timercmp(&tv, &callback_timers[n].tvwhen, >)) {
		tv = callback_timers[n].tvwhen;
	    }
	}
	if (timercmp(&tvnow, &tv, >)) {
	    tv.tv_sec = 0;
	    tv.tv_usec = 0;
	} else {
	    timersub(&tv, &tvnow, &tv);
	}
        FD_SET(termpipe[0], &rfds);
	ret = select_function(maxfd + 1, &rfds, NULL, NULL, &tv);
	switch (ret) {
	case -1:
	    if (errno == EAGAIN || errno == EINTR) {
		break;
	    }
	    elog(LOGERROR, "Select failed: %s.\n", strerror(errno));
	    exit(EXIT_FAILURE);
	    break;
	case 0:
	default:
            if (FD_ISSET(termpipe[0], &rfds)) {
                elog(LOGDEBUG, "Shutting down.");
                pre_exec_cleanup();
                exit(EXIT_SUCCESS);
            }
	    periodic();
	    gettimeofday(&tvnow, NULL);
	    n = -1;
	    while ((n = bit_find(cb_timers_tab, n + 1, MAX_TIMERS - 1)) != -1) {
		if (timercmp(&tvnow, &callback_timers[n].tvwhen, >)) {
		    tv.tv_sec = 0;
		    tv.tv_usec = 0;
		} else {
		    timersub(&callback_timers[n].tvwhen, &tvnow, &tv);
		}
		if (tv.tv_sec == 0 && tv.tv_usec < 50000) {
		    bit_clr(cb_timers_tab, n);
		    bit_set(cb_timers_freetab, n);
		    if (callback_timers[n].callback != NULL) {
			callback_timers[n].callback(callback_timers[n].param);
		    }
		}
	    }
	    if (ret == 0) {
		break;
	    }
	    n = -1;
	    while ((n = bit_find((uint32_t *)((void *)&rfds), n + 1,
                                 maxfd)) != -1)
            {
		if (!FD_ISSET(n, &readfds)) {
		    continue;
		}
		if (n == bf_fd) {
		    bfir_process();
		} else {
		    fd_callbacks[n](n);
		}
	    }
	    break;
	}
    }
    
    return 0;
}
